#include "prefix.h"

#include <assert.h>
#include <deque>
#include <math.h>
#include <algorithm>
#include <new>

#include "fpoint.h"
#include "bezier.h"
#include "rect.h"
#include "knot.h"
#include "part.h"

/*************************************************

	xWGȐ𕪊NX

**************************************************/

BZP_NAMESPACE_START

class StackData	//X^bN̌^
{
public:
	StackData() {
	}
	StackData(const StackData &source)
		:m_forwardA(source.m_forwardA)
		,m_backB(source.m_backB)
		,m_anchorB(source.m_anchorB)
		,m_b_t(source.m_b_t)
	{

	}
	FPoint m_forwardA;
	FPoint m_backB;
	FPoint m_anchorB;
	double m_b_t;
};
typedef std::deque<StackData> Stack;

#ifdef min
#undef min
#endif
#ifdef max
#undef max
#endif

class BezierMember
{
public:
	BezierMember()
		:anchorA(0,0)
		,forwardA(0,0)
		,backB(0,0)
		,anchorB(0,0)
	{

	}
	Stack stack;	//܂ĂȂ̕
	FPoint anchorA;	//O̒ʉߓ_
	FPoint forwardA;	//O̒ʉߓ_̑O̕_
	FPoint backB;		//̒ʉߓ_̌̕_	
	FPoint anchorB;	//̒ʉߓ_
};


Bezier::Bezier(const FPoint & anchorA,const FPoint & forwardA,const FPoint & backB,const FPoint & anchorB)
	:m_m(new BezierMember)
	,m_a_t(0.0)
	,m_b_t(1.0)
{
	BZP_BAD_ALLOC_TEST(m_m.get());

	m_m->anchorA=anchorA;
	m_m->forwardA=forwardA;
	m_m->backB=backB;
	m_m->anchorB=anchorB;
}

Bezier::Bezier(const Knot *knotA,const Knot *knotB)
	:m_m(new BezierMember)
	,m_a_t(0.0)
	,m_b_t(1.0)
{
	BZP_BAD_ALLOC_TEST(m_m.get());

	m_m->anchorA=knotA->GetPoint(PART_ANCHOR);
	m_m->forwardA=knotA->GetPoint(PART_FORWARDCONTROL);
	m_m->backB=knotB->GetPoint(PART_BACKCONTROL);
	m_m->anchorB=knotB->GetPoint(PART_ANCHOR);
}

Bezier::~Bezier() throw()
{
}

//A̒ʉߓ_̍W𓾂
FPoint Bezier::GetAnchorA() const
{
	return m_m->anchorA;
}

//ȂO̕_𓾂
FPoint Bezier::GetForwardA() const
{
	return m_m->forwardA;
}

//B̌̕_𓾂
FPoint Bezier::GetBackB() const
{
	return m_m->backB;
}

//B̒ʉߓ_̍W𓾂
FPoint Bezier::GetAnchorB() const
{
	return m_m->anchorB;
}


//_xWGȐ̑{͈͓ɂ\
bool Bezier::IsPointInArea(const FPoint & point,const FPoint & area) const
{
	if(point.x>MaxX()+area.x/2.0) return false;
	if(point.x<MinX()-area.x/2.0) return false;
	if(point.y>MaxY()+area.y/2.0) return false;
	if(point.y<MinY()-area.y/2.0) return false;
	return true;
}

//Ȑlp̂Ƃɂ邩
bool Bezier::IsOutRect(const FRect &rect) const
{
	FRect bzRect;
	SetMaxRect(&bzRect);	

	return !bzRect.Intersect(rect);
}

//Ȑlp̒ɂ邩
bool Bezier::IsInRect(const FRect &rect) const
{
	FRect bzRect;
	SetMaxRect(&bzRect);	

	return bzRect.IsInRect(rect);
}

//ȐƂ肤ő̋`
void  Bezier::SetMaxRect(FRect *rect) const
{
	rect->Set(MinX(),MinY(),MaxX(),MaxY());
}

//xWGȐ̍ł傫ꍇ̑傫[ɏ
bool Bezier::IsNarrow(const FPoint & maxSize) const
{
	return MaxWidth()<maxSize.x && MaxHeight() < maxSize.y;
}

//xWGȐ̉ۂɂ͂ꍇ
double Bezier::MaxWidth() const
{
	return MaxX()-MinX();
}

//xWGȐ̏cۂɂ͂ꍇ
double Bezier::MaxHeight() const
{
	return MaxY()-MinY();
}

//w肵𑜓xŉ2hbgȓɎ܂Ă邩
bool Bezier::IsNarrowWidthByDot(double resolution) const
{
	double width=floor(MaxX()*resolution)
		-floor(MinX()*resolution);
	return width<=1;
}

//w肵𑜓xŏc2hbgȓɎ܂Ă邩
bool Bezier::IsNarrowHeightByDot(double resolution) const
{
	double  height=floor(MaxY()*resolution)
		-floor(MinY()*resolution);
	return height<=1;
}

//w肵cʒuʂ\邩
bool Bezier::DoesPassLine(double y) const
{
	return MinY()<=y && y<=MaxY();
}

//Ȑ̒ʂ肤ő̉ʒu
double Bezier::MaxX() const
{
	return std::max(std::max(m_m->anchorA.x,m_m->forwardA.x)
		,std::max(m_m->backB.x,m_m->anchorB.x));
}

//Ȑ̒ʂ肤ŏ̉ʒu
double Bezier::MinX() const
{
	return std::min(std::min(m_m->anchorA.x,m_m->forwardA.x)
		,std::min(m_m->backB.x,m_m->anchorB.x));
}

//Ȑ̂Ƃ肤ő̏cʒu
double Bezier::MaxY() const
{
	return std::max(std::max(m_m->anchorA.y,m_m->forwardA.y)
		,std::max(m_m->backB.y,m_m->anchorB.y));
}

//Ȑ̂Ƃ肤ŏ̏cʒu
double Bezier::MinY() const
{
	return std::min(std::min(m_m->anchorA.y,m_m->forwardA.y)
		,std::min(m_m->backB.y,m_m->anchorB.y));
}

//ɂȂĂ邩
//mɋȐł͂Ȃꍇ̂ݡ{͒łfalseɂȂꍇ
bool Bezier::IsNotCurve() const
{
	return m_m->anchorA==m_m->forwardA && m_m->backB==m_m->anchorB;
}

//xWGȐ̕
void Bezier::Divide(double t)
{
	assert(0.0<=t&&t<=1.0);
	FPoint forward2A;	//ȂO_
	forward2A.TPoint(m_m->anchorA,m_m->forwardA,t);

	FPoint back2B;	//B̌_
	back2B.TPoint(m_m->backB,m_m->anchorB,t);

	FPoint centerAB;
	centerAB.TPoint(m_m->forwardA,m_m->backB,t);

	FPoint back2A;	//Ǎ_
	back2A.TPoint(forward2A,centerAB,t);

	FPoint forward2B;	//B̑O_
	forward2B.TPoint(centerAB,back2B,t);

	FPoint anchorAB;	//^񒆂̒ʉߓ_
	anchorAB.TPoint(back2A,forward2B,t);

	//BX^bNɐς
	StackData data;
	data.m_forwardA=forward2B;
	data.m_backB=back2B;
	data.m_anchorB=m_m->anchorB;
	data.m_b_t=m_b_t;
	m_m->stack.push_back(data);

	//AVȌ̂߂ɏ
	m_m->forwardA=forward2A;
	m_m->backB=back2A;
	m_m->anchorB=anchorAB;
	m_b_t=m_a_t*(1.0-t)+m_b_t*t;
}

//̒ʉߓ_Ɠ_ƂbaseSizeƂ 
double Bezier::Distance(const FPoint & point,const FPoint & baseSize) const
{
	return hypot((point.x-m_m->anchorB.x)/baseSize.x
		,(point.y-m_m->anchorB.y)/baseSize.y);
}

//X^bNɎcĂ邽߂ɏ
bool Bezier::Next()	
{
	if(m_m->stack.empty()) {
		return false;
	}

	StackData data;
	data=*(m_m->stack.rbegin());
	m_m->stack.pop_back();

	m_m->anchorA=m_m->anchorB;
	m_m->forwardA=data.m_forwardA;
	m_m->backB=data.m_backB;
	m_m->anchorB=data.m_anchorB;
	m_a_t=m_b_t;
	m_b_t=data.m_b_t;
	return true;
}

BZP_NAMESPACE_END
