#include "prefix.h"

#include <new>

#include "clip.h"
#include "path.h"
#include "convertertable.h"
#include "dimptr.h"
#include "part.h"
#include "partmask.h"
#include "convertertag.h"
#include "rect.h"
#include "fpoint.h"
#include "selectmode.h"
#include "exception.h"
#include "error.h"

/*************************************************

	pX𕡐؂

**************************************************/

BZP_NAMESPACE_START

void CopyParts(Path *destPath,const Path *srcPath,int part)
{
	BZP_PARAMETER_ERROR(destPath!=NULL && srcPath!=NULL
		,BZPERROR_NULL_ARGUMENT);

	int size;
	size=PathToTableSize(srcPath,part);

	DimPtr <PATHTABLE> buffer(reinterpret_cast<PATHTABLE*>(new char[size]));
	BZP_BAD_ALLOC_TEST(buffer.get());

	PathToTable(srcPath,part,buffer.get(),size);
	TableToPath(destPath,buffer.get(),size);
}

int ClipPart(const Path *srcPath,bool cutKnot)
{
	BZP_PARAMETER_ERROR(srcPath!=NULL,BZPERROR_NULL_ARGUMENT);
	
	int part=PART_ALL;
	if(srcPath->SelectedKnotCount()>0 && cutKnot) {
		part=PART_KNOT|PART_SUBGROUP;
	}
	return part;
}

//IĂpẌꕔ𕡐
void Copy(Path *destPath,const Path *srcPath,bool cutKnot)
{
	BZP_PARAMETER_ERROR(destPath!=NULL && srcPath!=NULL
		,BZPERROR_NULL_ARGUMENT);

	int part=ClipPart(srcPath,cutKnot)|PART_SELECTED;
	CopyParts(destPath,srcPath,part);
}

//IĂpẌꕔ폜
void Cut(Path *destPath,const Path *srcPath,bool cutKnot)
{
	BZP_PARAMETER_ERROR(destPath!=NULL && srcPath!=NULL
		,BZPERROR_NULL_ARGUMENT);

	int part=ClipPart(srcPath,cutKnot)|PART_UNSELECTED;
	CopyParts(destPath,srcPath,part);
}

//\t
void Paste(Path *destPath,const Path *clipboardPath,SIZE imageSize)
{
	BZP_PARAMETER_ERROR(destPath!=NULL,BZPERROR_NULL_ARGUMENT);

	//pX摜͈͓̔ɂȂΈʒu炷
	FRect clipRect;
	clipboardPath->GetOuterRect(&clipRect,PART_CURVE|PART_SUBGROUP,1.0);

	FRect imageRect(0,0,imageSize.cx,imageSize.cy);

	Path clipboardPath2;
	clipboardPath2.Copy(clipboardPath);
	if(!imageRect.Intersect(clipRect)) {
		clipboardPath2.Move(-clipRect.TopLeft());
	}

	destPath->SelectCurve(SELECTMODE_UNSELECTALL,0);
	destPath->SelectSubGroup(SELECTMODE_UNSELECTALL,0);
	destPath->Append(&clipboardPath2);
}

//IĂ镔Ał邩
bool CanCopy(const Path *path)
{
	BZP_PARAMETER_ERROR(path!=NULL,BZPERROR_NULL_ARGUMENT);

	return path->SelectedCurveCount()
		+path->SelectedSubGroupCount()>0;
}

//TuO[vł邩
bool CanGroup(const Path *path)
{
	BZP_PARAMETER_ERROR(path!=NULL,BZPERROR_NULL_ARGUMENT);

	return (path->SelectedCurveCount()
		+path->SelectedSubGroupCount()>=2);
}

//TuO[v΂点邩
bool CanUngroup(const Path *path)
{
	BZP_PARAMETER_ERROR(path!=NULL,BZPERROR_NULL_ARGUMENT);

	return path->SelectedSubGroupCount()>0;
}

BZP_NAMESPACE_END
