#include "prefix.h"

#include "converterai.h"
#include "knotmode.h"
#include "error.h"
#include "exception.h"
#include "fpoint.h"
#include "converterbuffer.h"
#include "path.h"
#include "rect.h"
#include "runtimeconfig.h"

#include <stdlib.h>
#include <ctype.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>

/*************************************************

	Illustlator`(*.ai)ɕϊ

**************************************************/

BZP_NAMESPACE_START

BZP_CONFIG_USING

////////////o///////////////

class AIWriter : public PathToBuffer
{
public:
	AIWriter();
	virtual ~AIWriter() throw();

	virtual void WriteStartPath(const Rect &boundRect);
	virtual void WriteEndPath();
	virtual void WriteStartCurve(bool isSelected,bool isClosed);
	virtual void WriteEndCurve();
	virtual void WriteStartSubGroup(bool isSelected);
	virtual void WriteEndSubGroup();
	virtual void WriteKnot(const FPoint & back,const FPoint & anchor,const FPoint & forward,int mode);

protected:
	void WriteFPointString(const FPoint &point);
	void WriteString(const char *format,...);
private:
	bool m_firstKnot;
	bool m_isCurveClosed;
	int m_lastMode;
	FPoint m_lastForward;

	FPoint m_closerAnchor;
	FPoint m_closerBack;
	int m_closerMode;
};

///////////////////////////////////////////////////////

AIWriter::AIWriter()
{
	
}

AIWriter::~AIWriter() throw()
{
}

//pX̎n܂̏o
void AIWriter::WriteStartPath(const Rect &boundRect)
{
#ifndef NDEBUG
	PathToBuffer::WriteStartPath(boundRect);
#endif

	SetBoundRect(boundRect);

	WriteString("%%!PS-Adobe-3.0\r"
		"%%%%Creator: BZPath Library\r"
		"%%%%BoundingBox: %d %d %d %d\r"
		"%%%%EndComments\r"
		"%%%%BeginProlog\r"
		"%%%%EndProlog\r"
		"%%%%BeginSetup\r"
		"%%%%EndSetup\r"
		"0 0 0 0 k\r"
		"1 1 1 0 K\r"
		,0,0,boundRect.Width(),boundRect.Height());
}

//pX̏Ȉo
void AIWriter::WriteEndPath()
{
#ifndef NDEBUG
	PathToBuffer::WriteEndPath();
#endif
	WriteString("%%%%Trailer\r"
		"%%%%EOF\r");
}

//̎n܂̏o
void AIWriter::WriteStartCurve(bool isSelected,bool isClosed)
{
#ifndef NDEBUG
	PathToBuffer::WriteStartCurve(isSelected,isClosed);
#endif
	m_firstKnot=true;
	m_isCurveClosed=isClosed;
}


//̏Ȉo
void AIWriter::WriteEndCurve()
{
#ifndef NDEBUG
	PathToBuffer::WriteEndCurve();
#endif

	int curveTerm=GetConfigInt(AI_CURVE_TERMINAL);
	if(m_isCurveClosed) {
		FPoint dummy(0,0);
		WriteKnot(m_closerBack,m_closerAnchor,dummy,m_closerMode);
		curveTerm=tolower(curveTerm);
	} else {
		curveTerm=toupper(curveTerm);
	}
	WriteChar(curveTerm);
	WriteChar('\r');
}

//TuO[v̎n܂̏o
void AIWriter::WriteStartSubGroup(bool isSelected)
{
#ifndef NDEBUG
	PathToBuffer::WriteStartSubGroup(isSelected);
#endif
	WriteString("u\r");
}

//TuO[v̏Ȉo
void AIWriter::WriteEndSubGroup()
{
#ifndef NDEBUG
	PathToBuffer::WriteEndSubGroup();
#endif
	WriteString("U\r");
}

//pڂ̏o
void AIWriter::WriteKnot(const FPoint & back,const FPoint & anchor,const FPoint & forward,int mode)
{
#ifndef NDEBUG
	PathToBuffer::WriteKnot(back,anchor,forward,mode);
#endif
	if(!m_firstKnot && (m_lastMode&KNOT_CURVED_FORWARD)!=0) {
		WriteFPointString(m_lastForward);
	}
	if(!m_firstKnot && (mode&KNOT_CURVED_BACK)!=0) {
		WriteFPointString(back);
	}
	WriteFPointString(anchor);

	char mark;
	if(m_firstKnot) {
		mark='m';
		m_firstKnot=false;

		if(m_isCurveClosed) {
			m_closerAnchor=anchor;
			m_closerBack=back;
			m_closerMode=mode;
		}
	} else if((m_lastMode&KNOT_CURVED_FORWARD)!=0) {
		if((mode&KNOT_CURVED_BACK)!=0) {
			if((mode&KNOT_LINKED)!=0) {
				mark='c';
			} else {
				mark='C';
			}
		} else {
			mark='Y';
		}
	} else if((mode&KNOT_CURVED_BACK)!=0) {
		if((mode&KNOT_LINKED)!=0) {
			mark='v';
		} else {
			mark='V';
		}
	} else {
		mark='l';
	}
	m_lastMode=mode;
	m_lastForward=forward;

	WriteString("%c\r",mark);
}

//W̏o
void AIWriter::WriteFPointString(const FPoint & point)
{
	WriteString("%.4f %.4f ",point.x-m_boundRect->left,m_boundRect->bottom-point.y);
}

//t̏o
void AIWriter::WriteString(const char *format,...)
{
	assert(format!=NULL);

	va_list arguments;
	va_start (arguments,format);

	int size=256;
	char *buffer=NULL;
	FOR(;;) {
		char *newBuffer=(char*)realloc(buffer,size);
		if(newBuffer==NULL) {
			free(buffer);
			throw Exception(BZPERROR_OUT_OF_MEMORY);
		}
		buffer=newBuffer;
		int outSize=_vsnprintf(buffer,size,format,arguments);
		if(outSize>=0) {
			size=outSize;
			break;
		}
		size+=256;
	}

	if(size>0) {
		try {
			WriteData(buffer,size);
		}
		catch(...) {
			free(buffer);
			throw;
		}
	}
	free(buffer);
	va_end(arguments);
}

//////////////////ǂݎApX쐬///////////////////

class AIReader : public BufferToPath
{
public:
	AIReader();
	virtual ~AIReader() throw();

	virtual void ReadStartPath(Rect *boundRect);
	virtual void ReadStartSubGroup();
	virtual void ReadEndSubGroup(bool *isSelected);
	virtual void ReadEndCurve(bool *isSelected,bool *isCurved,bool *isOverlappedCloser);
	virtual void ReadKnot(FPoint *back,FPoint *anchor,FPoint *forward,int *mode);
	virtual bool IsSubGroupOpened();
	virtual bool IsCurve();
	virtual bool IsKnot();
protected:
	void ReadProlog();
	void ReadHeader();
	void SkipScriptSetup();
	bool CheckKeyword(const char *keyword) const;
	void SkipToKeyword(const char *keyword);
	void ReadBoundingBox(Rect *rect);
	void NextLine();
	void ReadOperator2(const FPoint &back,const FPoint &anchor
		,FPoint *forward,int *mode);
	char CheckOperator();
	const bool IsNotIgnoreOperator(char ope) const;
	void SkipIgnoreOperator();
	int Read3FPointString(FPoint *arg1,FPoint *arg2,FPoint *arg3) const;
	void ReverseFPoint(FPoint *point) const;
	void SkipSpace();
private:
};

AIReader::AIReader()
{
}

AIReader::~AIReader() throw()
{
}

//pX̎n܂
void AIReader::ReadStartPath(Rect *boundRect)
{
#ifndef NDEBUG
	BufferToPath::ReadStartPath(boundRect);
#endif
	assert(boundRect!=NULL);

	ReadProlog();
	boundRect->Set(*m_boundRect);
	SkipScriptSetup();
}

//wb_ǂ
void AIReader::ReadProlog()
{
	ReadHeader();
	SkipToKeyword("%%EndProlog");
	NextLine();
}

//擪EndComments܂šBoundingBoxǂޡ
void AIReader::ReadHeader()
{
	SkipToKeyword("%%BoundingBox:");
	ReadBoundingBox(m_boundRect.get());
	SkipToKeyword("%%EndComments");
	NextLine();
}

//{̑̕Ou΂
void AIReader::SkipScriptSetup()
{
	SkipToKeyword("%%BeginSetup");
	NextLine();
	SkipToKeyword("%%EndSetup");
	NextLine();
}

//Ɠ
bool AIReader::CheckKeyword(const char *keyword) const
{
	assert(keyword!=NULL);
	int keyLength=strlen(keyword);
	if(GetPointer()+keyLength>Size()) {
		throw Exception(BZPERROR_READ_BEYOND_BUFFER);
	}

	if(strncmp(keyword,GetCurrent(),keyLength)!=0) {
		return false;
	}

	return true;
}

//Ɠs܂Ŕ΂
void AIReader::SkipToKeyword(const char *keyword)
{
	while(!CheckKeyword(keyword)) {
		NextLine();
	}
}

//`ǂ
void AIReader::ReadBoundingBox(Rect *rect)
{
	assert(rect!=NULL);

	int count=sscanf(GetCurrent(),"%*s %d %d %d %d"
		,&(rect->left)
		,&(rect->top)
		,&(rect->right)
		,&(rect->bottom) );
	if(count!=4) {
		throw Exception(BZPERROR_ILLEGAL_DATA_FORMAT);
	}
	NextLine();
}

//̍sɐi߂
void AIReader::NextLine()
{
	int length=strcspn(GetCurrent(),"\r\n");
	IncrementPointer(length+1);
}

//TuO[vn܂邩
bool AIReader::IsSubGroupOpened()
{
#ifndef NDEBUG
	BufferToPath::IsSubGroupOpened();
#endif
	SkipIgnoreOperator();
	return CheckOperator()=='u';
}

//TuO[v̎n܂
void AIReader::ReadStartSubGroup()
{
#ifndef NDEBUG
	BufferToPath::ReadStartSubGroup();
#endif
	assert(CheckOperator()=='u');
	NextLine();
}

//TuO[v̏I
void AIReader::ReadEndSubGroup(bool *isSelected)
{
#ifndef NDEBUG
	BufferToPath::ReadEndSubGroup(isSelected);
#endif
	*isSelected=false;
	SkipIgnoreOperator();
	if(CheckOperator()!='U') {
		throw Exception(BZPERROR_ILLEGAL_DATA_FORMAT);
	}
	NextLine();
}

//n܂邩
bool AIReader::IsCurve()
{
#ifndef NDEBUG
	BufferToPath::IsCurve();
#endif
	SkipIgnoreOperator();
	return CheckOperator()=='m';
}

//̏I
void AIReader::ReadEndCurve(bool *isSelected,bool *isClosed,bool *isOverlappedCloser)
{
#ifndef NDEBUG
	BufferToPath::ReadEndCurve(isSelected,isClosed,isOverlappedCloser);
#endif
	*isSelected=false;
	SkipIgnoreOperator();
	char ope=CheckOperator();
	switch(ope) {
	case 'n':	case 'f':	case 's':	case 'b':
		*isClosed=true;
		break;

	case 'N':	case 'F':	case 'S':	case 'B':
		*isClosed=false;
		break;
	}
	*isOverlappedCloser=true;	//Ō̓_͍ŏ̓_ƏdȂ
	NextLine();
}

//pڂ
bool AIReader::IsKnot()
{
#ifndef NDEBUG
	BufferToPath::IsKnot();
#endif
	SkipIgnoreOperator();
	char ope=CheckOperator();
	switch(ope) {
	case 'm':
	case 'l':	case 'L':
	case 'c':	case 'C':
	case 'v':	case 'V':
	case 'y':	case 'Y':
		return true;
	}
	return false;
}

//p
void AIReader::ReadKnot(FPoint *back,FPoint *anchor,FPoint *forward,int *mode)
{
#ifndef NDEBUG
	BufferToPath::ReadKnot(back,anchor,forward,mode);
#endif
	char ope=CheckOperator();
	switch(ope) {
	case 'm':
	case 'l':
	case 'y':
		*mode=KNOT_LINKED;
		break;

	case 'L':
	case 'Y':
		*mode=0;

	case 'c':
	case 'v':
		*mode=KNOT_CURVED_BACK|KNOT_LINKED;
		break;
	case 'C':
	case 'V':
		*mode=KNOT_CURVED_BACK;
		break;
	}

	FPoint dummy;
	FPoint *arg1,*arg2,*arg3;
	int argCount;

	switch(ope) {
	case 'm':
	case 'l':	case 'L':
		arg1=anchor;
		arg2=arg3=&dummy;
		argCount=2;
		break;

	case 'y':	case 'Y':	
		arg2=anchor;
		arg1=arg3=&dummy;
		argCount=4;
		break;

	case 'v':	case 'V':
		arg1=back;
		arg2=anchor;
		arg3=&dummy;
		argCount=4;
		break;

	case 'c':	case 'C':
		arg1=&dummy;
		arg2=back;
		arg3=anchor;
		argCount=6;
		break;
	}

	int readCount=Read3FPointString(arg1,arg2,arg3);
	if(readCount!=argCount) {
		throw Exception(BZPERROR_ILLEGAL_DATA_FORMAT);
	}

	NextLine();
	ReadOperator2(*back,*anchor,forward,mode);

	switch(ope) {
	case 'm':
		*back=anchor->OppositePoint(*forward);
		if((*mode&KNOT_CURVED_FORWARD)!=0) {
			*mode|=KNOT_CURVED_BACK;
		}
		break;

	case 'l':	case 'L':
	case 'y':	case 'Y':	
		*back=*anchor;
		break;
	}
}

//̖߂̓ǂ
void AIReader::ReadOperator2(const FPoint &back,const FPoint &anchor
		,FPoint *forward,int *mode)
{
	SkipIgnoreOperator();
	char ope=CheckOperator();

	FPoint dummy;
	switch(ope) {
	case 'l':	case 'L':
	case 'v':	case 'V':
		*forward=anchor;
		*mode&=~KNOT_LINKED;
		break;

	case 'c':	case 'C':
	case 'y':	case 'Y':
		Read3FPointString(forward,&dummy,&dummy);
		*mode|=KNOT_CURVED_FORWARD;
		break;

	default:
		*forward=anchor.OppositePoint(back);
		if((*mode&KNOT_CURVED_BACK)!=0) {
			*mode|=KNOT_CURVED_FORWARD;
		}
		break;
	}
}

//̖ߡ
char AIReader::CheckOperator()
{
	const char *ptr=GetCurrent();
	SkipSpace();
	if(*ptr=='%') {
		return '%';
	}

	int length=strcspn(ptr,"\r\n");
	const char *r_ptr=ptr+length;

	while(!isalpha(*r_ptr) && ptr!=r_ptr) {
		--r_ptr;
	}
	if(IsNotIgnoreOperator(*r_ptr)) {
		return *r_ptr;
	}
	return 0;
}

void AIReader::SkipSpace()
{
	const char *ptr=GetCurrent();
	int length=0;
	while(isspace(ptr[length])) {
		++length;
	}
	IncrementPointer(length);
}

//ǂݔ΂Ȃ߂
const bool AIReader::IsNotIgnoreOperator(char ope) const
{
	switch(ope) {
	case 'm':
	case 'l':	case 'L':
	case 'c':	case 'C':
	case 'v':	case 'V':
	case 'y':	case 'Y':
	case 'n':	case 'f':	case 's':	case 'b':
	case 'N':	case 'F':	case 'S':	case 'B':
	case 'u':	case 'U':
		return true;
	}
	return false;
}

//閽߂΂
void AIReader::SkipIgnoreOperator()
{
	FOR(;;) {
		SkipSpace();
		char ope=CheckOperator();
		if(ope=='%') {
			if(CheckKeyword("%%Trailer")) {
				return;
			}
		} else if(ope!=0) {
			return;
		}
		NextLine();
	}
}

//̍W3ǂ
int AIReader::Read3FPointString(FPoint *arg1,FPoint *arg2,FPoint *arg3) const
{
	int count=sscanf(GetCurrent()
		," %lf %lf %lf %lf %lf %lf"
		,&(arg1->x),&(arg1->y)
		,&(arg2->x),&(arg2->y)
		,&(arg3->x),&(arg3->y)	);

	ReverseFPoint(arg1);
	ReverseFPoint(arg2);
	ReverseFPoint(arg3);	

	return count;
}

//ai͍W㉺tȂ̂łЂԂ
void AIReader::ReverseFPoint(FPoint *point) const
{
	point->y=m_boundRect->bottom - point->y;	
}
			
///////////////////Ăяo֐/////////////////

/*ϊ̃f[^\̃oCgvZ*/
int PathToAISize(const Path *path,int part)
{
	BZP_PARAMETER_ERROR(path!=NULL,BZPERROR_NULL_ARGUMENT);

	AIWriter writer;
	path->PathToData(&writer,part);
	return writer.Size();
}

/*pXAI`ɕϊ*/
int PathToAI(const Path *path,int part,char *buffer,int maxSize)
{
	BZP_PARAMETER_ERROR(path!=NULL && buffer!=NULL
		,BZPERROR_NULL_ARGUMENT);

	AIWriter writer;
	writer.SetBuffer(buffer,maxSize);
	path->PathToData(&writer,part);

	return writer.Size();
}

/*AI`pXɂ*/
void AIToPath(Path *path,const char *buffer,int size)
{
	BZP_PARAMETER_ERROR(path!=NULL && buffer!=NULL
		,BZPERROR_NULL_ARGUMENT);

	AIReader reader;
	reader.SetBuffer(buffer,size);
	path->DataToPath(&reader);
}

BZP_NAMESPACE_END
