#include "prefix.h"

#include <string.h>

#include "converterbuffer.h"
#include "error.h"
#include "exception.h"
#include "rect.h"
#include "path.h"

/*************************************************

	pXƉ炩̌`ɕϊăobt@[ɓ

**************************************************/

/*͊OȂ̂ŃoCgςȂ悤ɂ*/
#define LONGSIZE 4		/*LONG^̑傫*/
#define SHORTSIZE 2		/*SHORT^̑傫*/
#define CHARSIZE 1		/*char^̑傫*/
#define CHARBITS 8		/*char^̃rbg*/
#define FIXEDPOINT_MAX 15	/*Œ菬_̍ől*/
#define FIXEDPOINT_MIN -16	/*Œ菬_̍ŏl*/
#define FIXEDPOINT_INTSIZE 1	/*Œ菬_̐̑傫*/
#define FIXEDPOINT_FLOATSIZE 3	/*Œ菬_̏̑傫*/
#define FIXEDPOINT_FLOATMULTI (1<<(CHARBITS*FIXEDPOINT_FLOATSIZE))	/*Œ菬_̏𐮐ɕϊ鎞̔{*/
#define FIXEDPOINT_FLOATMAX (FIXEDPOINT_FLOATMULTI-1) /*Œ菬_̏𐮐ɕϊƂ̍ől*/

BZP_NAMESPACE_START

////////////////PathToBufferNX///////////////////

PathToBuffer::PathToBuffer()
	:m_buffer(NULL)
	,m_bufferSize(0)
	,m_pointer(0)
{
}

PathToBuffer::~PathToBuffer() throw()
{
}

//obt@̐ݒ
void PathToBuffer::SetBuffer(char *buffer,int size)
{
	assert(buffer!=NULL);
	m_buffer=buffer;
	m_bufferSize=size;
}

//Kvȃobt@ʂ𓾂
int PathToBuffer::Size() const
{
	return GetPointer();
}

//݂̓ǂݏ̈ʒu𓾂
LONG PathToBuffer::GetPointer() const
{
	return m_pointer;
}

//Kvȃobt@ʂvZ邾ŁAobt@ւ̏oȂ
bool PathToBuffer::CountOnly() const
{
	return m_buffer==NULL;
}

//f[^o
void PathToBuffer::WriteData(const char *data,int size,LONG offset)
{
	if(!CountOnly()) {
		if(offset<0) {	
			if(GetPointer()+size>m_bufferSize) {
				throw Exception(BZPERROR_NOT_ENOUGH_BUFFER_SIZE);
			}
		} else {
			assert(offset+size<=m_bufferSize);
		}
		assert(m_buffer!=NULL && data!=NULL);
		LONG pointer;
		if(offset<0) {
			pointer=GetPointer();
		} else {
			pointer=offset;
		}
		memcpy(m_buffer+pointer,data,size);
	}
	if(offset<0) {
		IncrementPointer(size);
	}
}

//LONG^̏o
void PathToBuffer::WriteLong(LONG value,LONG offset)
{
	assert(sizeof(LONG)>=LONGSIZE);
	SubWriteInt(value,LONGSIZE,offset);
}

//SHORT^̏o
void PathToBuffer::WriteShort(SHORT value)
{
	assert(sizeof(SHORT)>=SHORTSIZE);
	SubWriteInt(value,SHORTSIZE);
}

//char^o
void PathToBuffer::WriteChar(char value)
{
	assert(sizeof(char)>=CHARSIZE);
	SubWriteInt(value,CHARSIZE);
}

//bool^̏o
void PathToBuffer::WriteBool(bool flag)
{
	WriteChar(flag?TRUE:FALSE);
}

//̏o͂̉
void PathToBuffer::SubWriteInt(LONG value,int intSize,LONG offset)
{
	assert(intSize<=LONGSIZE);
	const int charBits=8;

	unsigned char valchar[LONGSIZE];

	FOR(int i=0;i<intSize;++i) {
		LONG shift=charBits*(intSize-1-i);
		valchar[i]=static_cast<unsigned char>( value>>shift );
	}
	WriteData(reinterpret_cast<char*>(valchar),intSize,offset);
}

//Œ菬_̍W̏o
void PathToBuffer::WriteFixedPointPoint(const FPoint & point)
{
	assert(m_boundRect->Width()>0);
	double relx=(point.x-m_boundRect->left)/(m_boundRect->right-m_boundRect->left);
	WriteFixedPoint(relx);

	assert(m_boundRect->Height()>0);
	double rely=(point.y-m_boundRect->top)/(m_boundRect->bottom-m_boundRect->top);
	WriteFixedPoint(rely);
}

//_Œ菬_ɂďo
void PathToBuffer::WriteFixedPoint(double value)
{
	assert(FIXEDPOINT_INTSIZE<sizeof(LONG)
		&& FIXEDPOINT_FLOATSIZE<sizeof(LONG));

	double valInt;
	valInt=floor(value);
	double valFloat;
	valFloat=value-valInt;

	LONG lValInt=static_cast<LONG>(valInt);
	LONG lValFloat=static_cast<int>(valFloat*FIXEDPOINT_FLOATMULTI);

	if(lValInt<FIXEDPOINT_MIN) {
		lValInt=FIXEDPOINT_MIN;
		lValFloat=0;
	} else if(lValInt>FIXEDPOINT_MAX) {
		lValInt=FIXEDPOINT_MAX;
		lValFloat=FIXEDPOINT_FLOATMAX;
	}

	SubWriteInt(lValInt,FIXEDPOINT_INTSIZE);
	SubWriteInt(lValFloat,FIXEDPOINT_FLOATSIZE);
}

//`o
void PathToBuffer::WriteRect(const Rect &rect)
{
	WriteLong(rect.left);
	WriteLong(rect.top);
	WriteLong(rect.right);
	WriteLong(rect.bottom);
}

//|C^i߂
void PathToBuffer::IncrementPointer(int size)
{
	m_pointer+=size;
	if(!CountOnly() && m_pointer>m_bufferSize) {
		throw Exception(BZPERROR_READ_BEYOND_BUFFER);
	}
}

////////////////BufferToPathNX///////////////////

BufferToPath::BufferToPath()
:m_buffer(NULL)
,m_bufferSize(0)
,m_pointer(0)
{

}

BufferToPath::~BufferToPath() throw()
{

}

//obt@̐ݒ
void BufferToPath::SetBuffer(const char *buffer,int size)
{
	assert(buffer!=NULL);
	m_buffer=buffer;
	m_bufferSize=size;
}

//݂̓ǂݏ̈ʒu𓾂
LONG BufferToPath::GetPointer() const
{
	return m_pointer;
}

//ݒ肵obt@̑傫
int BufferToPath::Size() const
{
	assert(m_buffer!=NULL);
	return m_bufferSize;
}

//obt@̃AhX
const char *BufferToPath::GetBuffer() const
{
	return m_buffer;
}

//obt@݂̌̃AhX
const char *BufferToPath::GetCurrent() const
{
	return GetBuffer()+GetPointer();
}

//f[^ǂ
void BufferToPath::ReadData(char *data,int size)
{
	CheckData(data,size);
	IncrementPointer(size);
}

//|C^i߂
void BufferToPath::IncrementPointer(int size)
{
	m_pointer+=size;
	if(m_pointer>m_bufferSize) {
		throw Exception(BZPERROR_READ_BEYOND_BUFFER);
	}
}

//f[^\dataɂ顃|C^͐i߂Ȃ
void BufferToPath::CheckData(char *data,int size) const
{
	assert(m_buffer!=NULL && data!=NULL);
	if(GetPointer()+size>m_bufferSize) {
		throw Exception(BZPERROR_READ_BEYOND_BUFFER);
	}

	memcpy(data,m_buffer+GetPointer(),size);
}

//LONG^𓾂
LONG BufferToPath::ReadLong()
{
	assert(sizeof(LONG)>=LONGSIZE);
	return SubReadInt(LONGSIZE);
}

//SHORT^𓾂
SHORT BufferToPath::ReadShort()
{
	assert(sizeof(SHORT)>=SHORTSIZE);
	return static_cast<SHORT>(SubReadInt(SHORTSIZE));	
}

//char^ǂ
char BufferToPath::ReadChar()
{
	assert(sizeof(char)>=CHARSIZE);
	return static_cast<char>(SubReadInt(CHARSIZE));
}

//bool^ǂ
bool BufferToPath::ReadBool()
{
	return ReadChar()!=FALSE;
}

//Ȃ
ULONG BufferToPath::SubReadUInt(int intSize)
{
	ULONG value=static_cast<ULONG>
		(SubReadInt(intSize));

	if(intSize<sizeof(LONG)) {
		value&=(1<<(CHARBITS*intSize))-1;
	}
	return value;
}

//ǂމ
LONG BufferToPath::SubReadInt(int intSize)
{
	LONG value=SubCheckInt(intSize);
	IncrementPointer(intSize);
	return value;
}

//ǂ݂鉺
LONG BufferToPath::SubCheckInt(int intSize) const
{
	assert(intSize<=LONGSIZE);

	unsigned char valchar[LONGSIZE];

	CheckData(reinterpret_cast<char*>(valchar),intSize);

	LONG value=0;
	if(static_cast<char>(valchar[0])<0) {
		value=-1;
	}

	FOR(int i=0;i<intSize;++i) {
		value<<=CHARBITS;
		value|=valchar[i];
	}

	return value;
}

//`𓾂
void BufferToPath::ReadRect(Rect *rect)
{
	assert(rect!=NULL);
	rect->left=ReadLong();
	rect->top=ReadLong();
	rect->right=ReadLong();
	rect->bottom=ReadLong();
}

//Œ菬_𓾂ĕ_ɂ
double BufferToPath::ReadFixedPoint()
{
	LONG lValInt=SubReadInt(FIXEDPOINT_INTSIZE);
	ULONG lValFloat=SubReadUInt(FIXEDPOINT_FLOATSIZE);
	double dValFloat=static_cast<double>(lValFloat)/(1<<(FIXEDPOINT_FLOATSIZE*CHARBITS));
	return lValInt+dValFloat;
}

//WŒ菬_瓾
FPoint BufferToPath::ReadFixedPointPoint()
{
	FPoint point;

	double relx=ReadFixedPoint();
	point.x=relx*(m_boundRect->right-m_boundRect->left)+m_boundRect->left;

	double rely=ReadFixedPoint();
	point.y=rely*(m_boundRect->bottom-m_boundRect->top)+m_boundRect->top;

	return point;
}

//SHORT^ǂ
SHORT BufferToPath::CheckShort() const
{
	return static_cast<SHORT>(SubCheckInt(SHORTSIZE));
}

//LONG^ǂ
LONG BufferToPath::CheckLong() const
{
	return SubCheckInt(LONGSIZE);
}

BZP_NAMESPACE_END
