#include "prefix.h"

#include "convertertable.h"
#include "knotmode.h"
#include "error.h"
#include "exception.h"
#include "fpoint.h"
#include "rect.h"
#include "converterbuffer.h"
#include "path.h"

/*************************************************

	e[u`ɕϊ

**************************************************/

BZP_NAMESPACE_START


//////////////o//////////////

class TableWriter : public PathToBuffer
{
public:
	TableWriter();
	virtual ~TableWriter() throw();

	virtual void WriteStartPath(const Rect &boundRect);
	virtual void WriteEndPath();
	virtual void WriteStartCurve(bool isSelected,bool isClosed);
	virtual void WriteStartSubGroup(bool isSelected);
	virtual void WriteEndSubGroup();
	virtual void WriteKnot(const FPoint & back,const FPoint & anchor,const FPoint & forward,int mode);

protected:
	void ClearTable(PATHTABLE *table) const;
	void WriteTable(const PATHTABLE &table);
private:
};

TableWriter::TableWriter()
{
}

TableWriter::~TableWriter() throw()
{

}

//pX̎n܂̏o
void TableWriter::WriteStartPath(const Rect &boundRect)
{
#ifndef NDEBUG
	PathToBuffer::WriteStartPath(boundRect);
#endif
	PATHTABLE table;
	ClearTable(&table);

	table.type=PATHTABLE_HEADER;
	table.header.boundRect=boundRect;

	WriteTable(table);
}

//pX̏Ȉo
void TableWriter::WriteEndPath()
{
#ifndef NDEBUG
	PathToBuffer::WriteEndPath();
#endif
	PATHTABLE table;

	ClearTable(&table);
	table.type=PATHTABLE_END;

	WriteTable(table);
}

//̎n܂̏o
void TableWriter::WriteStartCurve(bool isSelected,bool isClosed)
{
#ifndef NDEBUG
	PathToBuffer::WriteStartCurve(isSelected,isClosed);
#endif
	PATHTABLE table;

	ClearTable(&table);
	table.type=PATHTABLE_CURVE;
	table.curve.isSelected=isSelected;
	table.curve.isClosed=isClosed;

	WriteTable(table);
}

//TuO[v̎n܂̏o
void TableWriter::WriteStartSubGroup(bool isSelected)
{
#ifndef NDEBUG
	PathToBuffer::WriteStartSubGroup(isSelected);
#endif
	PATHTABLE table;
	ClearTable(&table);

	table.type=PATHTABLE_SUBGROUP_OPEN;
	table.subGroup.isSelected=isSelected;

	WriteTable(table);
}

//TuO[v̏Ȉo
void TableWriter::WriteEndSubGroup()
{
#ifndef NDEBUG
	PathToBuffer::WriteEndSubGroup();
#endif
	PATHTABLE table;

	ClearTable(&table);
	table.type=PATHTABLE_SUBGROUP_CLOSE;
	table.subGroup.isSelected=false;
	WriteTable(table);
}

//pڂ̏o
void TableWriter::WriteKnot(const FPoint & back,const FPoint & anchor,const FPoint & forward,int mode)
{
#ifndef NDEBUG
	PathToBuffer::WriteKnot(back,anchor,forward,mode);
#endif
	PATHTABLE table;

	ClearTable(&table);
	table.type=PATHTABLE_KNOT;
	table.knot.back=back;
	table.knot.anchor=anchor;
	table.knot.forward=forward;
	table.knot.mode=mode;
	WriteTable(table);
}

//e[u
void TableWriter::ClearTable(PATHTABLE *table) const
{
	memset(table,0,sizeof(PATHTABLE));
}

//e[uo
void TableWriter::WriteTable(const PATHTABLE &table)
{
	WriteData((const char*)&table,sizeof(PATHTABLE));
}

//////////////////ǂݎApX쐬///////////////////

class TableReader : public BufferToPath
{
public:
	TableReader();
	virtual ~TableReader() throw();

	virtual void ReadStartPath(Rect *boundRect);
	virtual void ReadEndPath();
	virtual void ReadStartSubGroup();
	virtual void ReadEndSubGroup(bool *isSelected);
	virtual void ReadStartCurve();
	virtual void ReadEndCurve(bool *isSelected,bool *isCurved,bool *isOverlappedEndCurve);
	virtual void ReadKnot(FPoint *back,FPoint *anchor,FPoint *forward,int *mode);
	virtual bool IsSubGroupOpened();
	virtual bool IsCurve();
	virtual bool IsKnot();
protected:
	void ReadTable(PATHTABLE *table);
	bool CheckType(int type) const;
private:
	bool m_isCurveSelected;
	bool m_isCurveClosed;
	bool m_isSubGroupSelected;
};

TableReader::TableReader()
{
}

TableReader::~TableReader() throw()
{

}

//pX̎n܂
void TableReader::ReadStartPath(Rect *boundRect)
{
#ifndef NDEBUG
	BufferToPath::ReadStartPath(boundRect);
#endif
	assert(boundRect!=NULL);

	PATHTABLE table;
	ReadTable(&table);
	if(table.type!=PATHTABLE_HEADER) {
		throw Exception(BZPERROR_ILLEGAL_DATA_FORMAT);
	}
	boundRect->Set(table.header.boundRect);
}

//pX̏I
void TableReader::ReadEndPath()
{
#ifndef NDEBUG
	BufferToPath::ReadEndPath();
#endif
	PATHTABLE table;
	ReadTable(&table);
	if(table.type!=PATHTABLE_END) {
		throw Exception(BZPERROR_ILLEGAL_DATA_FORMAT);
	}
}

//TuO[vn܂邩
bool TableReader::IsSubGroupOpened()
{
#ifndef NDEBUG
	BufferToPath::IsSubGroupOpened();
#endif
	return CheckType(PATHTABLE_SUBGROUP_OPEN);
}

//TuO[v̎n܂
void TableReader::ReadStartSubGroup()
{
#ifndef NDEBUG
	BufferToPath::ReadStartSubGroup();
#endif
	PATHTABLE table;
	ReadTable(&table);
	assert(table.type==PATHTABLE_SUBGROUP_OPEN);

	m_isSubGroupSelected=table.subGroup.isSelected;
}

//TuO[v̏I
void TableReader::ReadEndSubGroup(bool *isSelected)
{
#ifndef NDEBUG
	BufferToPath::ReadEndSubGroup(isSelected);
#endif
	*isSelected=m_isSubGroupSelected;

	PATHTABLE table;
	ReadTable(&table);
	assert(table.type==PATHTABLE_SUBGROUP_CLOSE);
}

//n܂邩
bool TableReader::IsCurve()
{
#ifndef NDEBUG
	BufferToPath::IsCurve();
#endif
	return CheckType(PATHTABLE_CURVE);
}

//̎n܂
void TableReader::ReadStartCurve()
{
#ifndef NDEBUG
	BufferToPath::ReadStartCurve();
#endif
	PATHTABLE table;
	ReadTable(&table);
	assert(table.type==PATHTABLE_CURVE);

	m_isCurveSelected=table.curve.isSelected;
	m_isCurveClosed=table.curve.isClosed;
}

//̏I
void TableReader::ReadEndCurve(bool *isSelected,bool *isClosed,bool *isOverlappedCloser)
{
#ifndef NDEBUG
	BufferToPath::ReadEndCurve(isSelected,isClosed,isOverlappedCloser);
#endif
	*isSelected=m_isCurveSelected;
	*isClosed=m_isCurveClosed;
	*isOverlappedCloser=false;
}

bool TableReader::IsKnot()
{
#ifndef NDEBUG
	BufferToPath::IsKnot();
#endif
	return CheckType(PATHTABLE_KNOT);
}

//p
void TableReader::ReadKnot(FPoint *back,FPoint *anchor,FPoint *forward,int *mode)
{
#ifndef NDEBUG
	BufferToPath::ReadKnot(back,anchor,forward,mode);
#endif
	PATHTABLE table;
	ReadTable(&table);
	assert(table.type==PATHTABLE_KNOT);

	*back=table.knot.back;
	*anchor=table.knot.anchor;
	*forward=table.knot.forward;
	*mode=table.knot.mode;
}

//e[uǂ
void TableReader::ReadTable(PATHTABLE *table)
{
	ReadData((char*)table,sizeof(PATHTABLE));
}

//ނr
bool TableReader::CheckType(int type) const
{
	int value;
	CheckData(reinterpret_cast<char*>(&value)
		,sizeof(int));
	return type==value;
}

///////////////////Ăяo֐/////////////////

/*ϊ̃e[ũoCgvZ*/
int PathToTableSize(const Path *path,int part)
{
	BZP_PARAMETER_ERROR(path!=NULL,BZPERROR_NULL_ARGUMENT);

	TableWriter writer;
	path->PathToData(&writer,part);
	return writer.Size();
}

/*ϊ̃e[u̐vZ*/
int PathToTableCount(const Path *path,int part)
{
	int size=PathToTableSize(path,part);
	assert(size%sizeof(PATHTABLE)==0);
	return size/sizeof(PATHTABLE);
}

/*pXe[u`ɕϊ*/
int PathToTable(const Path *path,int part,PATHTABLE *table,int maxSize)
{
	BZP_PARAMETER_ERROR(path!=NULL && table!=NULL
		,BZPERROR_NULL_ARGUMENT);

	TableWriter writer;
	writer.SetBuffer(reinterpret_cast<char*>(table),maxSize);
	path->PathToData(&writer,part);
	return writer.Size();
}

/*e[u`pXɂ*/
void TableToPath(Path *path,const PATHTABLE *table,int size)
{
	BZP_PARAMETER_ERROR(path!=NULL && table!=NULL
		,BZPERROR_NULL_ARGUMENT);

	TableReader reader;
	reader.SetBuffer(reinterpret_cast<const char *>(table)
		,size);
	path->DataToPath(&reader);
}

BZP_NAMESPACE_END
