#include "prefix.h"

#include "convertertag.h"
#include "knotmode.h"
#include "error.h"
#include "exception.h"
#include "fpoint.h"
#include "converterbuffer.h"
#include "path.h"

/*************************************************

	Op̃^O`ɕϊ

**************************************************/

#define LONGSIZE 4	/*long^̑傫*/

/*f[^̐擪̎ʕ
萔̓RpCɂďԂςĂ܂̂
rbg삵Ă*/
#define SIGNATURE (((((('B'<<8)+'Z')<<8)+'P')<<8)+'D')

BZP_NAMESPACE_START

enum
{
	CURRENT_VERSION,	//݂̃o[W

	AVAILABLE_OLDEST_VERSION=0,	//̃R[hōꂽf[^ǂ߂ŌẪo[W
};

enum
{
	TAG_END_PATH,	//pX̏I
	TAG_OPEN_SUBGROUP,	//TuO[vn߂
	TAG_CLOSE_SUBGROUP,	//TuO[vI
	TAG_CURVE,	//n߂
	TAG_KNOT,	//p
};


////////////o///////////////

class TagWriter : public PathToBuffer
{
public:
	TagWriter();
	virtual ~TagWriter() throw();

	virtual void WriteStartPath(const Rect &boundRect);
	virtual void WriteEndPath();
	virtual void WriteStartCurve(bool isSelected,bool isClosed);
	virtual void WriteStartSubGroup(bool isSelected);
	virtual void WriteEndSubGroup();
	virtual void WriteKnot(const FPoint & back,const FPoint & anchor,const FPoint & forward,int mode);

protected:
	LONG WriteTag(SHORT tag);
	void WriteTagSize(LONG pointer);
private:
};

///////////////////////////////////////////////////////

TagWriter::TagWriter()
{
	
}

TagWriter::~TagWriter() throw()
{
}

//pX̎n܂̏o
void TagWriter::WriteStartPath(const Rect &boundRect)
{
#ifndef NDEBUG
	PathToBuffer::WriteStartPath(boundRect);
#endif

	SetBoundRect(boundRect);

	WriteLong(SIGNATURE);
	WriteLong(CURRENT_VERSION);
	WriteLong(AVAILABLE_OLDEST_VERSION);
	WriteRect(boundRect);
}

//pX̏Ȉo
void TagWriter::WriteEndPath()
{
#ifndef NDEBUG
	PathToBuffer::WriteEndPath();
#endif
	long pointer=WriteTag(TAG_END_PATH);
	WriteTagSize(pointer);
}

//̎n܂̏o
void TagWriter::WriteStartCurve(bool isSelected,bool isClosed)
{
#ifndef NDEBUG
	PathToBuffer::WriteStartCurve(isSelected,isClosed);
#endif
	long pointer=WriteTag(TAG_CURVE);
	WriteBool(isSelected);
	WriteBool(isClosed);
	WriteTagSize(pointer);
}

//TuO[v̎n܂̏o
void TagWriter::WriteStartSubGroup(bool isSelected)
{
#ifndef NDEBUG
	PathToBuffer::WriteStartSubGroup(isSelected);
#endif
	long pointer=WriteTag(TAG_OPEN_SUBGROUP);
	WriteBool(isSelected);
	WriteTagSize(pointer);
}

//TuO[v̏Ȉo
void TagWriter::WriteEndSubGroup()
{
#ifndef NDEBUG
	PathToBuffer::WriteEndSubGroup();
#endif
	long pointer=WriteTag(TAG_CLOSE_SUBGROUP);
	WriteTagSize(pointer);
}

//pڂ̏o
void TagWriter::WriteKnot(const FPoint & back,const FPoint & anchor,const FPoint & forward,int mode)
{
#ifndef NDEBUG
	PathToBuffer::WriteKnot(back,anchor,forward,mode);
#endif
	long pointer=WriteTag(TAG_KNOT);
	WriteBool((mode&KNOT_SELECTED)!=0);
	WriteBool((mode&KNOT_LINKED)!=0);
	WriteBool((mode&KNOT_CURVED_FORWARD)!=0);
	WriteBool((mode&KNOT_CURVED_BACK)!=0);
	WriteFixedPointPoint(back);
	WriteFixedPointPoint(anchor);
	WriteFixedPointPoint(forward);
	WriteTagSize(pointer);
}

//^Oo
//^Ȏ傫ꏊԂ
LONG TagWriter::WriteTag(SHORT tag)
{
	WriteShort(tag);
	LONG pointer=GetPointer();
	WriteLong(0);
	return pointer;
}

//pointerʒuɃ^Ȏ傫o
void TagWriter::WriteTagSize(LONG pointer)
{
	WriteLong(GetPointer()-pointer-LONGSIZE,pointer);
}

//////////////////ǂݎApX쐬///////////////////


class TagReader : public BufferToPath
{
public:
	TagReader();
	virtual ~TagReader() throw();

	virtual void ReadStartPath(Rect *boundRect);
	virtual void ReadEndPath();
	virtual void ReadStartSubGroup();
	virtual void ReadEndSubGroup(bool *isSelected);
	virtual void ReadStartCurve();
	virtual void ReadEndCurve(bool *isSelected,bool *isCurved,bool *isOverlappedCloser);
	virtual void ReadKnot(FPoint *back,FPoint *anchor,FPoint *forward,int *mode);
	virtual bool IsSubGroupOpened();
	virtual bool IsCurve();
	virtual bool IsKnot();
protected:
	void ReadTag(SHORT *tag,int *size,LONG *pointer);
	bool CheckTag(SHORT tag) const;
	void SkipTag(int size,LONG pointer);

private:
	LONG m_version;	/*o[Wԍ*/
	bool m_isCurveSelected;	/*̐IĂ邩*/
	bool m_isCurveClosed;	/*̐Ă邩*/
	bool m_isSubGroupSelected;	/*̃TuO[vIĂ邩*/
};

TagReader::TagReader()
{
}

TagReader::~TagReader() throw()
{
}

//pX̎n܂
void TagReader::ReadStartPath(Rect *boundRect)
{
#ifndef NDEBUG
	BufferToPath::ReadStartPath(boundRect);
#endif
	assert(boundRect!=NULL);

	long signature=ReadLong();
	if(signature!=SIGNATURE) {
		throw Exception(BZPERROR_ILLEGAL_DATA_FORMAT);
	}

	m_version=ReadLong();

	long availableOldestVersion=ReadLong();
	if(availableOldestVersion>CURRENT_VERSION) {
		throw Exception(BZPERROR_CANT_READ_NEW_FORMAT);
	}

	ReadRect(boundRect);
	SetBoundRect(*boundRect);	
}

//pX̏I
void TagReader::ReadEndPath()
{
#ifndef NDEBUG
	BufferToPath::ReadEndPath();
#endif
	SHORT tag;
	int size;
	LONG pointer;
	ReadTag(&tag,&size,&pointer);
	if(tag!=TAG_END_PATH) {
		throw Exception(BZPERROR_ILLEGAL_DATA_FORMAT);
	}
	SkipTag(size,pointer);
}

//TuO[vn܂邩
bool TagReader::IsSubGroupOpened()
{
#ifndef NDEBUG
	BufferToPath::IsSubGroupOpened();
#endif
	return CheckTag(TAG_OPEN_SUBGROUP);
}

//TuO[v̎n܂
void TagReader::ReadStartSubGroup()
{
#ifndef NDEBUG
	BufferToPath::ReadStartSubGroup();
#endif
	SHORT tag;
	int size;
	LONG pointer;
	ReadTag(&tag,&size,&pointer);
	assert(tag==TAG_OPEN_SUBGROUP);

	m_isSubGroupSelected=ReadBool();
	SkipTag(size,pointer);
}

//TuO[v̏I
void TagReader::ReadEndSubGroup(bool *isSelected)
{
#ifndef NDEBUG
	BufferToPath::ReadEndSubGroup(isSelected);
#endif
	*isSelected=m_isSubGroupSelected;

	SHORT tag;
	int size;
	LONG pointer;
	ReadTag(&tag,&size,&pointer);
	assert(tag==TAG_CLOSE_SUBGROUP);
	SkipTag(size,pointer);
}

//n܂邩
bool TagReader::IsCurve()
{
#ifndef NDEBUG
	BufferToPath::IsCurve();
#endif
	return CheckTag(TAG_CURVE);
}

//̎n܂
void TagReader::ReadStartCurve()
{
#ifndef NDEBUG
	BufferToPath::ReadStartCurve();
#endif
	SHORT tag;
	int size;
	LONG pointer;
	ReadTag(&tag,&size,&pointer);
	if(tag!=TAG_CURVE) {
		throw Exception(BZPERROR_ILLEGAL_DATA_FORMAT);
	}

	m_isCurveSelected=ReadBool();
	m_isCurveClosed=ReadBool();
	SkipTag(size,pointer);
}

//̏I
void TagReader::ReadEndCurve(bool *isSelected,bool *isClosed,bool *isOverlappedCloser)
{
#ifndef NDEBUG
	BufferToPath::ReadEndCurve(isSelected,isClosed,isOverlappedCloser);
#endif
	*isSelected=m_isCurveSelected;
	*isClosed=m_isCurveClosed;
	*isOverlappedCloser=false;
}

bool TagReader::IsKnot()
{
#ifndef NDEBUG
	BufferToPath::IsKnot();
#endif
	return CheckTag(TAG_KNOT);
}

//p
void TagReader::ReadKnot(FPoint *back,FPoint *anchor,FPoint *forward,int *mode)
{
#ifndef NDEBUG
	BufferToPath::ReadKnot(back,anchor,forward,mode);
#endif
	SHORT tag;
	int size;
	LONG pointer;
	ReadTag(&tag,&size,&pointer);

	int l_mode=0;
	bool isSelected=ReadBool();
	if(isSelected) l_mode|=KNOT_SELECTED;

	bool isLinked=ReadBool();
	if(isLinked) l_mode|=KNOT_LINKED;

	bool isForwardCurved=ReadBool();
	if(isForwardCurved) l_mode|=KNOT_CURVED_FORWARD;

	bool isBackCurved=ReadBool();
	if(isBackCurved) l_mode|=KNOT_CURVED_BACK;

	*mode=l_mode;

	*back=ReadFixedPointPoint();
	*anchor=ReadFixedPointPoint();
	*forward=ReadFixedPointPoint();

	SkipTag(size,pointer);
}

//^Oǂ
void TagReader::ReadTag(SHORT *tag,int *size,LONG *pointer)
{
	assert(tag!=NULL && size!=NULL && pointer!=NULL);
	*tag=ReadShort();
	*size=ReadLong();
	*pointer=GetPointer();
}

//^Or
bool TagReader::CheckTag(SHORT tag) const
{
	return tag==CheckShort();
}

//^ȌI܂Ń|C^i߂
void TagReader::SkipTag(int size,LONG pointer)
{
	if(GetPointer()>pointer+size) {
		throw Exception(BZPERROR_ILLEGAL_DATA_FORMAT);
	}
	IncrementPointer(pointer+size-GetPointer());
}

///////////////////Ăяo֐/////////////////

/*ϊ̃f[^\̃oCgvZ*/
int PathToTagSize(const Path *path,int part)
{
	BZP_PARAMETER_ERROR(path!=NULL,BZPERROR_NULL_ARGUMENT);

	TagWriter writer;
	path->PathToData(&writer,part);
	return writer.Size();
}

/*pX^O`ɕϊ*/
int PathToTag(const Path *path,int part,char *buffer,int maxSize)
{
	BZP_PARAMETER_ERROR(path!=NULL && buffer!=NULL
		,BZPERROR_NULL_ARGUMENT);

	TagWriter writer;
	writer.SetBuffer(buffer,maxSize);
	path->PathToData(&writer,part);

	return writer.Size();
}

/*^O`pXɂ*/
void TagToPath(Path *path,const char *buffer,int size)
{
	BZP_PARAMETER_ERROR(path!=NULL && buffer!=NULL
		,BZPERROR_NULL_ARGUMENT);

	TagReader reader;
	reader.SetBuffer(buffer,size);
	path->DataToPath(&reader);
}

BZP_NAMESPACE_END
