#include "prefix.h"

#include <memory>
#include <new>

#include "curve.h"
#include "knot.h"
#include "part.h"
#include "bezier.h"
#include "runtimeconfig.h"
#include "error.h"
#include "selectmode.h"
#include "ptrdeque.h"
#include "findresult.h"
#include "partmask.h"
#include "index.h"
#include "graph.h"
#include "movecurvemode.h"
#include "deleteknotmode.h"
#include "insertknotmode.h"
#include "converterbase.h"
#include "exception.h"
#include "rect.h"
#include "transform.h"

/******************************************************

	ȐNX̒`

*******************************************************/

BZP_NAMESPACE_START

BZP_CONFIG_USING
BZP_PARTMASK_USING

typedef PtrDeque<Knot> KnotDeque;

class CurveMember
{
public:
	KnotDeque knots;	/*p*/
};

Curve::Curve()
	:m_m(new CurveMember)
	,m_isClosed(false)
	,m_isSelected(true)
{
	BZP_BAD_ALLOC_TEST(m_m.get());
}

Curve::~Curve() throw()
{
}

//pڂ̐
int Curve::KnotCount() const
{
	return m_m->knots.Count();
}

//index̎wpڂ𓾂
Knot *Curve::GetKnot(int index)
{
	return m_m->knots.At(index);
}

//indexԖڂ̌pڂ̃AhX
const Knot *Curve::GetKnot(int index) const
{
	return m_m->knots.At(index);
}

//indeẍǑpڂ𓾂
const Knot *Curve::GetPreviousKnot(int index) const
{
	index=m_m->knots.NormalizeIndex(index);
	if(index==INDEX_FIRST) {
		if(IsClosed()) {
			return GetKnot(INDEX_LAST);
		} else {
			return NULL;
		}
	}
	return GetKnot(index-1);
}

//index̎̌pڂ𓾂
const Knot *Curve::GetNextKnot(int index) const
{
	int nextIndex=GetNextKnotIndex(index);
	if(nextIndex==INDEX_NONE) {
		return NULL;
	}
	return GetKnot(nextIndex);
}

//index̎̌pڂ𓾂
Knot *Curve::GetNextKnot(int index)
{
	int nextIndex=GetNextKnotIndex(index);
	if(nextIndex==INDEX_NONE) {
		return NULL;
	}
	return GetKnot(nextIndex);
}

//index̎̌pڂ̔ԍ
//ĂindexŌ̌pڂȂŏ̌pڂ̔ԍ
int Curve::GetNextKnotIndex(int index) const
{
	int nextIndex;
	nextIndex=m_m->knots.NormalizeIndex(index)+1;
	if(nextIndex==KnotCount()) {
		if(IsClosed()) {
			nextIndex=INDEX_FIRST;
		} else {
			nextIndex=INDEX_NONE;
		}
	}
	return nextIndex;
}

//IĂ邩
bool Curve::IsSelected() const
{
	return m_isSelected;
}

//index̑ǑpڂIĂ邩
bool Curve::IsPreviousKnotSelected(int index) const
{
	const Knot *prevKnot=GetPreviousKnot(index);
	bool isPrevSelected=false;
	if(prevKnot!=NULL) {
		isPrevSelected=prevKnot->IsSelected();
	}
	return isPrevSelected;
}

//index̎̌pڂIĂ邩
bool Curve::IsNextKnotSelected(int index) const
{
	const Knot *nextKnot=GetNextKnot(index);
	bool isNextSelected=false;
	if(nextKnot!=NULL) {
		isNextSelected=nextKnot->IsSelected();
	}
	return isNextSelected;
}

//ėւɂȂĂ邩
bool Curve::IsClosed() const
{
	return m_isClosed;
}

//I
//isSelectedfalseȂIO
void Curve::Select(bool isSelected)
{
	m_isSelected=isSelected;
	if(!isSelected && GetConfigBool(CONFIG_UNSELECT_CURVE_THEN_UNSELECT_KNOT)) {
		SelectAllKnots(false);
	}
}

//ėւɂ
//isClosedfalseȂJ
void Curve::Close(bool isClosed)
{
	assert(!(isClosed && KnotCount()==1));
	m_isClosed=isClosed;
}

//pڂȐɂȂĂ邩
//partPART_ANCHORȂÔǂ炩ȐȂtrue
//PART_FORWARDCONTROLȂOȐȂtrue
//PART_BACKCONTROLȂȐȂtrue
bool Curve::IsCurved(int part,int knotIndex) const
{
	return GetKnot(knotIndex)->IsCurved(part);
}

//̖[mڑ
int Curve::Chain(int knotIndexA,int knotIndexB,bool chainForward,Curve *curveB,bool deleteKnotB)
{
	//vύX	ONcurveBڑɓs̗ǂ悤ɕύXꂽԂ
	//߂Ă܂̂łʂCurveIuWFNgɕĂ̂ق
	assert(curveB!=NULL);

	//chainForward͑OɂȂ
	int symbolizeB=curveB->SymbolizeKnotIndex(knotIndexB);

	if( (!chainForward	&& symbolizeB==INDEX_FIRST)
		|| (chainForward && symbolizeB!=INDEX_FIRST)) {

		curveB->Reverse();
		if(symbolizeB==INDEX_FIRST) {
			knotIndexB=INDEX_LAST;
		} else {
			knotIndexB=INDEX_FIRST;
		}
	}

	if(deleteKnotB) {
		GetKnot(knotIndexA)->Copy(curveB->GetKnot(knotIndexB));
		curveB->DeleteKnot(DELETEKNOT_JUST,knotIndexB);
		if(curveB->KnotCount()==0) {
			return knotIndexA;
		}
	}

	if(this==curveB) {
		if(KnotCount()!=1) {
			Close();
			if(deleteKnotB) {
				return knotIndexA;
			} else {
				return knotIndexB;
			}
		}
		return knotIndexA;
	}

	int insertIndex;
	int newIndexA;
	if(!chainForward) {
		insertIndex=INDEX_PRECEDE;
		if(deleteKnotB) {
			newIndexA=curveB->KnotCount();
		} else {
			newIndexA=curveB->KnotCount()-1;
		}
	} else {
		insertIndex=INDEX_APPEND;
		if(deleteKnotB) {
			newIndexA=KnotCount()-1;
		} else {
			newIndexA=KnotCount();
		}
	}

	m_m->knots.InsertAfter(curveB->GetKnots(),insertIndex);
	curveB->ReleaseAllKnots();

	return newIndexA;
}

//𔽓]
void Curve::Reverse()
{
	int knotCount=KnotCount();
	FOR(int i=0;i<knotCount;++i) {
		GetKnot(i)->Reverse();
	}
	m_m->knots.Reverse();
}

//𑊑ΓIɈړ
void Curve::MoveRel(const FPoint & rel) 
{
	int knotCount=KnotCount();
	FOR(int i=0;i<knotCount;++i) {
		GetKnot(i)->MoveRel(rel);
	}
}

//ړ
void Curve::Move(const FPoint & newPoint,const FPoint & oldPoint,MOVECURVEMODE mode,int knotIndex)
{
	FPoint rel=newPoint-oldPoint;
	switch(mode) {
	case MOVECURVE_JUST:
		MoveRel(rel);
		break;

	case MOVECURVE_SEGMENT:
		GetKnot(knotIndex)->MoveRel(rel);
		GetNextKnot(knotIndex)->MoveRel(rel);
		break;
#if 0	//쐻
	case MOVECURVE_BEND:
		assert(false);
		MoveCurveBend(newPoint,oldPoint,mode,knotIndex);
		break;
#endif
	default:
		assert(false);
	}
}

#if 0	//vύX	쐻
//̐߂Ȃ
void Curve::MoveCurveBend(const FPoint & newPoint,const FPoint & oldPoint,MOVECURVEMODE mode,int knotIndex)
{

	//̕ς邩ςȂ
	Knot *knotA=GetKnot(knotIndex);
	Knot *knotB=GetNextKnot(knotIndex);
	assert(knotB!=NULL);

	FPoint anchorA=knotA->GetPoint(PART_ANCHOR);
	FPoint anchorB=knotA->GetPoint(PART_ANCHOR);

	FPoint d=anchorB-anchorA;

	FPoint move;
	move.x=(newPoint.x-oldPoint.x)/(d.x-d.y);
	move.y=(newPoint.y-oldPoint.y)/(d.y+d.x);


	new.x=old.x+x*dx-y*dy
	new.y=old.y+x*dy+y*dx

	x*dx=(nx-ox)+y*dy
	x*dy=(nx-ox)-y*dx

	(nx-ox)*dy+y*dy2=(nx-ox)*dx-y*dx2
	y(dx2+dy2)=(nx-ox)(dx-dy)
	y=(nx-ox)(dx-dy)/(dx2+dy2);

	double h=oldPoint.y/(d.y+d.x);

	FPoint forwardA=knotA->GetPoint(PART_FORWARDCONTROL);
	double forwardAh=forwardA


	FPoint backB=knotB->GetPoint(PART_BACKCONTROL);


}
#endif

//pڂ̔z𓾂
const KnotDeque &Curve::GetKnots() const
{
	return m_m->knots;
}

//SĂ̌pڂLȂ
void Curve::ReleaseAllKnots()
{
	m_m->knots.ReleaseAll();
}

//pڂǉ
int Curve::InsertKnot(INSERTKNOTMODE mode,double t,int knotIndex)
{
	Knot *curKnot=GetKnot(knotIndex);
	Knot *nextKnot=GetNextKnot(knotIndex);
	assert(curKnot!=NULL && nextKnot!=NULL);

	Bezier bezier(curKnot,nextKnot);
	bezier.Divide(t);

	FPoint tpoint=bezier.GetAnchorB();

	std::auto_ptr<Knot> newKnot(new Knot);
	BZP_BAD_ALLOC_TEST(newKnot.get());

	newKnot->SetPoint(tpoint,PART_ANCHOR);

	if(curKnot->IsCurved(PART_FORWARDCONTROL)
		|| nextKnot->IsCurved(PART_BACKCONTROL)) {
		curKnot->SetPoint(bezier.GetForwardA(),PART_FORWARDCONTROL);
		newKnot->SetPoint(bezier.GetBackB(),PART_BACKCONTROL);

		bezier.Next();
		newKnot->SetPoint(bezier.GetForwardA(),PART_FORWARDCONTROL);
		nextKnot->SetPoint(bezier.GetBackB(),PART_BACKCONTROL);

		newKnot->Curve(true,PART_ANCHOR);
		newKnot->Link();
	} else if(mode==INSERTKNOT_STRAIGHT_CURVE) {
		double magnification=GetConfigDouble(CONFIG_MAKE_CURVE_MAGNIFICATION);
		FPoint forward=tpoint.TPoint(nextKnot->GetPoint(PART_ANCHOR),magnification);
		newKnot->SetPoint(forward,PART_FORWARDCONTROL);
		FPoint back=tpoint.TPoint(curKnot->GetPoint(PART_ANCHOR),magnification);
		newKnot->SetPoint(back,PART_BACKCONTROL);
		newKnot->Curve(true,PART_ANCHOR);
		newKnot->Link();
	} else {
		newKnot->SetPoint(tpoint,PART_FORWARDCONTROL);
		newKnot->SetPoint(tpoint,PART_BACKCONTROL);
	}
	int index=m_m->knots.InsertAfter(newKnot.get(),knotIndex);
	newKnot.release();
	return index;
}

//̖[Ɍpڂ
int Curve::AddEndKnot(const FPoint & point,int knotIndex)
{
	int insertIndex=SymbolizeKnotIndex(knotIndex);
	if(insertIndex==INDEX_FIRST) {
		insertIndex=INDEX_PRECEDE;
	}

	std::auto_ptr<Knot> newKnot(new Knot);
	BZP_BAD_ALLOC_TEST(newKnot.get());

	newKnot->MovePoint(point,PART_ANCHOR);

	int afterKnotIndex=m_m->knots.InsertAfter(newKnot.get(),insertIndex);
	newKnot.release();
	return afterKnotIndex;
}

//pڂ폜
void Curve::DeleteKnot(DELETEKNOTMODE mode,int knotIndex)
{
	m_m->knots.Remove(knotIndex);

	if(KnotCount()==1) {
		Close(false);
	} else if(mode==DELETEKNOT_MIMIC) {
//vύX	
	} else if(mode==DELETEKNOT_ROUNDCORNER) {
//vύX	
	}
}

//pڂI
void Curve::SelectKnot(SELECTMODE mode,int knotIndex)
{
	Select(true);

	Knot *curKnot=NULL;
	if(knotIndex!=INDEX_NONE
		&& mode!=SELECTMODE_ALL
		&& mode!=SELECTMODE_UNSELECTALL) {
		curKnot=GetKnot(knotIndex);
	}

	switch(mode) {
	case SELECTMODE_ADD:
		curKnot->Select(true);
		break;

	case SELECTMODE_SINGLE:
		SelectAllKnots(false);
		curKnot->Select(true);
		break;

	case SELECTMODE_UNSELECT:
		curKnot->Select(false);
		break;

	case SELECTMODE_TOGGLE:
		curKnot->Select(!curKnot->IsSelected());
		break;

	case SELECTMODE_ALL:
		SelectAllKnots(true);
		break;

	case SELECTMODE_UNSELECTALL:
		SelectAllKnots(false);
		break;
	}
}

//SĂ̌pڂIEI
void Curve::SelectAllKnots(bool isSelected)
{
	int knotCount=KnotCount();
	FOR(int i=0;i<knotCount;++i) {
		GetKnot(i)->Select(isSelected);
	}
}

//pڂIĂ邩
bool Curve::IsSelectedKnot(int knotIndex) const
{
	if(!IsSelected()) {
		return false;
	}
	return GetKnot(knotIndex)->IsSelected();
}

//pڂ̕ȂĂ邩
void Curve::LinkKnot(bool isLinked,int knotIndex)
{
	GetKnot(knotIndex)->Link(isLinked);
}

//pڂɂȂĂȐEɂ
void Curve::CurveKnot(bool isCurved,int part,int knotIndex)
{
	GetKnot(knotIndex)->Curve(isCurved,part);
}

//IĂpڂ̐
int Curve::SelectedKnotCount() const
{
	int count=0;
	FOR(int i=0;i<KnotCount();++i) {
		if(GetKnot(i)->IsSelected()) {
			++count;
		}
	}
	return count;
}

//pڂ̔ԍ̈ʒu\lɂ
int Curve::SymbolizeKnotIndex(int knotIndex) const
{
	return m_m->knots.SymbolizeIndex(knotIndex);
}

//IĂpڂ𑊑ΓIɈړ
void Curve::MoveSelectedKnots(const FPoint & rel)
{
	if(IsSelected()) {
		int knotCount=KnotCount();
		FOR(int i=0;i<knotCount;++i) {
			Knot *knot=GetKnot(i);
			if(knot->IsSelected()) {
				knot->MoveRel(rel);
			}
		}
	}
}

//pڂ̒ʉߓ__̍W𓾂
FPoint Curve::GetPoint(int part,int knotIndex) const
{
	return GetKnot(knotIndex)->GetPoint(part);
}

//pڂ̒ʉߓ__̍Wړ
void Curve::MovePoint(const FPoint & point,int part,int knotIndex)
{
	GetKnot(knotIndex)->MovePoint(point,part);
}

//pڂ̏ԂύXɒʉߓ__̍WύX
void Curve::SetPoint(const FPoint & point,int part,int knotIndex)
{
	GetKnot(knotIndex)->SetPoint(point,part);
}

//𗼑֐L΂
void Curve::MoveWControls(const FPoint & point,bool grippedForward,int knotIndex)
{
	GetKnot(knotIndex)->MoveWControls(point,grippedForward);
}

//pڂ̏ԂW𓯂ɂ
//srcKnotIndex̎wpڂ̓edestKnotIndex̎wpڂɎʂ
void Curve::CopyKnot(int destKnotIndex,int srcKnotIndex,const Curve *srcCurve)
{
	const Knot *srcKnot=srcCurve->GetKnot(srcKnotIndex);
	CopyKnot(srcKnot,destKnotIndex);
}

//pڂ̏ԂW𓯂ɂ
//srcKnot̓eknotIndex̎wpڂɎʂ
void Curve::CopyKnot(const Knot *srcKnot,int knotIndex)
{
	GetKnot(knotIndex)->Copy(srcKnot);
}

//pڂ̏ԂW𓯂ɂ
//knotIndex̎wpڂ̂̓edestKnotɎʂ
void Curve::CopyKnotTo(Knot *destKnot,int knotIndex) const
{
	destKnot->Copy(GetKnot(knotIndex));
}

//ƌpڂ`
void Curve::Draw(const Graph *graph) const
{
	if(graph->DoesDraw()
		&& (IsSelected() || graph->DoesDrawUnselected())) {
		DrawSegments(graph);
		if(graph->DoesDrawKnot()) {
			DrawKnots(graph);
		}
	}
}

//`
void Curve::DrawSegments(const Graph *graph) const
{
	int knotCount=KnotCount();
	FOR(int i=0;i<knotCount;++i) {
		const Knot *curKnot=GetKnot(i);
		const Knot *nextKnot=GetNextKnot(i);

		if(nextKnot!=NULL) {
			graph->DrawCurve(curKnot->GetPoint(PART_ANCHOR)
				,curKnot->GetPoint(PART_FORWARDCONTROL)
				,nextKnot->GetPoint(PART_BACKCONTROL)
				,nextKnot->GetPoint(PART_ANCHOR));
		}
	}
}

//ȐIĂΌpڂ`
//IĂ炸pڂǗĂꍇ́~`
void Curve::DrawKnots(const Graph *graph) const
{
	int knotCount=KnotCount();
	if(IsSelected()) {
		FOR(int i=0;i<knotCount;++i) {
			GetKnot(i)->DrawControlHandles(graph
				,IsPreviousKnotSelected(i),IsNextKnotSelected(i));
		}
	} else {
		if(knotCount==1) {
			GetKnot(0)->DrawSingleKnot(graph);
		}
	}
}

//IꂽO[v̐ƌpڂ`
void Curve::DrawSelectedSubGroup(const Graph *graph) const
{
	DrawSegments(graph);

	if(graph->DoesDrawKnot()) {
		int knotCount=KnotCount();
		FOR(int i=0;i<knotCount;++i) {
			GetKnot(i)->DrawAnchor(graph,true);
		}
	}
}

//IĂȂO[v̐ƁAǗpڂ΁~`
void Curve::DrawUnselectedSubGroup(const Graph *graph) const
{
	if(graph->DoesDrawUnselected()) {
		if(KnotCount()>=2) {
			DrawSegments(graph);
		} else if(graph->DoesDrawKnot()) {
			GetKnot(INDEX_FIRST)->DrawSingleKnot(graph);
		}
	}
}

//ʉߓ_`
void Curve::DrawAnchors(const Graph *graph) const
{
	if(IsSelected()) {
		int knotCount=KnotCount();
		FOR(int i=0;i<knotCount;++i) {
			GetKnot(i)->DrawAnchor(graph);
		}
	}
}

//iT
bool Curve::Find(FindResult *findResult,const Graph *graph,int curveIndex,int mask) const
{
	if(FindKnots(findResult,graph,curveIndex,mask)) {
		return true;
	}
	return FindCurve(findResult,graph,mask);
}

//pڂT
bool Curve::FindKnots(FindResult *findResult,const Graph *graph,int curveIndex,int mask) const
{
	int maskCurveSelect=mask;
	if(!IsSelected()) {
		maskCurveSelect=MaskUnselectedCurve(mask);
		if(IsNoPart(maskCurveSelect)) {
			return false;
		}
	}

	int maskKnot=MaskKnot(maskCurveSelect);
	if(IsNoPart(maskKnot)) {
		return false;
	}
	
	//̂̂قǎOɕ\̂ŕ\̋tŒT
	FOR(int i=KnotCount()-1;i>=0;--i) {
		if(!ContainControl(maskKnot)) {
			if(i==KnotCount()-1 && !IsClosed()) {
				if(!ContainLastAnchor(maskKnot)) {
					continue;
				}
			} else if(i==INDEX_FIRST && !IsClosed()) {
				if(!ContainFirstAnchor(maskKnot)) {
					continue;
				}
			} else if(!ContainJointAnchor(maskKnot)) {
				continue;
			}
		}

		if(ContainEndAnchor(maskKnot)
			&& findResult->IsIgnoreKnot(i,curveIndex)) {
			continue;
		}

		int anchorPart=PART_JOINTANCHOR;
		if(!IsClosed()) {
			if(KnotCount()==1) {
				anchorPart=PART_ENDANCHOR;
			} else if(i==0) {
				anchorPart=PART_FIRSTANCHOR;
			} else if(i==KnotCount()-1) {
				anchorPart=PART_LASTANCHOR;
			}
		}

		bool isPrevSelected=IsPreviousKnotSelected(i);
		bool isNextSelected=IsNextKnotSelected(i);

		if(GetKnot(i)->Find(findResult,graph,maskKnot,anchorPart,isPrevSelected,isNextSelected)) {
			findResult->SetKnotIndex(i);
			findResult->SetKnotOnSelectedCurve(IsSelected());
			return true;
		}
	}
	return false;
}

//̑SĂ̐߂T
bool Curve::FindCurve(FindResult *findResult,const Graph *graph,int mask) const
{
	int maskCurve=MaskCurve(mask);
	if(IsNoPart(maskCurve)) {
		return false;
	}

	if(IsSelectedOpposite(maskCurve,IsSelected())) {
		return false;
	}

	if(IsClosedOpposite(maskCurve,IsClosed())) {
		return false;
	}
	
	int knotCount=KnotCount();
	FOR(int i=0;i<knotCount;++i) {
		const Knot *curKnot=GetKnot(i);
		const Knot *nextKnot=GetNextKnot(i);
		if(nextKnot==NULL) break;

		if(FindSegment(findResult,graph,curKnot,nextKnot)) {
			findResult->SetSelect(IsSelected());
			findResult->SetClose(IsClosed());
			findResult->SetKnotIndex(i);
			return true;
		}
	}
	return false;
}

//pڂ̊Ԃ̈߂T
bool Curve::FindSegment(FindResult *findResult,const Graph *graph,const Knot *knotA,const Knot *knotB) const
{
	Bezier bezier(knotA,knotB);

	FPoint curveArea=graph->GetCurveArea();	//{͈
	double minDistance=2.0;	//ԋ߂ꏊ̋
	double min_t=-1.0;	//ԋ߂ꏊ̃xWGȐ̒`ŌƂ̔}ϐ̒l

	FPoint point=findResult->GetPoint();

	FOR(;;) {
		if(bezier.IsPointInArea(point,curveArea)) {
			if(!bezier.IsNarrow(curveArea)) {
				bezier.Divide();
			} else {
				double distance=bezier.Distance(point,curveArea);
				if(distance<minDistance) {
					minDistance=distance;
					min_t=bezier.GetCenter_t();
				}
				if(!bezier.Next()) break;
			}
		} else if(!bezier.Next()) break;
	}

	//ꍇ2.0ȏɂ͂ȂȂ͂Ȃ̂
	//2.0Ȃ猩ĂȂ
	if(minDistance>=2.0) {
		return false;
	}

	findResult->SetPart(PART_CURVE);
	findResult->SetCurve_t(min_t);
	return true;
}

//̋`I
void Curve::SelectBox(SELECTMODE mode,const FRect &rect)
{
	bool isSingleKnotInRect=false;
	if(KnotCount()==1
		&& GetKnot(INDEX_FIRST)->IsAnchorInRect(rect)) {

		isSingleKnotInRect=true;
	}

	if(isSingleKnotInRect || IsInRect(rect)) {
		Select(true);
	}
	
	if(IsSelected()) {
		if(mode==SELECTMODE_SINGLE) {
			SelectAllKnots(false);
		}
		int knotCount=KnotCount();
		FOR(int i=0;i<knotCount;++i) {
			GetKnot(i)->SelectBox(mode,rect);
		}
	}
}

//`͈͓̔ɂ邩
bool Curve::IsInRect(const FRect &rect) const
{
	int knotCount=KnotCount();
	FOR(int i=0;i<knotCount;++i) {
		const Knot *curKnot=GetKnot(i);
		const Knot *nextKnot=GetNextKnot(i);
		if(nextKnot!=NULL) {
			Bezier bezier(curKnot,nextKnot);
			do {
				if(!bezier.IsOutRect(rect)) {
					bezier.Divide();
					if(bezier.IsInRect(rect)) {
						return true;
					}
				}
			} while(bezier.Next());
		}
	}
	return false;
}

//炩̃f[^[`ɕϊ
void Curve::PathToData(PathTo *const converter,int part) const
{
	assert(converter!=NULL);

	if(ContainCurve(part) 
		&& IsSelectedOpposite(part,IsSelected()) ) {
		return;
	}

	int knotCount=KnotCount();
	int startKnot=INDEX_FIRST;
	bool cut=false;
	if(!ContainCurve(part)) {
		FOR(int i=0;i<knotCount;++i) {
			const Knot *knot=GetKnot(i);
			if(IsSelectedOpposite(part,knot->IsSelected() && IsSelected() )) {
				cut=true;
				if(IsClosed()) {
					startKnot=i;
				}
			}
		}
	}

	bool processingCurve=false;
	FOR(int i=0;i<knotCount;++i) {
		int index=i+startKnot;
		if(index>=knotCount) index-=knotCount;

		const Knot *knot=GetKnot(index);
		if(ContainCurve(part)
			||	(!IsSelectedOpposite(part,knot->IsSelected() && IsSelected() )) ) {

			if(!processingCurve) {
				converter->WriteStartCurve(IsSelected(),IsClosed() && !cut);
				processingCurve=true;
			}
			knot->PathToData(converter);
		} else if(processingCurve) {
			converter->WriteEndCurve();
			processingCurve=false;
		}
	}

	if(processingCurve) {
		converter->WriteEndCurve();
	}
}

//炩̃f[^[`pXɕϊ
void Curve::DataToPath(ToPath *const converter)
{
	assert(converter!=NULL);
	converter->ReadStartCurve();

	FOR(;converter->IsKnot();) {
		std::auto_ptr<Knot> newKnot(new Knot);
		BZP_BAD_ALLOC_TEST(newKnot.get());

		newKnot->DataToPath(converter);
		m_m->knots.Add(newKnot.get());
		newKnot.release();
	}
	bool isOverlappedCloser=false;
	converter->ReadEndCurve(&m_isSelected,&m_isClosed,&isOverlappedCloser);
	
	if(isOverlappedCloser) {
		const Knot *lastKnot=GetKnot(INDEX_LAST);
		Knot *firstKnot=GetKnot(INDEX_FIRST);
		if(lastKnot->GetPoint(PART_ANCHOR)
			==firstKnot->GetPoint(PART_ANCHOR)) {

			firstKnot->SetPoint(lastKnot->GetPoint(PART_BACKCONTROL)
				,PART_BACKCONTROL);
			firstKnot->Link(lastKnot->IsLinked());
			firstKnot->Curve(lastKnot->IsCurved(PART_BACKCONTROL)
				,PART_BACKCONTROL);

			DeleteKnot(DELETEKNOT_JUST,INDEX_LAST);		
		}
	}
}

//𕡐
void Curve::Copy(const Curve *srcCurve)
{
	Select(srcCurve->IsSelected());
	Close(srcCurve->IsClosed());

	int knotCount=srcCurve->KnotCount();
	FOR(int i=0;i<knotCount;++i) {
		std::auto_ptr<Knot> newKnot(new Knot);
		BZP_BAD_ALLOC_TEST(newKnot.get());
		newKnot->Copy(srcCurve->GetKnot(i));
		AddKnot(newKnot.get());
		newKnot.release();
	}
}

//pڃIuWFNg
void Curve::AddKnot(Knot *knot)
{
	m_m->knots.Add(knot);
}

//pX͂ދ`
void Curve::GetOuterRect(FRect *rect,int part,double precision) const
{
	if(ContainKnot(part)) {
		GetOuterKnotsRect(rect,part,precision);
	} else if(ContainCurve(part)) {
		GetOuterCurveRect(rect,precision);
	}
}

//pX͂ދ`_܂ޡ
void Curve::GetOuterKnotsRect(FRect *rect,int part,double precision) const
{
	rect->Clear();
	int knotCount=KnotCount();
	FOR(int i=0;i<knotCount;++i) {
		FRect knotRect;
		const Knot *knot=GetKnot(i);
		if(!IsSelectedOpposite(part,knot->IsSelected())) {
			GetKnot(i)->GetOuterRect(&knotRect,precision);
		}
		rect->Union(knotRect);
	}
}

//pX͂ދ`̕
void Curve::GetOuterCurveRect(FRect *rect,double precision) const
{
	rect->Clear();

	int knotCount=KnotCount();
	FOR(int i=0;i<knotCount;++i) {
		const Knot *curKnot=GetKnot(i);
		const Knot *nextKnot=GetNextKnot(i);
		if(nextKnot==NULL) break;

		Bezier bezier(curKnot,nextKnot);
		FRect segmentRect;
		GetOuterSegmentRect(&segmentRect,&bezier,precision);
		
		rect->Union(segmentRect);
	}
}

//`ɓ_̍W܂߂
static void IncludePoint(FRect *rect,const FPoint &point)
{
	if(point.x<rect->left) {
		rect->left=point.x;
	}
	if(rect->right<point.x) {
		rect->right=point.x;
	}
	if(point.y<rect->top) {
		rect->top=point.y;
	}
	if(rect->bottom<point.y) {
		rect->bottom=point.y;
	}
}

//߂͂ދ`
void Curve::GetOuterSegmentRect(FRect *rect,Bezier *bezier,double precision) const
{
	FRect segRect;
	segRect.Set(bezier->GetAnchorA(),bezier->GetAnchorB());
	segRect.Normalize();
	if(segRect.left==segRect.right) {
		segRect.right+=precision;
	}
	if(segRect.top==segRect.bottom) {
		segRect.bottom+=precision;
	}

	FOR(;;) {
		FRect tmp(segRect);
		tmp.Inflate(precision);
		if(bezier->GetForwardA().IsInRect(tmp)
			&& bezier->GetBackB().IsInRect(tmp) ) {
			IncludePoint(&segRect,bezier->GetForwardA());
			IncludePoint(&segRect,bezier->GetBackB());
			if(!bezier->Next()) break;
		} else {
			bezier->Divide();
			IncludePoint(&segRect,bezier->GetAnchorB());
		}
	}

	rect->Set(segRect);
}

//Wϊ
void Curve::Transform(const TRANSFORM &transform)
{
	int knotCount=KnotCount();
	FOR(int i=0;i<knotCount;++i) {
		GetKnot(i)->Transform(transform);
	}
}

//P_Ő؂ĊJ
void Curve::CutClosedCurvePoint(double t,int knotIndex)
{
	assert(IsClosed());

	InsertKnot(INSERTKNOT_STRAIGHT_JUST,t,knotIndex);
	RotateKnots(GetNextKnotIndex(knotIndex));

	SelectAllKnots(false);
	Close(false);

	std::auto_ptr<Knot> newKnot(new Knot);
	BZP_BAD_ALLOC_TEST(newKnot.get());
	CopyKnotTo(newKnot.get(),INDEX_FIRST);
	newKnot->Select(true);
	AddKnot(newKnot.get());
	newKnot.release();
}

//ĂknotIndex̌pڂŏɂȂ悤ɏԂ炷
void Curve::RotateKnots(int knotIndex)
{
	assert(IsClosed());
	m_m->knots.Rotate(knotIndex);
}

//JĂP_Ő؂ĕ
void Curve::CutOpenCurvePoint(Curve *destCurve,double t,int knotIndex)
{
	assert(!IsClosed());

	InsertKnot(INSERTKNOT_STRAIGHT_JUST,t,knotIndex);

	CutOpenCurveSegment(destCurve,knotIndex);

	Select(false);

	std::auto_ptr<Knot> newKnot(new Knot);
	BZP_BAD_ALLOC_TEST(newKnot.get());
	destCurve->CopyKnotTo(newKnot.get(),INDEX_FIRST);
	newKnot->Select(false);
	AddKnot(newKnot.get());
	newKnot.release();
}

//Ă̐߂폜ĊJ
void Curve::CutClosedCurveSegment(int knotIndex)
{
	assert(IsClosed());
	RotateKnots(GetNextKnotIndex(knotIndex));
	Close(false);
}

//JĂ̐߂폜ĕ
void Curve::CutOpenCurveSegment(Curve *destCurve,int knotIndex)
{
	int index=m_m->knots.NormalizeIndex(
		GetNextKnotIndex(knotIndex));
	assert(index!=INDEX_NONE);

	int count=KnotCount()-index;
	FOR(int i=0;i<count;++i) {
		destCurve->AddKnot(m_m->knots.At(index));
		m_m->knots.ReleaseItem(index);
	}
}

BZP_NAMESPACE_END
