#ifndef BZP_CURVE_H
#define BZP_CURVE_H

#include "site.h"

#ifdef __cplusplus
#include <memory>
#endif

/*************************************************

	ȐNX̐錾

**************************************************/

BZP_NAMESPACE_START

#ifdef __cplusplus

class Bezier;
class Knot;
class FPoint;
class FRect;
class Graph;
class FindResult;
class Data;
class PathTo;
class ToPath;
class TRANSFORM;
template<class T> class PtrDeque;
typedef PtrDeque<Knot> KnotDeque;
enum MOVECURVEMODE;
enum SELECTMODE;
enum DELETEKNOTMODE;
enum INSERTKNOTMODE;

class CurveMember;

class Curve  
{
public:
	Curve();
	~Curve() throw();

	bool IsSelected() const;
	bool IsClosed() const;
	void Close(bool isClosed=true);
	void Select(bool isSelected=true);
	bool IsCurved(int part,int knotIndex) const;

	void Draw(const Graph *graph) const;
	void DrawSelectedSubGroup(const Graph *graph) const;
	void DrawUnselectedSubGroup(const Graph *graph) const;
	void DrawAnchors(const Graph *graph) const;

	bool Find(FindResult *findResult,const Graph *graph,int curveIndex,int mask) const;

	int KnotCount() const;

	int Chain(int knotIndexA,int knotIndexB,bool chainForward,Curve *curveB,bool deleteKnotB);
	void Reverse();
	void MoveRel(const FPoint & rel);
	void Move(const FPoint & newPoint,const FPoint & oldPoint,MOVECURVEMODE mode,int knotIndex);

	int InsertKnot(INSERTKNOTMODE mode,double t,int knotIndex);
	int AddEndKnot(const FPoint & point,int knotIndex);
	void AddEndKnot(Knot *knot,int knotIndex);
	void SelectKnot(SELECTMODE mode,int knotIndex);
	bool IsSelectedKnot(int knotIndex) const;
	void MoveSelectedKnots(const FPoint & rel);
	void DeleteKnot(DELETEKNOTMODE mode,int knotIndex);
	void LinkKnot(bool isLinked,int knotIndex);
	void CurveKnot(bool isCurved,int part,int knotIndex);
	int SymbolizeKnotIndex(int knotIndex) const;
	const Knot *GetKnot(int index) const;
	const Knot *GetNextKnot(int index) const;
	int SelectedKnotCount() const;

	FPoint GetPoint(int part,int knotIndex) const;
	void MovePoint(const FPoint & point,int part,int knotIndex);
	void MoveWControls(const FPoint & point,bool grippedForward,int knotIndex);
	void SetPoint(const FPoint & point,int part,int knotIndex);

	void CopyKnot(int destKnotIndex,int srcKnotIndex,const Curve *srcCurve);
	void CopyKnot(const Knot *srcKnot,int knotIndex);
	void CopyKnotTo(Knot *destKnot,int knotIndex) const;
	void AddKnot(Knot *knot);

	void SelectBox(SELECTMODE mode,const FRect &rect);
	bool IsInRect(const FRect &rect) const;

	void PathToData(PathTo *const converter,int part) const;
	void DataToPath(ToPath *const converter);
	void Copy(const Curve *srcCurve);

	void GetOuterRect(FRect *rect,int part,double precision) const;
	void Transform(const TRANSFORM &transform);
	void CutClosedCurvePoint(double t,int knotIndex);
	void CutOpenCurvePoint(Curve *destCurve,double t,int knotIndex);
	void CutClosedCurveSegment(int knotIndex);
	void CutOpenCurveSegment(Curve *destCurve,int knotIndex);

protected:
	void RotateKnots(int knotIndex);
	void GetOuterKnotsRect(FRect *rect,int part,double precision) const;
	void GetOuterCurveRect(FRect *rect,double precision) const;
	void GetOuterSegmentRect(FRect *rect,Bezier *bezier,double precision) const; 
	void SelectAllKnots(bool isSelected);
	bool IsPreviousKnotSelected(int index) const;
	bool IsNextKnotSelected(int index) const;
	Knot *GetKnot(int index);
	const Knot *GetPreviousKnot(int index) const;
	Knot *GetNextKnot(int index);
	int GetNextKnotIndex(int index) const;

	const KnotDeque &GetKnots() const;
	void ReleaseAllKnots();

	void DrawSegments(const Graph *graph) const;
	void DrawKnots(const Graph *graph) const;

	bool FindKnots(FindResult *findResult,const Graph *graph,int curveIndex,int mask) const;
	bool FindCurve(FindResult *findResult,const Graph *graph,int mask) const;
	bool FindSegment(FindResult *findResult,const Graph *graph,const Knot *knotA,const Knot *knotB) const;

	void MoveCurveBend(const FPoint & oldPoint,const FPoint & newPoint,MOVECURVEMODE mode,int knotIndex);
private:
	std::auto_ptr<CurveMember> m_m;

	bool m_isSelected;	/*IĂ邩*/
	bool m_isClosed;	/*Ă邩*/
};

#endif /*__cplusplus*/

BZP_NAMESPACE_END

#endif
