#include "prefix.h"

#include <string.h>
#include <new>

#include "diff.h"
#include "path.h"
#include "convertertable.h"
#include "dimptr.h"
#include "part.h"
#include "exception.h"
#include "error.h"


/*************************************************

	pX̍

**************************************************/

BZP_NAMESPACE_START

struct tagDiffHeader
{
	int frontOffset;	//Ⴂ܂ł̑ÕoCg
	int tailOffset;	//Ⴂ܂ł̌̃oCg
	int prevSize;	//ȌԂ̑傫
	int curSize;	//݂̏Ԃ̑傫
};
typedef tagDiffHeader DiffHeader;

//̍ő̑傫
int DifferenceSize(const Path *previousPath,const Path *currentPath)
{
	BZP_PARAMETER_ERROR(previousPath!=NULL && currentPath!=NULL
		,BZPERROR_NULL_ARGUMENT);

	int size;
	size=PathToTableSize(previousPath,PART_ALL);
	size+=PathToTableSize(currentPath,PART_ALL);
	return size+sizeof(DiffHeader)*2;
}

//
//߂l͍̃oCg Ȃ0
int CreateDifference(const Path *previousPath,const Path *currentPath,char *buffer,int maxSize)
{
	BZP_PARAMETER_ERROR(previousPath!=NULL && currentPath!=NULL && buffer!=NULL
		,BZPERROR_NULL_ARGUMENT);

	int prevSize=PathToTable(previousPath,PART_ALL
		,reinterpret_cast<PATHTABLE*>(buffer+sizeof(DiffHeader)),maxSize-sizeof(DiffHeader));
	int curSize=PathToTable(currentPath,PART_ALL
		,reinterpret_cast<PATHTABLE*>(buffer+sizeof(DiffHeader)+prevSize),maxSize-sizeof(DiffHeader)-prevSize);

	const char *prevFront=buffer+sizeof(DiffHeader);
	const char *curFront=prevFront+prevSize;

	int frontOffset;
	if(curSize<=prevSize) {
		frontOffset=curSize;
	} else {
		frontOffset=prevSize;
	}

	FOR(int i=0;i<frontOffset;++i) {
		char prev=*prevFront;
		++prevFront;
		char cur=*curFront;
		++curFront;
		if(prev!=cur) {
			frontOffset=i;
			break;
		}
	}

	if(frontOffset==curSize && curSize==prevSize) {
		return 0;
	}

	int tailOffset;
	if(curSize<=prevSize) {
		tailOffset=curSize-frontOffset;
	} else {
		tailOffset=prevSize-frontOffset;
	}

	const char *prevTail=buffer+sizeof(DiffHeader)+prevSize;
	const char *curTail=prevTail+curSize;
	FOR(int i=0;i<tailOffset;++i) {
		--prevTail;
		char prev=*prevTail;
		--curTail;
		char cur=*curTail;
		if(prev!=cur) {
			tailOffset=i;
			break;
		}
	}

	int prevDiffSize=prevSize-frontOffset-tailOffset;
	memcpy(buffer+sizeof(DiffHeader),buffer+sizeof(DiffHeader)+frontOffset,prevDiffSize);
	int curDiffSize=curSize-frontOffset-tailOffset;
	memcpy(buffer+sizeof(DiffHeader)+prevDiffSize,buffer+sizeof(DiffHeader)+prevSize+frontOffset,curDiffSize);

	DiffHeader *diffHeader=reinterpret_cast<DiffHeader*>(buffer);
	diffHeader->frontOffset=frontOffset;
	diffHeader->tailOffset=tailOffset;
	diffHeader->prevSize=prevSize;
	diffHeader->curSize=curSize;

	return sizeof(DiffHeader)+prevDiffSize+curDiffSize;
}

//pXɖ߂E蒼
void UndoRedoPath(bool undo,Path *destPath,const Path *currentPath,const char *diff,int size)
{
	BZP_PARAMETER_ERROR(destPath!=NULL
		&& currentPath!=NULL && diff!=NULL
			,BZPERROR_NULL_ARGUMENT);

	const DiffHeader *diffHeader
		=reinterpret_cast<const DiffHeader*>(diff);

	int frontOffset=diffHeader->frontOffset;
	int tailOffset=diffHeader->tailOffset;
	assert(sizeof(DiffHeader)+diffHeader->prevSize
		+diffHeader->curSize-(frontOffset+tailOffset)*2
			<=size);

	int beforeSize,afterSize;
	if(undo) {
		beforeSize=diffHeader->curSize;
		afterSize=diffHeader->prevSize;
	} else {
		beforeSize=diffHeader->prevSize;
		afterSize=diffHeader->curSize;
	}


	int curSize;
	curSize=PathToTableSize(currentPath,PART_ALL);

	int bufferSize=afterSize;
	if(curSize>afterSize) {
		bufferSize=curSize;
	}

	DimPtr<char> buffer(new char[bufferSize]);
	BZP_BAD_ALLOC_TEST(buffer.get());

	assert(beforeSize==curSize || frontOffset+tailOffset<=curSize);

	PathToTable(currentPath,PART_ALL
		,reinterpret_cast<PATHTABLE*>(buffer.get()),bufferSize);

	//̋ʕړ
	memcpy(buffer.get()+afterSize-tailOffset
		,buffer.get()+curSize-tailOffset
		,tailOffset);

	//Ԃ̍ʂ
	const char *diffUndoRedo=diff+sizeof(DiffHeader);
	if(!undo) {
		diffUndoRedo+=beforeSize-(frontOffset+tailOffset);
	}
	memcpy(buffer.get()+frontOffset
		,diffUndoRedo
		,afterSize-(frontOffset+tailOffset));

	TableToPath(destPath
		,reinterpret_cast<PATHTABLE*>(buffer.get()),afterSize);
}

BZP_NAMESPACE_END
