#include "prefix.h"

#include <deque>
#include <algorithm>
#include <new>

#include "fill.h"
#include "curvepointers.h"
#include "curve.h"
#include "knot.h"
#include "bezier.h"
#include "path.h"
#include "fpoint.h"
#include "index.h"
#include "part.h"
#include "exception.h"
#include "error.h"

/*************************************************

	̈hԂ

**************************************************/

BZP_NAMESPACE_START

class PointQue : public std::deque<FPoint>
{

};

class FillMember
{
public:
	CurvePointers curvePtrs;	//߂Ēu
	PointQue pointQue;	//W߂Ēu
};

Fill::Fill()
	:m_m(new FillMember)
	,m_antiAliasing(4)
	,m_option(0)
#if !defined(NDEBUG) || defined(BZP_THROW_BAD_PARAMETER)
	,m_startScan(false)
#endif
{
	BZP_BAD_ALLOC_TEST(m_m.get());
}

Fill::~Fill() throw()
{

}

/*pX𒊏o
path@pX
part@`TuO[v̐PART_CLOSEDł蕪邱Ƃ͂łȂ
*/
void Fill::AddPath(const Path *path,int part)
{
	BZP_PARAMETER_ERROR(!m_startScan,BZPERROR_CALLED_MISTAKE_ORDER);
	BZP_PARAMETER_ERROR(path!=NULL,BZPERROR_NULL_ARGUMENT);
	path->ExtractCurve(&(m_m->curvePtrs),part);
}

//E炩ɂ邽߂ɉ{ŃTvO邩
void Fill::SetAntiAliasing(int antiAliasing)
{
	BZP_PARAMETER_ERROR(!m_startScan,BZPERROR_CALLED_MISTAKE_ORDER);
	BZP_PARAMETER_ERROR(antiAliasing>=1,BZPERROR_ZERO_ANTIALIASING);
	m_antiAliasing=antiAliasing;
}

//IvV̐ݒ
void Fill::SetOption(int option)
{
	BZP_PARAMETER_ERROR(!m_startScan,BZPERROR_CALLED_MISTAKE_ORDER);
	m_option=option;
}

/*sh顏̍s珇ԂɌĂяow肳ꂽs̓_͎̂Ă
left buffer̐擪̉W
y@cW
buffer obt@̃AhX	1hbg1oCg
size obt@̑傫	*/
void Fill::ScanLine(int left,int y,double *buffer,int size)
{
	BZP_PARAMETER_ERROR(buffer!=NULL,BZPERROR_NULL_ARGUMENT);

#if !defined(NDEBUG) || defined(BZP_THROW_BAD_PARAMETER)
	m_startScan=true;
#endif

	ClearBuffer(buffer,size);

	ResolveCurves(y);

	FOR(int i=0;i<m_antiAliasing;++i) {
		FillBuffer(left,y*m_antiAliasing+i,buffer,size);
	}

	FOR(int x=0;x<size;++x) {
		if(buffer[x]>1.0) buffer[x]=1.0;
	}
}

//Ɨp̃obt@
void Fill::ClearBuffer(double *buffer,int size)
{
	FOR(int i=0;i<size;++i) {
		buffer[i]=0;
	}
}

//WBAOɂ邩
static bool IsAPreviouslyThanB(const FPoint &pointA,const FPoint &pointB)
{
	return pointA.y<pointB.y || (pointA.y==pointB.y && pointA.x<pointB.x);
}

//yʂ𕪉
void Fill::ResolveCurves(int y)
{
	int beforePointCount=m_m->pointQue.size();

	int curveCount=m_m->curvePtrs.Count();
	FOR(int i=0;i<curveCount;) {
		const Curve *curve=m_m->curvePtrs.GetCurve(i);
		if(DoesDrawOpenCurve() || curve->IsClosed()) {
			if(DoesCurvePassLine(y,curve)) {
				ResolveCurve(curve);
				m_m->curvePtrs.Remove(i);
				--curveCount;
			} else {
				++i;
			}
		} else {
			m_m->curvePtrs.Remove(i);
			--curveCount;
		}
	}

	if(beforePointCount<m_m->pointQue.size()) {
		std::sort(m_m->pointQue.begin()+beforePointCount,m_m->pointQue.end(),IsAPreviouslyThanB);
		if(beforePointCount>0) {
			std::inplace_merge(m_m->pointQue.begin(),m_m->pointQue.begin()+beforePointCount,m_m->pointQue.end(),IsAPreviouslyThanB);
		}
	}
}

//yʂ邩
bool Fill::DoesCurvePassLine(int y,const Curve *curve)
{
	int knotCount=curve->KnotCount();
	FOR(int i=0;i<knotCount;++i) {
		const Knot *curKnot=curve->GetKnot(i);
		const Knot *nextKnot=curve->GetNextKnot(i);
		if(nextKnot==NULL) break;
	
		Bezier bezier(curKnot,nextKnot);
		if(bezier.DoesPassLine(y)) return true;
	}
	return false;
}

//𕪉
void Fill::ResolveCurve(const Curve *curve)
{
	int knotCount=curve->KnotCount();
	FOR(int i=0;i<knotCount;++i) {
		const Knot *curKnot=curve->GetKnot(i);
		const Knot *nextKnot=curve->GetNextKnot(i);
		if(nextKnot==NULL) break;

		ResolveSegment(curKnot,nextKnot);
	}

	if(DoesDrawOpenCurve() && !curve->IsClosed()) {
		FPoint pointA=curve->GetKnot(INDEX_FIRST)
			->GetPoint(PART_ANCHOR);
		FPoint pointB=curve->GetKnot(INDEX_LAST)
			->GetPoint(PART_ANCHOR);
		ResolveSegment(pointA,pointB);
	}
}

//P߂𕪉
void Fill::ResolveSegment(const Knot *knotA,const Knot *knotB)
{
	Bezier bezier(knotA,knotB);
	SubResolveSegment(&bezier);
}

void Fill::ResolveSegment(const FPoint &pointA,const FPoint &pointB)
{
	Bezier bezier(pointA,pointA,pointB,pointB);
	SubResolveSegment(&bezier);
}

void Fill::SubResolveSegment(Bezier *bezier)
{
	FOR(;;) {
		if(bezier->IsNarrowHeightByDot(m_antiAliasing)) {
			FPoint a=bezier->GetAnchorA();
			a.y=floor(a.y*m_antiAliasing);
			FPoint b=bezier->GetAnchorB();
			b.y=floor(b.y*m_antiAliasing);
			if(a.y>b.y) {
				m_m->pointQue.push_back(b);
			} else if(a.y<b.y) {
				m_m->pointQue.push_back(a);
			}
 			if(!bezier->Next()) break;
		} else {
			bezier->Divide();
		}
	}		//***vύX ̋Eh邩hȂ
}

//obt@̂Psh
void Fill::FillBuffer(int left,int y,double *buffer,int size)
{
	DisposePreviousLine(y);

	bool doesPut=false;

	PointQue::iterator it=m_m->pointQue.begin();
	PointQue::iterator endIt=m_m->pointQue.end();
	double start;
	for(;it!=endIt && (*it).y==y;++it) {
		if(!doesPut) {
			doesPut=true;
			start=(*it).x;
		} else {
			doesPut=false;
			SubFillBuffer(left,start,(*it).x,buffer,size);
		}
	}

	m_m->pointQue.erase(m_m->pointQue.begin(),it);
}

void Fill::SubFillBuffer(int left,double start,double end,double *buffer,int size)
{
	start-=left;
	end-=left;
	if(start<0) {
		if(end<0) {
			return;
		}
		start=0;
	}
	if(end>size) {
		if(start>=size) {
			return;
		}
		end=size;
	}
	if(start>=end) return;

	int startInt=static_cast<int>(floor(start));
	int endInt=static_cast<int>(floor(end));

	if(startInt==endInt) {
		buffer[startInt]+=(end-start)/m_antiAliasing;
	} else {
		buffer[startInt]+=(1.0-(start-startInt))/m_antiAliasing;

		FOR(int x=startInt+1;x<endInt;++x) {
			buffer[x]+=1.0/m_antiAliasing;
		}

		if(endInt<size) {
			buffer[endInt]+=(end-endInt)/m_antiAliasing;
		}
	}
}

//yȑO̍s̍W
void Fill::DisposePreviousLine(int y)
{
	PointQue::iterator it=m_m->pointQue.begin();
	PointQue::iterator endIt=m_m->pointQue.end();

	FOR(;it!=endIt && (*it).y<y;++it);

	m_m->pointQue.erase(m_m->pointQue.begin(),it);
}

//JhԂ
bool Fill::DoesDrawOpenCurve() const
{
	return (m_option&FILL_NOT_FILL_OPEN_CURVE)==0;
}

BZP_NAMESPACE_END
