#include "prefix.h"

#include <assert.h>
#include <math.h>

#include "fpoint.h"
#include "rect.h"
#include "mathdefine.h"

/*************************************************

	_ł̍WNX̒`

**************************************************/

BZP_NAMESPACE_START

//x,yŏ
FPoint::FPoint(double x,double y)
{
	this->x=x;
	this->y=y;
}

//POINTŏ
FPoint::FPoint(POINT pointInt)
{
	this->x=pointInt.x;
	this->y=pointInt.y;
}

//Rs[RXgN^
FPoint::FPoint(const FPoint & point)
{
	this->x=point.x;
	this->y=point.y;
}

FPoint::FPoint(const FPOINT &point)
{
	this->x=point.x;
	this->y=point.y;
}

FPoint::~FPoint() throw()
{
}

//FPoint̑
FPoint FPoint::operator =(const FPoint &point)
{
	this->x=point.x;
	this->y=point.y;
	return *this;
}

//FPOINT̑
FPoint FPoint::operator =(const FPOINT &point)
{
	this->x=point.x;
	this->y=point.y;
	return *this;
}

//POINT̑
FPoint FPoint::operator =(POINT pointInt)
{
	this->x=pointInt.x;
	this->y=pointInt.y;
	return *this;
}

//Z
FPoint FPoint::operator -(const FPoint & point) const
{
	FPoint result(*this);
	result-=point;
	return result;
}

//Z
FPoint FPoint::operator +(const FPoint & point) const
{
	FPoint result(*this);
	result+=point;
	return result;
}

//Zđ
FPoint FPoint::operator -=(const FPoint & point)
{
	this->x-= point.x;
	this->y-= point.y;
	return *this;
}

//Zđ
FPoint FPoint::operator +=(const FPoint & point)
{
	this->x+= point.x;
	this->y+= point.y;
	return *this;
}

//x yɂꂼ|Z
FPoint FPoint::operator *(double v) const
{
	double x=this->x*v;
	double y=this->y*v;
	return FPoint(x,y);
}

//x yɂꂼꊄZ
FPoint FPoint::operator /(double v) const
{
	double x=this->x/v;
	double y=this->y/v;
	return FPoint(x,y);
}

//r
bool FPoint::operator ==(const FPoint & point) const
{
	return ( (this->x == point.x)
		&& (this->y == point.y) );
}

//WłȂ
bool FPoint::operator !=(const FPoint & point) const
{
	return !(operator ==(point));
}

//tɂ
FPoint FPoint::operator -() const
{
	return FPoint(-x,-y);
}

//B܂t̊܂Ői񂾍W𓾂
FPoint FPoint::TPoint(const FPoint & pointB,double t) const
{
	FPoint point;
	point.TPoint(*this,pointB,t);
	return point;
}

//ABt̊܂Ői񂾍W
void FPoint::TPoint(const FPoint &pointA,const FPoint &pointB,double t)
{
#define TPOINT(a,b,t) (a+(b-a)*t)

	this->x= TPOINT(pointA.x,pointB.x,t);
	this->y= TPOINT(pointA.y,pointB.y,t);
}

//BƂ̒ԓ_
FPoint FPoint::Center(const FPoint & pointB) const
{
	return TPoint(pointB,0.5);
}

//BɌlengthړ_
FPoint FPoint::LeavePoint(const FPoint & pointB,double length) const
{
	double distance=Distance(pointB);	
	if(distance==0) {
		return *this;
	}
	return TPoint(pointB,length/distance);
}

//POINT\̂ɕϊ
FPoint::operator POINT() const
{
	POINT point;
	point.x= static_cast<int>(this->x + ((this->x>=0)?0.5:-0.5) );
	point.y= static_cast<int>(this->y + ((this->y>=0)?0.5:-0.5) );
	return point;
}

//Wϊ̉
static const double SubMapping(double frompt
	,int fromStart,int fromEnd
	,int toStart,int toEnd)
{
	int fromInterval=fromEnd-fromStart;
	assert(fromInterval>0);
	int toInterval=toEnd-toStart;

	return toStart+(frompt-fromStart)*toInterval/fromInterval;
}

//fromRecttoRect֍Wϊ
FPoint FPoint::Mapping(const Rect &fromRect, const Rect &toRect) const
{
	double x=SubMapping(this->x,fromRect.left,fromRect.right
		,toRect.left,toRect.right);
	double y=SubMapping(this->y,fromRect.top,fromRect.bottom
		,toRect.top,toRect.bottom);
	return FPoint(x,y);
}

//target𒆐SƂareȃ傫͈͓̔ɍW邩
bool FPoint::IsInArea(const FPoint & target,int area) const
{
	FRect targetRect;
	targetRect.SizedRect(target,area);
	return IsInRect(targetRect);
}

//W`̒ɂ邩
bool FPoint::IsInRect(const FRect &rect) const
{
	return (
			(rect.left<=rect.right && rect.left<=x && x<rect.right)
			||(rect.left>rect.right && rect.right<=x && x<rect.left)
		) && (
			(rect.top<=rect.bottom && rect.top<=y && y<rect.bottom)
			||(rect.top>rect.bottom && rect.bottom<=y && y<rect.top)
		);
}

//pointƂ̋
double FPoint::Distance(const FPoint & point) const
{
	return hypot(this->x-point.x,this->y-point.y);
}

//Wrootc΂߂̕Ɍ肷
void FPoint::FixAngle(const FPoint & root)
{
	double angle=BZP_ATAN_XY(this->x-root.x,this->y-root.y);
	angle+=BZP_PI*2+BZP_PI*2/8/2;
	double left=fmod(angle,BZP_PI*2/8);
	angle=angle-left;
	double length=Distance(root);
	this->x=root.x+cos(angle)*length;
	this->y=root.y+sin(angle)*length;
}

//W_𒆐Sɉ]
FPoint FPoint::Roll(double angle) const
{
	double newx=cos(angle)*this->x - sin(angle)*this->y;
	double newy=sin(angle)*this->x + cos(angle)*this->y;
	return FPoint(newx,newy);
}

//݂̓_𒆐SɂApoint̔Α̍W
FPoint FPoint::OppositePoint(const FPoint &point) const
{
	FPoint opposite=*this;
	opposite-=point;
	opposite+=*this;
	return opposite;
}


BZP_NAMESPACE_END
