#include "prefix.h"

#include <windows.h>
#include <assert.h>

#include "gdiobject.h"
#include "exception.h"
#include "error.h"

/*************************************************

	GDIIuWFNgL

**************************************************/

BZP_NAMESPACE_START


#ifndef NDEBUG
static int s_debug_dc_count=0;
static int s_debug_obj_count=0;
#endif

////////////GdiDC///////////////

GdiDC::GdiDC()
	:m_hDC(NULL)
	,m_penObject(NULL)
	,m_brushObject(NULL)
	,m_bitmapObject(NULL)
	,m_oldPen(NULL)
	,m_oldBrush(NULL)
	,m_oldBitmap(NULL)
{

}

GdiDC::GdiDC(HDC hDC)
	:m_hDC(hDC)
	,m_penObject(NULL)
	,m_brushObject(NULL)
	,m_bitmapObject(NULL)
	,m_oldPen(NULL)
	,m_oldBrush(NULL)
	,m_oldBitmap(NULL)
	,m_autoDelete(true)
{

}

GdiDC::~GdiDC() throw()
{
	Delete();
}

//LĂDCԂ
HDC GdiDC::Get() const
{	
	return m_hDC;
}

//hDCL
void GdiDC::Attach(HDC hDC)
{
	assert(m_hDC==NULL);
	m_hDC=hDC;
	m_autoDelete=true;
#ifndef NDEBUG
	++s_debug_dc_count;
#endif
}

//hDC֘At顏L͂Ȃ
void GdiDC::Rent(HDC hDC)
{
	Delete();
	m_hDC=hDC;
	m_autoDelete=false;
}

//DC폜
void GdiDC::Delete() throw()
{
	if(m_hDC!=NULL) {
		DeselectObject(OBJ_PEN);
		DeselectObject(OBJ_BRUSH);
		DeselectObject(OBJ_BITMAP);
		if(m_autoDelete) {
#ifdef NDEBUG
			::DeleteDC(m_hDC);
#else
			BOOL success=::DeleteDC(m_hDC);
			assert(success);
			--s_debug_dc_count;
#endif
		}
	}
	m_hDC=NULL;
}

//objectI
void GdiDC::SelectObject(GdiObject *object)
{
	assert(object!=NULL);

	DWORD type=object->Type();
	DeselectObject(type);

	HGDIOBJ oldObject=::SelectObject(m_hDC,object->Get());
	if(oldObject==NULL) {
		throw Exception(BZPERROR_CANT_SELECT_GDI_OBJECT);
	}

	switch(type) {
	case OBJ_BITMAP:
		m_bitmapObject=object;
		m_oldBitmap=oldObject;
		break;

	case OBJ_PEN:
		m_penObject=object;
		m_oldPen=oldObject;
		break;

	case OBJ_BRUSH:
		m_brushObject=object;
		m_oldBrush=oldObject;
		break;

	default:
		assert(false);
	}

	object->Select(this);
}

//type̎ނ̃IuWFNgO
void GdiDC::DeselectObject(DWORD type) throw()
{
	HGDIOBJ oldObject;
	GdiObject *object;
	switch(type) {
	case OBJ_BITMAP:
		oldObject=m_oldBitmap;
		m_oldBitmap=NULL;
		object=m_bitmapObject;
		m_bitmapObject=NULL;
		break;

	case OBJ_PEN:
		oldObject=m_oldPen;
		m_oldPen=NULL;
		object=m_penObject;
		m_penObject=NULL;
		break;

	case OBJ_BRUSH:
		oldObject=m_oldBrush;
		m_oldBrush=NULL;
		object=m_brushObject;
		m_brushObject=NULL;
		break;

	default:
		assert(false);
	}
	if(object!=NULL) {
		object->Deselect();
		assert(oldObject!=NULL);
#ifdef NDEBUG
		::SelectObject(m_hDC,oldObject);
#else
		HGDIOBJ selectedObj;
		selectedObj=::SelectObject(m_hDC,oldObject);
		assert(selectedObj!=NULL);
#endif
	}
}

/////////////GdiObject/////////

GdiObject::GdiObject()
	:m_dc(NULL)
	,m_object(NULL)
{

}

GdiObject::GdiObject(DWORD type,HGDIOBJ object)
	:m_dc(NULL)
	,m_object(NULL)
{
	Attach(type,object);
}

GdiObject::GdiObject(DWORD type,int number)
	:m_dc(NULL)
	,m_object(NULL)
{
	SetStockObject(type,number);
}

GdiObject::~GdiObject() throw()
{
	Delete();
}

//GDIIuWFNgL
void GdiObject::Attach(DWORD type,HGDIOBJ object)
{
	Delete();
	assert(object!=NULL);		
	m_object=object;
	m_type=type;

#ifndef NDEBUG
	++s_debug_obj_count;
#endif
}

//XgbNIuWFNgL
void GdiObject::SetStockObject(DWORD type,int number)
{
	Delete();

	m_object=::GetStockObject(number);
	m_type=type;

#ifndef NDEBUG
	++s_debug_obj_count;
#endif
}

//LĂGDIIuWFNg𓾂
HGDIOBJ GdiObject::Get() const
{
	return m_object;
}

//GDIIuWFNg폜XgbNIuWFNgȂ؂藣
void GdiObject::Delete() throw()
{
	if(m_object!=NULL) {
		if(m_dc!=NULL) {
			m_dc->DeselectObject(Type());
			m_dc=NULL;
		}
//		if(!m_isStockObject) {
#ifdef NDEBUG
			::DeleteObject(m_object);
#else
			BOOL success=::DeleteObject(m_object);
			assert(success);
			--s_debug_obj_count;
#endif
//		}
		m_object=NULL;
	}
}

//DCɑIꂽƂGdiDCĂ΂
void GdiObject::Select(GdiDC *dc)
{
	m_dc=dc;		
}

//DC̑I~߂ƂGdiDCĂ΂
void GdiObject::Deselect() throw()
{
	m_dc=NULL;
}

//GDIIuWFNg̎
DWORD GdiObject::Type() const throw()
{
	return m_type;
}

GdiObject::operator HBRUSH()
{
	assert(Type()==OBJ_BRUSH);
	return static_cast<HBRUSH>(m_object);
}

GdiObject::operator HPEN()
{
	assert(Type()==OBJ_PEN);
	return static_cast<HPEN>(m_object);
}

GdiObject::operator HBITMAP()
{
	assert(Type()==OBJ_BITMAP);
	return static_cast<HBITMAP>(m_object);
}

GdiObject::operator HRGN()
{
	assert(Type()==OBJ_REGION);
	return static_cast<HRGN>(m_object);
}

#ifndef NDEBUG

void GdiResourceLeakCheck()
{
	assert(s_debug_dc_count==0 && s_debug_obj_count==0);
}

#endif

BZP_NAMESPACE_END
