#include "prefix.h"

#include <windows.h>
#include <assert.h>
#include <new>

#include "graph.h"
#include "error.h"
#include "fpoint.h"
#include "runtimeconfig.h"
#include "rect.h"
#include "drawmode.h"
#include "exception.h"
#include "gdiobject.h"
#include "partssize.h"

/*************************************************

	\Wϊ̃NX

**************************************************/

BZP_NAMESPACE_START

//pXhƂƓhȂƂWINAPI̐FAuVy
#define PAINTCOLOR	0xffffff
#define CLEARCOLOR	0
#define PAINTBRUSH	WHITE_BRUSH
#define CLEARBRUSH	BLACK_BRUSH
#define PAINTPEN	WHITE_PEN
#define	CLEARPEN	BLACK_PEN

//̖蕔̈ʒu
//0̏ꍇ͖̐[ړI̍Ww
//1̏ꍇ͖̕tړI̍Ww
#define ARROWCENTER	0.6

//`p̃rbg}bv
static GdiObject *s_beforeBitmap=NULL;	//ύXOL^rbg}bv
static GdiObject *s_afterBitmap=NULL;	//ύXL^rbg}bv
static SIZE s_allocateSize={0,0};	//rbg}bv̑傫
static int s_objectCount=0;	//GraphIuWFNg쐬񐔡0ɂȂÓIIuWFNg폜

static bool s_isInUse=false;	//`擮ĂŒ͎gȂ悤ɂ

static const int s_defaultPartsSize[PARTSSIZE_MAX]={
//ȋ傫̏ݒ
	7,	//ʉߓ_̑傫
	6,	//_̑傫
	7,	//IĂ炸Ǘ_̑傫
	8,	//O̕_ɂƂ̕
	10,	//̒

//i̒T͈̔͂̏ݒi\ƈĂĂǂj
	7,	//ʉߓ_̒T͈
	6,	//_̒T͈
	7,	//Ǘ_̒T͈
	8,	//O̕_ɂƂ̒T͈
	6,	//̒T͈
};

class GraphMember
{
public:
	GraphMember()
		:viewRect(0,0,0,0)
		,boundRect(0,0,0,0)
	{
		FOR(int i=0;i<PARTSSIZE_MAX;++i) {
			partsSize[i]=s_defaultPartsSize[i];
		}
	}
	GdiDC viewDC;	//\Ώۂ̃foCXReLXg
	Rect viewRect;	//\W`
	Rect boundRect;
	int partsSize[PARTSSIZE_MAX];
};

Graph::Graph()
	:m_m(new GraphMember)
	,m_color(0xffffff)
	,m_drawMode(DRAWMODE_VIEW)
	,m_drawDC(NULL)
	,m_beforeDC(NULL)
{
	BZP_BAD_ALLOC_TEST(m_m.get());
	if(s_beforeBitmap==NULL) {
		s_beforeBitmap=new GdiObject;
		BZP_BAD_ALLOC_TEST(s_beforeBitmap);
	}
	if(s_afterBitmap==NULL) {
		s_afterBitmap=new GdiObject;
		BZP_BAD_ALLOC_TEST(s_afterBitmap);
	}
	++s_objectCount;
}

Graph::~Graph() throw()
{
	--s_objectCount;
	if(s_objectCount==0) {
		DestroyStaticObjects();
	}
}

//\foCXReLXgݒ
void Graph::SetViewDC(HDC hDC)
{
	BZP_PARAMETER_ERROR(hDC!=NULL,BZPERROR_NULL_DC);
	m_m->viewDC.Rent(hDC);
}

//W\`ݒ
void Graph::SetBoundRect(const Rect &rect)
{
	BZP_PARAMETER_ERROR(!rect.IsEmpty(),BZPERROR_EMPTY_RECT);
	m_m->boundRect.Set(rect);	
}

void Graph::SetBoundRect(const RECT &rect)
{
	BZP_PARAMETER_ERROR(rect.right-rect.left!=0
		&&rect.bottom-rect.top!=0,BZPERROR_EMPTY_RECT);
	m_m->boundRect.Set(rect);
}

//\W\`ݒ
void Graph::SetViewRect(const Rect &rect)
{
	BZP_PARAMETER_ERROR(!rect.IsEmpty(),BZPERROR_EMPTY_RECT);
	m_m->viewRect.Set(rect);
}

void Graph::SetViewRect(const RECT &rect)
{
	BZP_PARAMETER_ERROR(rect.right-rect.left!=0
		&& rect.bottom-rect.top!=0,BZPERROR_EMPTY_RECT);
	m_m->viewRect.Set(rect);
}

//\Fݒ
void Graph::SetColor(COLORREF color)
{
	m_color=color;
}

//\F𔽓]Đݒ
void Graph::SetInvertColor(COLORREF color)
{
	SetColor((~color) & 0xffffff);
}

//\@̐ݒ
void Graph::SetDrawMode(int mode)
{
	m_drawMode=mode;
}

//݂̕\@𓾂
int Graph::GetDrawMode() const
{
	return m_drawMode;
}

//iT͈͂̑傫ݒ
void Graph::SetPartsSize(PARTSSIZE parts,int size)
{
	m_m->partsSize[parts]=size;
}

//iT͈͂̑傫𓾂
int Graph::GetPartsSize(PARTSSIZE parts) const
{
	return m_m->partsSize[parts];
}

//Wr[̕\̈ɂ邩
bool Graph::IsPointInView(POINT point) const
{
	assert(m_m->viewDC.Get()!=NULL);
	return ::PtVisible(m_m->viewDC.Get(),point.x,point.y)!=FALSE;
}

//rbg}bvgpĂ邩
//ɕ\XV邱Ƃ͂łȂ
bool Graph::IsInUse() const
{
	return s_isInUse;
}

//Ɨprbg}bv̊m
//\̈悪ȑO\̂傫΃rbg}bv傫
void Graph::AllocateBuffer()
{
	assert(m_m->viewDC.Get()!=NULL);

	Rect clipRect;
	CalcClipRect(&clipRect);

	SIZE bitmapSize;
	bitmapSize.cx=clipRect.Width();
	if(bitmapSize.cx<=0) bitmapSize.cx=1;
	bitmapSize.cy=clipRect.Height();
	if(bitmapSize.cy<=0) bitmapSize.cy=1;

	if((bitmapSize.cx<=s_allocateSize.cx)
		&&(bitmapSize.cy<=s_allocateSize.cy)) {
		return;
	}

	DestroyBitmap();

	s_beforeBitmap->Attach(OBJ_BITMAP
		,::CreateBitmap(bitmapSize.cx,bitmapSize.cy,1,1,NULL));
	s_afterBitmap->Attach(OBJ_BITMAP
		,::CreateBitmap(bitmapSize.cx,bitmapSize.cy,1,1,NULL));

	if(s_beforeBitmap->Get()==NULL||s_afterBitmap->Get()==NULL) {
		DestroyBitmap();
		throw Exception(BZPERROR_CANT_CREATE_BITMAP);
	}
	s_allocateSize=bitmapSize;
}

//Ɨprbg}bv̉
void Graph::DestroyBitmap()
{
	if(s_beforeBitmap!=NULL) {
		s_beforeBitmap->Delete();
	}
	if(s_afterBitmap!=NULL) {
		s_afterBitmap->Delete();
	}
	s_allocateSize.cx=s_allocateSize.cy=0;
}

//ÓIϐɊ蓖ĂIuWFNg폜
void Graph::DestroyStaticObjects()
{
	DestroyBitmap();
	delete s_beforeBitmap;
	s_beforeBitmap=NULL;
	delete s_afterBitmap;
	s_afterBitmap=NULL;
}

//rbg}bv[foCXReLXgɊ蓖Ă
void Graph::CreateMemoryDC(GdiDC *dc,GdiObject *bitmap)
{
	assert(dc!=NULL && bitmap->Get()!=NULL);

	HDC hDC=::CreateCompatibleDC(NULL);
	if(hDC==NULL) {
		throw Exception(BZPERROR_CANT_CREATE_DC);
	}
	dc->Attach(hDC);
	dc->SelectObject(bitmap);

	Rect clipRect;
	CalcClipRect(&clipRect);
	::SetWindowOrgEx(dc->Get(),clipRect.left,clipRect.top,NULL);

	GdiObject clearBrush(OBJ_BRUSH,CLEARBRUSH);
	::FillRect(hDC,&clipRect,HBRUSH(clearBrush));
}

//ĕ`̏pX̕ύXO
void Graph::PrepareBeforeBitmap()
{
	if(s_isInUse) {
		throw Exception(BZPERROR_GRAPH_LOCKED);
	}

	AllocateBuffer();

	assert(m_drawDC.get()==NULL);
	BZP_AUTOPTR_RESET(m_drawDC,GdiDC,new GdiDC);
	BZP_BAD_ALLOC_TEST(m_drawDC.get());
	CreateMemoryDC(m_drawDC.get(),s_beforeBitmap);

	s_isInUse=true;
}

//ĕ`̏pX̕ύX
void Graph::PrepareAfterBitmap()
{
	assert(s_isInUse && m_drawDC.get()!=NULL);
	m_beforeDC=m_drawDC;
	BZP_AUTOPTR_RESET(m_drawDC,GdiDC,new GdiDC);
	BZP_BAD_ALLOC_TEST(m_drawDC.get());
	CreateMemoryDC(m_drawDC.get(),s_afterBitmap);
}

//O̍\
void Graph::DrawDifference()
{
	assert(s_isInUse && m_drawDC.get()!=NULL && m_beforeDC.get()!=NULL);

	Rect clipRect;
	CalcClipRect(&clipRect);

	::BitBlt(m_drawDC->Get(),clipRect.left,clipRect.top
		,clipRect.Width()
		,clipRect.Height()
		,m_beforeDC->Get(),clipRect.left,clipRect.top
		,SRCINVERT);

	BZP_AUTOPTR_RESET(m_beforeDC,GdiDC,NULL);

	DrawAll();
}

//pX̑SĂ̕\
void Graph::DrawAll()
{
	HDC viewDC=m_m->viewDC.Get();

	assert(s_isInUse && m_drawDC.get()!=NULL && viewDC!=NULL);

	::SetBkMode(viewDC,TRANSPARENT);
	::SetBkColor(viewDC,m_color);
	::SetTextColor(viewDC,0);

	Rect clipRect;
	CalcClipRect(&clipRect);

	::BitBlt(viewDC,clipRect.left,clipRect.top
		,clipRect.Width()
		,clipRect.Height()
		,m_drawDC->Get(),clipRect.left,clipRect.top
		,SRCINVERT);

	BZP_AUTOPTR_RESET(m_drawDC,GdiDC,NULL);
	s_isInUse=false;
}

//\𒆎~
void Graph::CancelDraw() throw()
{
	assert(s_isInUse);
	s_isInUse=false;

	BZP_AUTOPTR_RESET(m_drawDC,GdiDC,NULL);
	BZP_AUTOPTR_RESET(m_beforeDC,GdiDC,NULL);
}


//Ȑ`
void Graph::DrawCurve(const FPoint & anchorA,const FPoint & forwardA,const FPoint & backB,const FPoint & anchorB) const
{
	const int pointCount=4;
	POINT points[pointCount];

	points[0]=BoundToView(anchorA);
	points[1]=BoundToView(forwardA);
	points[2]=BoundToView(backB);
	points[3]=BoundToView(anchorB);

	GdiObject paintPen(OBJ_PEN,PAINTPEN);
	m_drawDC->SelectObject(&paintPen);

	::PolyBezier(m_drawDC->Get(),points,pointCount);
}

//ʉߓ_`
void Graph::DrawAnchor(const FPoint &anchor,bool isSelected) const
{
	POINT viewAnchor=BoundToView(anchor);

	GdiObject paintBrush(OBJ_BRUSH,PAINTBRUSH);

	Rect anchorRect;
	int anchorSize=GetPartsSize(PARTSSIZE_ANCHORSIZE);
	anchorRect.SizedRect(viewAnchor,anchorSize);
	::FillRect(m_drawDC->Get(),&anchorRect,HBRUSH(paintBrush));
	
	if(!isSelected) {
		GdiObject clearBrush(OBJ_BRUSH,CLEARBRUSH);
		Rect digAnchorRect;
		assert(anchorSize-2>0);
		digAnchorRect.SizedRect(viewAnchor,anchorSize-2);
		::FillRect(m_drawDC->Get(),&digAnchorRect,HBRUSH(clearBrush));
	}
}

//_Ɩ_`
void Graph::DrawControlHandle(const FPoint & anchor,const FPoint & control) const
{
	POINT viewAnchor=BoundToView(anchor);
	POINT viewControl=BoundToView(control);

	GdiObject paintBrush(OBJ_BRUSH,PAINTBRUSH);
	m_drawDC->SelectObject(&paintBrush);
	GdiObject paintPen(OBJ_PEN,PAINTPEN);
	m_drawDC->SelectObject(&paintPen);

	Rect controlRect;
	int controlSize=GetPartsSize(PARTSSIZE_CONTROLSIZE);
	controlRect.SizedRect(viewControl,controlSize);
	::Ellipse(m_drawDC->Get(),controlRect.left,controlRect.top
		,controlRect.right,controlRect.bottom);

	::MoveToEx(m_drawDC->Get(),viewAnchor.x,viewAnchor.y,NULL);
	::LineTo(m_drawDC->Get(),viewControl.x,viewControl.y);
}

//_ŕ`
void Graph::DrawArrowControl(const FPoint & anchor,const FPoint & control) const
{
	POINT viewAnchor=BoundToView(anchor);
	POINT viewControl=BoundToView(control);

	int width=GetPartsSize(PARTSSIZE_ARROW_WIDTH);
	int height=GetPartsSize(PARTSSIZE_ARROW_LENGTH);

	double angle=BZP_ATAN_XY(viewControl.x-viewAnchor.x,viewControl.y-viewAnchor.y);

	FPoint fpoints[3];
	fpoints[0].x=height*ARROWCENTER;
	fpoints[0].y=0;
	fpoints[1].x=-height*(1.0-ARROWCENTER);
	fpoints[1].y=-width/2.0;
	fpoints[2].x=-height*(1.0-ARROWCENTER);
	fpoints[2].y=width/2.0;

	POINT points[3];
	FOR(int i=0;i<3;++i) {
		points[i]=fpoints[i].Roll(angle);
		points[i].x+=viewControl.x;
		points[i].y+=viewControl.y;
	}
	GdiObject rgn;
	rgn.Attach(OBJ_REGION
		,reinterpret_cast<HGDIOBJ>(::CreatePolygonRgn(points,3,ALTERNATE)));

	GdiObject paintBrush(OBJ_BRUSH,PAINTBRUSH);
	::FillRgn(m_drawDC->Get(),HRGN(rgn),HBRUSH(paintBrush));

	GdiObject paintPen(OBJ_PEN,PAINTPEN);
	m_drawDC->SelectObject(&paintPen);

	::MoveToEx(m_drawDC->Get(),viewAnchor.x,viewAnchor.y,NULL);
	::LineTo(m_drawDC->Get(),viewControl.x,viewControl.y);
}

//IĂ炸Ǘpڂ`
void Graph::DrawSingleKnot(const FPoint & anchor) const
{
	POINT viewAnchor=BoundToView(anchor);

	Rect anchorRect;
	int singleKnotSize=GetPartsSize(PARTSSIZE_SINGLEKNOTSIZE);
	anchorRect.SizedRect(viewAnchor,singleKnotSize);

	GdiObject paintPen(OBJ_PEN,PAINTPEN);
	m_drawDC->SelectObject(&paintPen);

	HDC hDC=m_drawDC->Get();
	::MoveToEx(hDC,anchorRect.left,anchorRect.top,NULL);
	::LineTo(hDC,anchorRect.right,anchorRect.bottom);
	::MoveToEx(hDC,anchorRect.left,anchorRect.bottom-1,NULL);
	::LineTo(hDC,anchorRect.right,anchorRect.top-1);
}

//`̍Wϊ
void Graph::BoundToView(Rect *viewRect,const FRect &boundRect) const
{
	FPoint cornerA(boundRect.left,boundRect.top);
	POINT viewPointA=BoundToView(cornerA);
	FPoint cornerB(boundRect.right,boundRect.bottom);
	POINT viewPointB=BoundToView(cornerB);
	viewRect->Set(viewPointA.x,viewPointA.y
		,viewPointB.x,viewPointB.y);
}


//`I}``̎lp܂͉~`
//Ȃ[foCXReLXgɂ͓_̕\܂Ȃ̂ŁA͉ʂɒɕ`
void Graph::DrawSelectShape(const FRect &rect,bool isCircle,bool dot) const
{
	HDC viewDC=m_m->viewDC.Get();
	int oldrop=::SetROP2(viewDC,R2_XORPEN);
	GdiObject brush(OBJ_BRUSH,NULL_BRUSH);
	m_m->viewDC.SelectObject(&brush);

	int penStyle;
	if(dot) {
		penStyle=PS_DOT;
	} else {
		penStyle=PS_SOLID;
	}
	GdiObject pen(OBJ_PEN,::CreatePen(penStyle,1,PAINTCOLOR));
	m_m->viewDC.SelectObject(&pen);

	Rect viewRect;
	BoundToView(&viewRect,rect);

	if(!isCircle) {
		::Rectangle(viewDC,viewRect.left,viewRect.top,viewRect.right,viewRect.bottom);
	} else {
		::Ellipse(viewDC,viewRect.left,viewRect.top,viewRect.right,viewRect.bottom);
	}
	::SetROP2(viewDC,oldrop);
}

//W\Wɕϊ
POINT Graph::BoundToView(const FPoint & boundPoint) const
{
	FPoint viewPoint;
	viewPoint=boundPoint.Mapping(m_m->boundRect,m_m->viewRect);
	return POINT(viewPoint);
}

//\WWɕϊ
FPoint Graph::ViewToBound(POINT viewPoint) const
{
	FPoint viewPointF=viewPoint;
	return viewPointF.Mapping(m_m->viewRect,m_m->boundRect);
}

//̒T͈̔͂WϊĕԂ
FPoint Graph::GetCurveArea() const
{
	POINT viewLeftTop={0,0};
	FPoint boundLeftTop=ViewToBound(viewLeftTop);
	POINT viewRightBottom;
	viewRightBottom.x=viewRightBottom.y=GetPartsSize(PARTSSIZE_CURVEAREA);
	FPoint boundRightBottom=ViewToBound(viewRightBottom);
	FPoint size=boundRightBottom-boundLeftTop;
	if(size.x<0) {
		size.x=-size.x;
	}
	if(size.y<0) {
		size.y=-size.y;
	}
	
	return size;
}

//\ʂ̃foCXReLXg
HDC Graph::GetViewDC() const
{
	return m_m->viewDC.Get();
}

//̃rbg}bṽfoCXReLXg
HDC Graph::GetBitmapDC() const
{
	assert(m_drawDC->Get()!=NULL);
	return m_drawDC->Get();
}

//XV`
void Graph::CalcClipRect(Rect *rect) const
{
	::GetClipBox(m_m->viewDC.Get(),static_cast<RECT*>(rect));
	if(!DoesDrawMargin()) {
		rect->Intersect(m_m->viewRect);
	}
}

//pڂ\邩
bool Graph::DoesDrawKnot() const
{
	return (m_drawMode&DRAWMODE_HIDE_KNOT)==0
		&& DoesDraw();
}

//gOɕ\邩
bool Graph::DoesDrawMargin() const
{
	return (m_drawMode&DRAWMODE_HIDE_MARGIN)==0;
}

//IĂȂƃO[v\邩
bool Graph::DoesDrawUnselected() const
{
	return (m_drawMode&DRAWMODE_HIDE_UNSELECTED)==0
		&& DoesDraw();
}

//O̕_ɂ邩
bool Graph::DoesDrawArrowControl() const
{
	return (m_drawMode&DRAWMODE_ARROW_FORWARDCONTROL)!=0
		&& DoesDraw();
}

//pX\邩
bool Graph::DoesDraw() const
{
	return (m_drawMode&DRAWMODE_HIDE_ALL)==0;
}

BZP_NAMESPACE_END
