#include "prefix.h"

#include <memory>
#include <new>

#include "group.h"
#include "curve.h"
#include "subgroup.h"
#include "fpoint.h"
#include "selectmode.h"
#include "ptrdeque.h"
#include "findresult.h"
#include "partmask.h"
#include "movecurvemode.h"
#include "graph.h"
#include "curvepointers.h"
#include "rect.h"
#include "part.h"
#include "index.h"
#include "converterbase.h"
#include "transform.h"

/*************************************************

	̋ȐƃTuO[v܂Ƃ߂NX̒`

**************************************************/

BZP_NAMESPACE_START

BZP_PARTMASK_USING

typedef PtrDeque<Curve> CurveDeque;
typedef PtrDeque<SubGroup> SubGroupDeque;

class GroupMember
{
public:
	CurveDeque curves;
	SubGroupDeque subGroups;
};

Group::Group()
	:m_m(new GroupMember)
{
	BZP_BAD_ALLOC_TEST(m_m.get());
}

Group::~Group() throw()
{
}

//pX`
void Group::DrawGroup(const Graph *graph) const
{
	if(!graph->DoesDraw()) return;

	int curveCount=CurveCount();
	FOR(int i=0;i<curveCount;++i) {
		GetCurve(i)->Draw(graph);
	}

	int subGroupCount=SubGroupCount();
	FOR(int i=0;i<subGroupCount;++i) {
		GetSubGroup(i)->DrawSubGroup(graph);
	}

	if(graph->DoesDrawKnot()) {
		FOR(int i=0;i<curveCount;++i) {
			GetCurve(i)->DrawAnchors(graph);
		}
	}
}

//LĂ̐
int Group::CurveCount() const
{
	return m_m->curves.Count();
}

//LĂTuO[v̐
int Group::SubGroupCount() const
{
	return m_m->subGroups.Count();
}

//vfȂ
bool Group::IsEmpty() const
{
	return (CurveCount()+SubGroupCount())==0;
}

//indexԖڂ̐𓾂
Curve *Group::GetCurve(int index)
{
	return m_m->curves.At(index);
}

//indexԖڂ̐𓾂
const Curve *Group::GetCurve(int index) const
{
	return m_m->curves.At(index);
}

//indexԖڂ̃TuO[v𓾂
SubGroup *Group::GetSubGroup(int index)
{
	return m_m->subGroups.At(index);
}

//indexԖڂ̃TuO[v𓾂
const SubGroup *Group::GetSubGroup(int index) const
{
	return m_m->subGroups.At(index);
}

//̏Ɍpڂǉ
int Group::InsertKnot(INSERTKNOTMODE mode,double t,int knotIndex,int curveIndex)
{
	return GetCurve(curveIndex)->InsertKnot(mode,t,knotIndex);
}

//̖[Ɍpڂ
int Group::AddEndKnot(const FPoint & point,int knotIndex,int curveIndex)
{
	return GetCurve(curveIndex)->AddEndKnot(point,knotIndex);
}

//pڂ폜
void Group::DeleteKnot(DELETEKNOTMODE mode,int knotIndex,int curveIndex)
{
	GetCurve(curveIndex)->DeleteKnot(mode,knotIndex);
}

//curveIndexw̌pڂ̐
int Group::KnotCount(int curveIndex) const
{
	return GetCurve(curveIndex)->KnotCount();
}

//pڂI
void Group::SelectKnot(SELECTMODE mode,int knotIndex,int curveIndex)
{
	if(mode==SELECTMODE_SINGLE) {
		SelectAllCurves(false);
		SelectAllSubGroups(false);
	}
	GetCurve(curveIndex)->SelectKnot(mode,knotIndex);	
}

//pڂIĂ邩
bool Group::IsSelectedKnot(int knotIndex,int curveIndex) const
{
	return GetCurve(curveIndex)->IsSelectedKnot(knotIndex);
}

//w肵pڂ̕ȂĂ邩
void Group::LinkKnot(bool isLinked,int knotIndex,int curveIndex)
{
	GetCurve(curveIndex)->LinkKnot(isLinked,knotIndex);
}

//w肵pڂȐɂȂĂ邩
void Group::CurveKnot(bool isCurved,int part,int knotIndex,int curveIndex)
{
	GetCurve(curveIndex)->CurveKnot(isCurved,part,knotIndex);
}

//pڂȐɂȂĂ邩
bool Group::IsCurved(int part,int knotIndex,int curveIndex) const
{
	return GetCurve(curveIndex)->IsCurved(part,knotIndex);
}

//Iꂽpڂ̐
int Group::SelectedKnotCount() const
{
	int count=0;
	FOR(int i=0;i<CurveCount();++i) {
		const Curve *curve=GetCurve(i);
		if(curve->IsSelected()) {
			count+=curve->SelectedKnotCount();
		}
	}
	return count;
}

//pڂ̔ԍ̈ʒu\lɂ
int Group::SymbolizeKnotIndex(int knotIndex,int curveIndex) const
{
	return GetCurve(curveIndex)->SymbolizeKnotIndex(knotIndex);
}

//IĂpڂ𑊑ΓIɈړ
void Group::MoveSelectedKnots(const FPoint & rel)
{
	int curveCount=CurveCount();
	FOR(int i=0;i<curveCount;++i) {
		GetCurve(i)->MoveSelectedKnots(rel);
	}
}

//pڂ̒ʉߓ__̍W𓾂
FPoint Group::GetPoint(int part,int knotIndex,int curveIndex) const
{
	return GetCurve(curveIndex)->GetPoint(part,knotIndex);
}

//pڂ̒ʉߓ__ړ
void Group::MovePoint(const FPoint & point,int part,int knotIndex,int curveIndex)
{
	GetCurve(curveIndex)->MovePoint(point,part,knotIndex);
}

//pڂ̏Ԃςɒʉߓ__̍WύX
void Group::SetPoint(const FPoint & point,int part,int knotIndex,int curveIndex)
{
	GetCurve(curveIndex)->SetPoint(point,part,knotIndex);
}

//pڂ̕𗼑ɐL΂
void Group::MoveWControls(const FPoint & point,bool grippedForward,int knotIndex,int curveIndex)
{
	GetCurve(curveIndex)->MoveWControls(point,grippedForward,knotIndex);
}

//Vǉ
int Group::NewCurve(int curveIndex)
{
	std::auto_ptr<Curve> newCurve(new Curve);
	BZP_BAD_ALLOC_TEST(newCurve.get());

	int afterCurveIndex=m_m->curves.InsertAfter(newCurve.get(),curveIndex);
	newCurve.release();
	return afterCurveIndex;
}

//̖[mȂ
void Group::Chain(int *knotIndexA,int *curveIndexA,bool chainForward,int knotIndexB,int curveIndexB,bool deleteKnotB)
{
	*knotIndexA=GetCurve(*curveIndexA)->Chain(*knotIndexA,knotIndexB,chainForward,GetCurve(curveIndexB),deleteKnotB);

	int newIndexA=m_m->curves.NormalizeIndex(*curveIndexA);
	curveIndexB=m_m->curves.NormalizeIndex(curveIndexB);

	if(newIndexA!=curveIndexB) {
		DeleteCurve(curveIndexB);
		if(newIndexA>curveIndexB) {
			newIndexA--;
		}
	}
	*curveIndexA=newIndexA;
}

//폜
void Group::DeleteCurve(int curveIndex)
{
	m_m->curves.Remove(curveIndex);
}

//TuO[v폜
void Group::DeleteSubGroup(int subGroupIndex)
{
	m_m->subGroups.Remove(subGroupIndex);
}

//I
void Group::SelectCurve(SELECTMODE mode,int curveIndex)
{
	Curve *curCurve=NULL;
	if(curveIndex!=INDEX_NONE
		&& mode!=SELECTMODE_ALL
		&& mode!=SELECTMODE_UNSELECTALL) {
		curCurve=GetCurve(curveIndex);
	}
	switch(mode) {
	case SELECTMODE_ADD:
		curCurve->Select(true);
		break;

	case SELECTMODE_SINGLE:
		SelectAllCurves(false);
		SelectAllSubGroups(false);
		curCurve->Select(true);
		break;

	case SELECTMODE_UNSELECT:
		curCurve->Select(false);
		break;

	case SELECTMODE_TOGGLE:
		curCurve->Select(!curCurve->IsSelected());
		break;

	case SELECTMODE_ALL:
		SelectAllCurves(true);
		break;

	case SELECTMODE_UNSELECTALL:
		SelectAllCurves(false);
		break;

	default:
		assert(false);
	}
}

//SĂ̐IEI
void Group::SelectAllCurves(bool isSelected)
{
	int curveCount=CurveCount();
	FOR(int i=0;i<curveCount;++i) {
		GetCurve(i)->Select(isSelected);
	}
}

//SẴTuO[vIEI
void Group::SelectAllSubGroups(bool isSelected)
{
	int subGroupCount=SubGroupCount();
	FOR(int i=0;i<subGroupCount;++i) {
		GetSubGroup(i)->Select(isSelected);
	}
}

//IĂ邩
bool Group::IsSelectedCurve(int curveIndex) const
{
	return GetCurve(curveIndex)->IsSelected();
}

//ړ
void Group::MoveCurve(const FPoint & newPoint,const FPoint & oldPoint,MOVECURVEMODE mode,int knotIndex,int curveIndex)
{
	if(mode==MOVECURVE_ALLSELECTED) {
		MoveSelectedCurves(newPoint-oldPoint);
	} else {
		GetCurve(curveIndex)->Move(newPoint,oldPoint,mode,knotIndex);
	}
}

//IĂ𑊑ΓIɈړ
void Group::MoveSelectedCurves(const FPoint & rel)
{
	int curveCount=CurveCount();
	FOR(int i=0;i<curveCount;++i) {
		Curve *curve=GetCurve(i);
		if(curve->IsSelected()) {
			curve->MoveRel(rel);
		}
	}
}

//wԍ
bool Group::IsSameCurve(int indexA,int indexB) const
{
	return m_m->curves.NormalizeIndex(indexA)==m_m->curves.NormalizeIndex(indexB);
}

//Ă邩
bool Group::IsClosedCurve(int curveIndex) const
{
	return GetCurve(curveIndex)->IsClosed();
}

//EJ
void Group::CloseCurve(bool isClosed,int curveIndex)
{
	GetCurve(curveIndex)->Close(isClosed);
}

//IꂽȐ̐
int Group::SelectedCurveCount() const
{
	int count=0;
	FOR(int i=0;i<CurveCount();++i) {
		if(GetCurve(i)->IsSelected()) {
			++count;
		}
	}
	return count;
}

//IuWFNgǉ
void Group::AddCurve(Curve *curve)
{
	m_m->curves.Add(curve);
}

//w肵LȂ
void Group::ReleaseCurve(int curveIndex)
{
	m_m->curves.ReleaseItem(curveIndex);
}

//TuO[vI
void Group::SelectSubGroup(SELECTMODE mode,int subGroupIndex)
{
	SubGroup *curSubGroup=NULL;
	if(subGroupIndex!=INDEX_NONE
		&& mode!=SELECTMODE_ALL
		&& mode!=SELECTMODE_UNSELECTALL) {

		curSubGroup=GetSubGroup(subGroupIndex);
	}
	switch(mode) {
	case SELECTMODE_ADD:
		curSubGroup->Select(true);
		break;

	case SELECTMODE_SINGLE:
		SelectAllCurves(false);
		SelectAllSubGroups(false);
		curSubGroup->Select(true);
		break;

	case SELECTMODE_UNSELECT:
		curSubGroup->Select(false);
		break;

	case SELECTMODE_TOGGLE:
		curSubGroup->Select(!curSubGroup->IsSelected());
		break;

	case SELECTMODE_ALL:
		SelectAllSubGroups(true);
		break;

	case SELECTMODE_UNSELECTALL:
		SelectAllSubGroups(false);
		break;

	default:
		assert(false);
	}
}

//TuO[vIĂ邩
bool Group::IsSelectedSubGroup(int subGroupIndex) const
{
	return GetSubGroup(subGroupIndex)->IsSelected();
}

//IĂTuO[v𑊑ΓIɈړ
void Group::MoveSelectedSubGroups(const FPoint & rel)
{
	int subGroupCount=SubGroupCount();
	FOR(int i=0;i<subGroupCount;++i) {
		SubGroup *subGroup=GetSubGroup(i);
		if(subGroup->IsSelected()) {
			subGroup->MoveRel(rel);
		}
	}
}

//IĂTuO[v̐
int Group::SelectedSubGroupCount() const
{
	int count=0;
	FOR(int i=0;i<SubGroupCount();++i) {
		if(GetSubGroup(i)->IsSelected()) {
			++count;
		}
	}
	return count;
}

//TuO[v𕪉ĐIuWFNg̃|C^W߂
void Group::ExtractCurve(CurvePointers *curvePtrs,int part) const
{
	/*ʂPART_CURVEPART_SUBGROUPwł顗łǂ
	PART_CURVEɂPART_SELECTED PART_UNSELECTED PART_CLOSED PART_OPENED
	PART_SUBGROUPɂPART_SELECTED PART_UNSELECTEDwł
	TuO[v̐Ă邩ǂ͋ʂȂ	*/

	if(ContainCurve(part)) {
		int curveCount=CurveCount();
		FOR(int i=0;i<curveCount;++i) {
			const Curve *curve=GetCurve(i);
			if(!IsSelectedOpposite(part,curve->IsSelected())
				&&!IsClosedOpposite(part,curve->IsClosed()) ) {
				curvePtrs->AddCurve(curve);
			}
		}
	}

	if(ContainSubGroup(part)) {
		int subGroupCount=SubGroupCount();
		FOR(int i=0;i<subGroupCount;++i) {
			const SubGroup *subGroup=GetSubGroup(i);
			if(!IsSelectedOpposite(part,subGroup->IsSelected()) ) {
				subGroup->ExtractCurve(curvePtrs,part);
			}
		}
	}
}

//IĂƃTuO[v܂Ƃ߂Ĉ̃TuO[vɂ
void Group::PutInGroup()
{
	if(SelectedCurveCount()+SelectedSubGroupCount()<=1) {
		return;
	}

	std::auto_ptr<SubGroup> newSubGroup(new SubGroup);
	BZP_BAD_ALLOC_TEST(newSubGroup.get());

	int curveCount=CurveCount();
	FOR(int i=0;i<curveCount;) {
		Curve *curve=GetCurve(i);
		if(curve->IsSelected()) {
			newSubGroup->AddCurve(curve);
			m_m->curves.ReleaseItem(i);
			--curveCount;
		} else {
			++i;
		}
	}

	int subGroupCount=SubGroupCount();
	FOR(int i=0;i<subGroupCount;) {
		SubGroup *subGroup=GetSubGroup(i);
		if(subGroup->IsSelected()) {
			newSubGroup->AddSubGroup(subGroup);
			m_m->subGroups.ReleaseItem(i);
			--subGroupCount;
		} else {
			++i;
		}
	}

	m_m->subGroups.Add(newSubGroup.get());
	newSubGroup.release();
}

//IĂTuO[v𕪉
void Group::Ungroup()
{
	SelectAllCurves(false);

	SubGroupDeque subGroupStock;
	int subGroupCount=SubGroupCount();
	FOR(int i=0;i<subGroupCount;) {
		SubGroup *subGroup=GetSubGroup(i);
		if(subGroup->IsSelected()) {
			subGroupStock.Add(subGroup);
			m_m->subGroups.ReleaseItem(i);
			--subGroupCount;
		} else {
			++i;
		}
	}

	int subGroupStockCount=subGroupStock.Count();
	FOR(int i=0;i<subGroupStockCount;++i) {
		SubGroup *subGroup=subGroupStock.At(i);

		int curveCount=subGroup->CurveCount();
		FOR(int j=0;j<curveCount;++j) {
			AddCurve(subGroup->GetCurve(0));
			subGroup->ReleaseCurve(0);
		}

		int subGroupCount=subGroup->SubGroupCount();
		FOR(int j=0;j<subGroupCount;++j) {
			AddSubGroup(subGroup->GetSubGroup(0));
			subGroup->ReleaseSubGroup(0);
		}
	}
}

//TuO[vǉ
void Group::AddSubGroup(SubGroup *subGroup)
{
	m_m->subGroups.Add(subGroup);
}

//w肵TuO[vLȂ
void Group::ReleaseSubGroup(int subGroupIndex)
{
	m_m->subGroups.ReleaseItem(subGroupIndex);
}

//pڂ̏ԂW𓯂ɂ
void Group::CopyKnot(int destKnotIndex,int destCurveIndex,int srcKnotIndex,int srcCurveIndex)
{
	Curve *srcCurve=GetCurve(srcCurveIndex);
	GetCurve(destCurveIndex)->CopyKnot(destKnotIndex,srcKnotIndex,srcCurve);
}

//pڂ̏ԂW𓯂ɂ
void Group::CopyKnot(const Knot *srcKnot,int knotIndex,int curveIndex)
{
	GetCurve(curveIndex)->CopyKnot(srcKnot,knotIndex);
}

//pڂ̏ԂW𓯂ɂ
void Group::CopyKnotTo(Knot *destKnot,int knotIndex,int curveIndex) const
{
	GetCurve(curveIndex)->CopyKnotTo(destKnot,knotIndex);
}

//W̎pX̕ʂT
//WpFindResultɓĂmask͒Tʂ̎ޡ
bool Group::FindGroup(FindResult *findResult,const Graph *graph,int mask) const
{
	if(FindKnots(findResult,graph,mask,true)) return true;
	if(FindKnots(findResult,graph,mask,false)) return true;
	if(FindCurves(findResult,graph,mask,true)) return true;
	if(FindCurves(findResult,graph,mask,false)) return true;
	if(FindSubGroups(findResult,graph,mask,true)) return true;
	if(FindSubGroups(findResult,graph,mask,false)) return true;
	
	findResult->SetPart(PART_NONE);
	return false;		
}

//̌pڂ̒ʉߓ_ƕ_T
bool Group::FindKnots(FindResult *findResult,const Graph *graph,int mask,bool isSelected) const
{
	int maskKnot=MaskKnot(mask);
	if(!IsNoPart(maskKnot)) {
		int curveCount=CurveCount();
		FOR(int i=0;i<curveCount;++i) {
			const Curve *curve=GetCurve(i);
			if(curve->IsSelected()==isSelected
				&& curve->Find(findResult,graph,i,maskKnot)) {
				findResult->SetCurveIndex(i);
				return true;
			}
		}
	}
	return false;
}

//T
bool Group::FindCurves(FindResult *findResult,const Graph *graph,int mask,bool isSelected) const
{
	int maskCurve=MaskCurve(mask);
	if(!IsNoPart(maskCurve)) {
		int curveCount=CurveCount();
		FOR(int i=0;i<curveCount;++i) {
			const Curve *curve=GetCurve(i);
			if(curve->IsSelected()==isSelected
				&& curve->Find(findResult,graph,i,maskCurve)) {
				findResult->SetCurveIndex(i);
				return true;
			}
		}
	}
	return false;
}

//TuO[vT
bool Group::FindSubGroups(FindResult *findResult,const Graph *graph,int mask,bool isSelected) const
{
	int maskGroup=MaskSubGroup(mask);
	if(!IsNoPart(maskGroup)) {
		int subGroupCount=SubGroupCount();
		FOR(int i=0;i<subGroupCount;++i) {
			const SubGroup *subGroup=GetSubGroup(i);
			if(subGroup->IsSelected()==isSelected
				&& subGroup->FindSubGroup(findResult,graph,maskGroup)) {
				findResult->SetSubGroupIndex(i);
				return true;
			}
		}
	}
	return false;
}

//`I
void Group::SelectBoxGroup(SELECTMODE mode,const FRect &rect)
{
	if(mode==SELECTMODE_SINGLE) {
		SelectAllCurves(false);
		SelectAllSubGroups(false);
	}

	int curveCount=CurveCount();
	FOR(int i=0;i<curveCount;++i) {
		GetCurve(i)->SelectBox(mode,rect);
	}

	int subGroupCount=SubGroupCount();
	FOR(int i=0;i<subGroupCount;++i) {
		GetSubGroup(i)->SelectBoxSubGroup(mode,rect);
	}
}

//炩̃f[^\`ɕϊ
void Group::PathToDataGroup(PathTo *const converter,int part) const
{
	if(ContainCurve(part)||ContainKnot(part)) {
		int curveCount=CurveCount();
		FOR(int i=0;i<curveCount;++i) {
			GetCurve(i)->PathToData(converter,part);
		}
	}

	if(ContainSubGroup(part)) {
		int subGroupCount=SubGroupCount();
		FOR(int i=0;i<subGroupCount;++i) {
			GetSubGroup(i)->PathToDataSubGroup(converter,part);
		}
	}
}

//炩̃f[^\`pXɕϊ
void Group::DataToPathGroup(ToPath *const converter)
{
	Destroy();

	FOR(;;) {
		if(converter->IsCurve()) {
			std::auto_ptr<Curve> newCurve(new Curve);
			BZP_BAD_ALLOC_TEST(newCurve.get());

			newCurve->DataToPath(converter);
			m_m->curves.Add(newCurve.get());
			newCurve.release();

		} else if(converter->IsSubGroupOpened()) {
			std::auto_ptr<SubGroup> newSubGroup(new SubGroup);
			BZP_BAD_ALLOC_TEST(newSubGroup.get());

			newSubGroup->DataToPathSubGroup(converter);
			m_m->subGroups.Add(newSubGroup.get());
			newSubGroup.release();
		} else {
			break;
		}
	}
}

//pX̓eSč폜
void Group::Destroy() throw()
{
	m_m->curves.Destroy();
	m_m->subGroups.Destroy();
}

//pX𕡐
void Group::CopyGroup(const Group *srcGroup)
{
	int curveCount=srcGroup->CurveCount();
	FOR(int i=0;i<curveCount;++i) {
		std::auto_ptr<Curve> newCurve(new Curve);
		BZP_BAD_ALLOC_TEST(newCurve.get());
		newCurve->Copy(srcGroup->GetCurve(i));
		AddCurve(newCurve.get());
		newCurve.release();
	}

	int subGroupCount=srcGroup->SubGroupCount();
	FOR(int i=0;i<subGroupCount;++i) {
		std::auto_ptr<SubGroup> newSubGroup(new SubGroup);
		BZP_BAD_ALLOC_TEST(newSubGroup.get());
		newSubGroup->CopySubGroup(srcGroup->GetSubGroup(i));
		AddSubGroup(newSubGroup.get());
		newSubGroup.release();
	}
}

//pX͂ދ`
void Group::GetOuterRectGroup(FRect *rect,int part,double precision) const
{
	rect->Clear();
	if(ContainCurve(part)||ContainKnot(part)) {
		int curveCount=CurveCount();
		FOR(int i=0;i<curveCount;++i) {
			const Curve *curve=GetCurve(i);
			if(!IsSelectedOpposite(part,curve->IsSelected())) {
				FRect curveRect;
				curve->GetOuterRect(&curveRect,part,precision);
				rect->Union(curveRect);
			}
		}
	}

	if(ContainSubGroup(part)) {
		int subGroupCount=SubGroupCount();
		FOR(int i=0;i<subGroupCount;++i) {
			const SubGroup *subGroup=GetSubGroup(i);
			if(!IsSelectedOpposite(part,subGroup->IsSelected())) {
				FRect subGroupRect;
				subGroup->GetOuterRectGroup(&subGroupRect,PART_CURVE|PART_SUBGROUP,precision);
				rect->Union(subGroupRect);
			}
		}
	}
}

/*Wϊ
	x'=ax+by+c
	y'=dx+ey+f	̌vZčWς*/
void Group::Transform(const TRANSFORM &transform)
{
	int curveCount=CurveCount();
	FOR(int i=0;i<curveCount;++i) {
		GetCurve(i)->Transform(transform);
	}

	int subGroupCount=SubGroupCount();
	FOR(int i=0;i<subGroupCount;++i) {
		GetSubGroup(i)->Transform(transform);
	}
}

//P_ŐؒfĐ𕪂
void Group::CutCurvePoint(double t,int knotIndex,int curveIndex)
{
	SelectCurve(SELECTMODE_SINGLE,curveIndex);

	Curve *curve=GetCurve(curveIndex);
	if(curve->IsClosed()) {
		curve->CutClosedCurvePoint(t,knotIndex);
	} else {
		std::auto_ptr<Curve> newCurve(new Curve);
		BZP_BAD_ALLOC_TEST(newCurve.get());
		curve->CutOpenCurvePoint(newCurve.get(),t,knotIndex);
		AddCurve(newCurve.get());
		newCurve.release();
	}
}

//߂폜Đ𕪂
void Group::CutCurveSegment(int knotIndex,int curveIndex)
{
	SelectCurve(SELECTMODE_SINGLE,curveIndex);

	Curve *curve=GetCurve(curveIndex);
	if(curve->IsClosed()) {
		curve->CutClosedCurveSegment(knotIndex);
	} else {
		std::auto_ptr<Curve> newCurve(new Curve);
		BZP_BAD_ALLOC_TEST(newCurve.get());
		curve->CutOpenCurveSegment(newCurve.get(),knotIndex);
		AddCurve(newCurve.get());
		newCurve.release();
	}
}

BZP_NAMESPACE_END