#ifndef BZP_GROUP_H
#define BZP_GROUP_H

#include "site.h"

#ifdef __cplusplus
#include <memory>
#endif

/*************************************************

	̋ȐƃTuO[v܂Ƃ߂NX̒`

**************************************************/

BZP_NAMESPACE_START

#ifdef __cplusplus

struct tagTRANSFORMDATA
{
	/*x'=ax+by+c y'=dx+ey+f*/
	double a,b,c,d,e,f;
};
typedef tagTRANSFORMDATA TRANSFORMDATA;

class Curve;
class Graph;
class FindResult;
class SubGroup;
class FPoint;
class FRect;
class Knot;
class Data;
class CurvePointers;
class PathTo;
class ToPath;
class TRANSFORM;
enum SELECTMODE;
enum MOVECURVEMODE;
enum DELETEKNOTMODE;
enum INSERTKNOTMODE;

class GroupMember;

class Group
{
public:
	Group();
	virtual ~Group() throw();


	int CurveCount() const;
	int SubGroupCount() const;
	bool IsEmpty() const;

	void DrawGroup(const Graph *graph) const;
	bool FindGroup(FindResult *findResult,const Graph *graph,int mask) const;
	void Destroy() throw();
	void CopyGroup(const Group *srcGroup);

	int InsertKnot(INSERTKNOTMODE mode,double t,int knotIndex,int curveIndex);
	int AddEndKnot(const FPoint & point,int knotIndex,int curveIndex);
	int KnotCount(int curveIndex) const;
	void SelectKnot(SELECTMODE mode,int knotIndex,int curveIndex);
	bool IsSelectedKnot(int knotIndex,int curveIndex) const;
	void MoveSelectedKnots(const FPoint & rel);
	void DeleteKnot(DELETEKNOTMODE mode,int knotIndex,int curveIndex);
	void LinkKnot(bool isLinked,int knotIndex,int curveIndex);
	void CurveKnot(bool isCurved,int part,int knotIndex,int curveIndex);
	int SymbolizeKnotIndex(int knotIndex,int curveIndex) const;
	void CopyKnot(const Knot *srcKnot,int knotIndex,int curveIndex);
	void CopyKnotTo(Knot *destKnot,int knotIndex,int curveIndex) const;
	bool IsCurved(int part,int knotIndex,int curveIndex) const;
	int SelectedKnotCount() const;

	FPoint GetPoint(int part,int knotIndex,int curveIndex) const;
	void MovePoint(const FPoint & point,int part,int knotIndex,int curveIndex);
	void SetPoint(const FPoint & point,int part,int knotIndex,int curveIndex);
	void MoveWControls(const FPoint & point,bool grippedForward,int knotIndex,int curveIndex);

	int NewCurve(int curveIndex);
	void DeleteCurve(int curveIndex);
	void Chain(int *knotIndexA,int *curveIndexA,bool chainForward,int knotIndexB,int curveIndexB,bool deleteKnotB);
	void SelectCurve(SELECTMODE selectMode,int curveIndex);
	bool IsSelectedCurve(int curveIndex) const;
	void MoveCurve(const FPoint & newPoint,const FPoint & oldPoint,MOVECURVEMODE mode,int knotIndex,int curveIndex);
	bool IsSameCurve(int indexA,int indexB) const;
	bool IsClosedCurve(int curveIndex) const;
	void AddCurve(Curve *curve);
	void ReleaseCurve(int curveIndex);
	const Curve *GetCurve(int index) const;
	Curve *GetCurve(int index);
	void CloseCurve(bool isClosed,int curveIndex);
	int SelectedCurveCount() const;

	void SelectSubGroup(SELECTMODE mode,int subGroupIndex);
	bool IsSelectedSubGroup(int subGroupIndex) const;
	void DeleteSubGroup(int subGroupIndex);
	void AddSubGroup(SubGroup *subGroup);
	void ReleaseSubGroup(int subGroupIndex);
	SubGroup *GetSubGroup(int index);
	void MoveSelectedSubGroups(const FPoint & rel);
	int SelectedSubGroupCount() const;

	void ExtractCurve(CurvePointers *curvePtrs,int part) const;
	void PutInGroup();
	void Ungroup();
	const SubGroup *GetSubGroup(int index) const;
	void GetOuterRectGroup(FRect *rect,int part,double precision) const;
	void Transform(const TRANSFORM &transform);

	void CutCurvePoint(double t,int knotIndex,int curveIndex);
	void CutCurveSegment(int knotIndex,int curveIndex);
protected:
	void MoveSelectedCurves(const FPoint & rel);
	void SelectAllCurves(bool isSelected);
	void SelectAllSubGroups(bool isSelected);
	void CopyKnot(int destKnotIndex,int destCurveIndex
		,int srcKnotIndex,int srcCurveIndex);

	void PathToDataGroup(PathTo *const converter,int part) const;
	void DataToPathGroup(ToPath *const converter);
	void SelectBoxGroup(SELECTMODE mode,const FRect &rect);

	bool FindKnots(FindResult *findResult,const Graph *graph,int mask,bool isSelected) const;
	bool FindCurves(FindResult *findResult,const Graph *graph,int mask,bool isSelected) const;
	bool FindSubGroups(FindResult *findResult,const Graph *graph,int mask,bool isSelected) const;

private:
	const std::auto_ptr<GroupMember> m_m;
};

#endif /*__cplusplus*/

BZP_NAMESPACE_END

#endif
