#include "prefix.h"

#include <algorithm>
#include <new>

#include "knot.h"
#include "knotmode.h"
#include "graph.h"
#include "part.h"
#include "findresult.h"
#include "partmask.h"
#include "selectmode.h"
#include "converterbase.h"
#include "fpoint.h"
#include "partssize.h"
#include "rect.h"
#include "transform.h"

/*************************************************

	Ȑ̌pڂ̃NX̒`

**************************************************/

BZP_NAMESPACE_START

BZP_PARTMASK_USING

class KnotMember
{
public:
	KnotMember()
		:anchor(0,0)
		,forward(0,0)
		,back(0,0)
	{

	}
	FPoint anchor;	//ʉߓ_̍W
	FPoint forward;	//O̕_̍W
	FPoint back;	//̕_̍W
};

Knot::Knot()
	:m_m(new KnotMember)
	,m_mode(KNOT_LINKED|KNOT_SELECTED)
	,m_forwardLength(0)
	,m_backLength(0)
{
	BZP_BAD_ALLOC_TEST(m_m.get());
}

Knot::~Knot() throw()
{
	const FPoint zero(0,0);
	m_m->anchor=zero;
	m_m->forward=zero;
	m_m->back=zero;
}

//pڂ̏Ԃ̂܂ܕԂ
int Knot::GetMode() const
{
	return m_mode;
}

//ʉߓ_IĂ邩
bool Knot::IsSelected() const
{
	return (m_mode&KNOT_SELECTED)!=0;
}

//ʉߓ_I
//selectfalseȂIO
void Knot::Select(bool isSelected)
{
	if(isSelected) {
		m_mode|=KNOT_SELECTED;
	} else {
		m_mode&=~KNOT_SELECTED;
	}
}

//ȂĂ邩
bool Knot::IsLinked() const
{
	return (m_mode&KNOT_LINKED)!=0;
}

//ȂE܂
void Knot::Link(bool isLinked)
{
	if(isLinked) {
		m_mode|=KNOT_LINKED;
	} else {
		m_mode&=~KNOT_LINKED;
	}
}

//ȐɂȂpڂł邩
bool Knot::IsCurved(int part) const
{
	bool isCurved;
	
	switch(part) {
	case PART_ANCHOR:
	case PART_KNOT:
	case PART_CONTROL:
		isCurved=(m_mode&(KNOT_CURVED_FORWARD|KNOT_CURVED_BACK))!=0;
		break;

	case PART_FORWARDCONTROL:
		isCurved=(m_mode&KNOT_CURVED_FORWARD)!=0;
		break;

	case PART_BACKCONTROL:
		isCurved=(m_mode&KNOT_CURVED_BACK)!=0;
		break;

	default:
		assert(false);
	}
	return isCurved;
}

//pڂȂȐɂ
void Knot::Curve(bool isCurved,int part)
{
	switch(part) {
	case PART_ANCHOR:
	case PART_KNOT:
	case PART_CONTROL:
		if(isCurved) {
			m_mode|=KNOT_CURVED_FORWARD|KNOT_CURVED_BACK;
		} else {
			DeleteControl(true);
			DeleteControl(false);
		}
		break;

	case PART_FORWARDCONTROL:
		if(isCurved) {
			m_mode|=KNOT_CURVED_FORWARD;
		} else {
			DeleteControl(true);
		}
		break;

	case PART_BACKCONTROL:
		if(isCurved) {
			m_mode|=KNOT_CURVED_BACK;
		} else {
			DeleteControl(false);
		}
		break;

	default:
		assert(false);
	}
}

//ʉߓ_ƕ_̍W
FPoint Knot::GetPoint(int part) const
{
	FPoint point;
	switch(part) {
	case PART_ANCHOR:
		point=m_m->anchor;
		break;

	case PART_FORWARDCONTROL:
		point=m_m->forward;
		break;

	case PART_BACKCONTROL:
		point=m_m->back;
		break;

	default:
		assert(false);
	}
	return point;
}

//pڂ`
void Knot::DrawControlHandles(const Graph *graph,bool isPrevSelected,bool isNextSelected) const
{
	if(graph->DoesDrawKnot()) {
		if((isPrevSelected || IsSelected())
			&& IsCurved(PART_BACKCONTROL)) {
			graph->DrawControlHandle(m_m->anchor,m_m->back);
		}
		if((isNextSelected || IsSelected())
			&& IsCurved(PART_FORWARDCONTROL)) {
			if(graph->DoesDrawArrowControl()) {
				graph->DrawArrowControl(m_m->anchor,m_m->forward);
			} else {
				graph->DrawControlHandle(m_m->anchor,m_m->forward);
			}
		}
	}
}

//IĂȂāAǗpڂ`
void Knot::DrawSingleKnot(const Graph *graph) const
{
	if(graph->DoesDrawKnot()) {
		graph->DrawSingleKnot(m_m->anchor);
	}
}

//ʉߓ_`
void Knot::DrawAnchor(const Graph *graph) const
{
	DrawAnchor(graph,IsSelected());
}

//ʉߓ_`IEIw肷
void Knot::DrawAnchor(const Graph *graph,bool isSelected) const
{
	if(graph->DoesDrawKnot()) {
		graph->DrawAnchor(m_m->anchor,isSelected);
	}
}

//pڂړ
void Knot::MovePoint(const FPoint & point,int part)
{
	switch(part) {
	case PART_ANCHOR:
	case PART_KNOT:
		//ʉߓ_܂Ōpڂړ
		MoveRel(point-m_m->anchor);
		break;

	//Е̕_܂ňړꍇ
	case PART_FORWARDCONTROL:
		MoveForwardControl(point);
		break;

	case PART_BACKCONTROL:
		MoveBackControl(point);
		break;

	default:
		assert(false);
	}
}

//O̕_ړ
void Knot::MoveForwardControl(const FPoint & point)
{
	FPoint relative=point-m_m->anchor;
	m_forwardLength=point.Distance(m_m->anchor);
	m_m->forward=point;

	if(IsLinked()) {
		if(m_forwardLength>0.0) {
			FPoint backRelative=relative*m_backLength/m_forwardLength;
			m_m->back=m_m->anchor-backRelative;
		}
	}
	m_mode|=KNOT_CURVED_FORWARD;
}

//̕_ړ
void Knot::MoveBackControl(const FPoint & point)
{
	FPoint relative=point-m_m->anchor;
	m_backLength=point.Distance(m_m->anchor);
	m_m->back=point;

	if(IsLinked()) {
		if(m_backLength>0.0) {
			FPoint forwardRelative=relative*m_forwardLength/m_backLength;
			m_m->forward=m_m->anchor-forwardRelative;
		}
	}
	m_mode|=KNOT_CURVED_BACK;
}

//𗼑֐L΂
void Knot::MoveWControls(const FPoint & point,bool grippedForward)
{
	m_mode|=KNOT_LINKED|KNOT_CURVED_FORWARD|KNOT_CURVED_BACK;
	FPoint opposite=m_m->anchor.OppositePoint(point);
	if(grippedForward) {
		m_m->forward=point;
		m_m->back=opposite;
	} else {
		m_m->back=point;
		m_m->forward=opposite;
	}
	CalcControlLength();
}

//pڂ̈ʒupointRelړ
void Knot::MoveRel(const FPoint & pointRel)
{
	m_m->anchor+=pointRel;
	m_m->forward+=pointRel;
	m_m->back+=pointRel;
}

//̏ԂɊւ炸WύX
void Knot::SetPoint(const FPoint & point,int part)
{
	switch(part) {
	case PART_ANCHOR:
		m_m->anchor=point;
		break;

	case PART_FORWARDCONTROL:
		m_m->forward=point;
		break;

	case PART_BACKCONTROL:
		m_m->back=point;
		break;

	default:
		assert(false);
	}
	CalcControlLength();
}

//_폜
void Knot::DeleteControl(bool deleteForward)
{
	if(deleteForward) {
		m_mode&=~(KNOT_LINKED|KNOT_CURVED_FORWARD);
		m_m->forward=m_m->anchor;
		m_forwardLength=0.0;
	} else {
		m_mode&=~(KNOT_LINKED|KNOT_CURVED_BACK);
		m_m->back=m_m->anchor;
		m_backLength=0.0;
	}
}

//pڂ̏Ԃ^
void Knot::Copy(const Knot *srcKnot)
{
	m_m->anchor=srcKnot->GetPoint(PART_ANCHOR);
	m_m->forward=srcKnot->GetPoint(PART_FORWARDCONTROL);
	m_m->back=srcKnot->GetPoint(PART_BACKCONTROL);
	m_mode=srcKnot->GetMode();
	CalcControlLength();
}

//O̕_ւ
void Knot::Reverse()
{
	FPoint point=m_m->forward;
	m_m->forward=m_m->back;
	m_m->back=point;
	CalcControlLength();
	int mode=m_mode&~(KNOT_CURVED_FORWARD|KNOT_CURVED_BACK);
	if(IsCurved(PART_FORWARDCONTROL)) {
		mode|=KNOT_CURVED_BACK;
	}
	if(IsCurved(PART_BACKCONTROL)) {
		mode|=KNOT_CURVED_FORWARD;
	}
	m_mode=mode;
}

//T
//anchorPart͒ʉߓ_wĂꍇ̏ڍׂȕʡJOINTANCHORFIRSTANCHORȂǡ
//isPrevSelected isNextSelected͑O̒ʉߓ_IĂ邩
bool Knot::Find(FindResult *findResult,const Graph *graph
	,int mask,int anchorPart
	,bool isPrevSelected,bool isNextSelected) const
{
	FPoint point=findResult->GetPoint();

	int maskAnchor=MaskAnchor(mask);
	if(!IsNoPart(maskAnchor) 
		&& !IsSelectedOpposite(mask,IsSelected())
		&& !IsCurvedOpposite(mask,IsCurved(PART_ANCHOR))
		&& !IsLinkedOpposite(mask,IsLinked()) ) {

		int anchorArea=graph->GetPartsSize(PARTSSIZE_ANCHORAREA);
		if(point.IsInArea(m_m->anchor,anchorArea)) {
			findResult->SetPart(anchorPart);
			findResult->SetSelect(IsSelected());
			findResult->SetCurve(IsCurved(PART_ANCHOR));
			findResult->SetLink(IsLinked());
			return true;
		}
	}

	PARTSSIZE forwardParts;
	if(!graph->DoesDrawArrowControl()) {
		forwardParts=PARTSSIZE_CONTROLAREA;
	} else {
		forwardParts=PARTSSIZE_ARROWAREA;
	}
	int controlArea=graph->GetPartsSize(forwardParts);
	int maskForwardControl=MaskForwardControl(mask);
	if(!IsNoPart(maskForwardControl) 
		&& (isNextSelected||IsSelected())
		&& IsCurved(PART_FORWARDCONTROL)) {

		if(point.IsInArea(m_m->forward,controlArea)) {
			findResult->SetPart(PART_FORWARDCONTROL);
			findResult->SetCurve(IsCurved(PART_FORWARDCONTROL));
			findResult->SetLink(IsLinked());
			return true;
		}
	}

	int maskBackControl=MaskBackControl(mask);
	if(!IsNoPart(maskBackControl)
		&& (isPrevSelected||IsSelected())
		&& IsCurved(PART_BACKCONTROL)) {

		if(point.IsInArea(m_m->back,controlArea)) {
			findResult->SetPart(PART_BACKCONTROL);
			findResult->SetCurve(IsCurved(PART_BACKCONTROL));
			findResult->SetLink(IsLinked());
			return true;
		}
	}
	return false;
}

//ʉߓ_`̒ɓĂ邩
bool Knot::IsAnchorInRect(const FRect &rect) const
{
	return m_m->anchor.IsInRect(rect);
}

//`I
void Knot::SelectBox(SELECTMODE mode,const FRect &rect)
{
	if(!m_m->anchor.IsInRect(rect)) {
		return;
	}

	bool isSelected=IsSelected();
	switch(mode) {
	case SELECTMODE_ADD:
	case SELECTMODE_SINGLE:
		isSelected=true;
		break;

	case SELECTMODE_UNSELECT:
		isSelected=false;
		break;

	case SELECTMODE_TOGGLE:
		isSelected=!isSelected;
		break;

	default:
		assert(false);
	}
	Select(isSelected);
}

//炩̃f[^\`ɕϊ
void Knot::PathToData(PathTo *const converter) const
{
	assert(converter!=NULL);
	converter->WriteKnot(m_m->back,m_m->anchor,m_m->forward,m_mode);
}

//炩̃f[^\`pXɕϊ
void Knot::DataToPath(ToPath *const converter)
{
	assert(converter!=NULL);

	converter->ReadKnot(&m_m->back,&m_m->anchor,&m_m->forward,&m_mode);

	CalcControlLength();
}

//̒vZ
void Knot::CalcControlLength()
{
	m_forwardLength=m_m->anchor.Distance(m_m->forward);
	m_backLength=m_m->anchor.Distance(m_m->back);
}

#ifdef min
# undef min
#endif

inline const double MIN3(double a,double b,double c)
{
	return std::min(std::min(a,b),c);
}

#ifdef max
# undef max
#endif

inline const double MAX3(double a,double b,double c)
{
	return std::max(std::max(a,b),c);
}

//_܂ރpX͂ދ`
void Knot::GetOuterRect(FRect *rect,double precision) const
{
	double left=MIN3(m_m->anchor.x,m_m->forward.x,m_m->back.x);
	double right=MAX3(m_m->anchor.x,m_m->forward.x,m_m->back.x)+1;
	if(left==right) right+=precision;
	double top=MIN3(m_m->anchor.y,m_m->forward.y,m_m->back.y);
	double bottom=MAX3(m_m->anchor.y,m_m->forward.y,m_m->back.y)+1;
	if(top==bottom) bottom+=precision;
	rect->Set(left,top,right,bottom);
}

inline const FPoint SubTransform(const FPoint &point,const TRANSFORM &transform)
{
	FPoint result;

	result.x=transform.x_x*point.x
		+transform.x_y*point.y
		+transform.x_c;

	result.y=transform.y_x*point.x
		+transform.y_y*point.y
		+transform.y_c;

	return result;
}

//Wϊ
void Knot::Transform(const TRANSFORM &transform)
{
	m_m->anchor=SubTransform(m_m->anchor,transform);
	m_m->forward=SubTransform(m_m->forward,transform);
	m_m->back=SubTransform(m_m->back,transform);
	CalcControlLength();
}



BZP_NAMESPACE_END
