#include "prefix.h"

#include <windows.h>
#include <memory>
#include <new>

#include "path.h"
#include "group.h"
#include "rect.h"
#include "converterbase.h"
#include "transform.h"
#include "graph.h"
#include "exception.h"
#include "error.h"

/*************************************************

	pX̃NX
	GrouppĂGroup̊֐g

**************************************************/

BZP_NAMESPACE_START

Path::Path()
	:m_boundRect(new Rect)
{
	BZP_BAD_ALLOC_TEST(m_boundRect.get());

#ifndef NDEBUG
	m_boundRect->Clear();
#endif
}

Path::~Path() throw()
{
}

Path::Path(const Path &source)
{
	Copy(&source);
}

const Path &Path::operator =(const Path &source)
{
	Copy(&source);
	return source;
}

void Path::Copy(const Path *source)
{
	Destroy();
	Rect sourceRect;
	source->GetBoundRect(&sourceRect);
	SetBoundRect(sourceRect);
	Append(source);
}

//W`̐ݒ
void Path::SetBoundRect(const Rect &rect)
{
	BZP_PARAMETER_ERROR(!rect.IsEmpty(),BZPERROR_EMPTY_RECT);
	m_boundRect->Set(rect);
}

void Path::SetBoundRect(const RECT &rect)
{
	BZP_PARAMETER_ERROR(rect.right-rect.left!=0
		&& rect.bottom-rect.top!=0,BZPERROR_EMPTY_RECT);
	m_boundRect->Set(rect);
}

//W`𓾂
void Path::GetBoundRect(RECT *rect) const
{
	BZP_PARAMETER_ERROR(rect!=NULL,BZPERROR_NULL_ARGUMENT);
	*rect=static_cast<RECT>(*m_boundRect);
}

//f[^\ɕϊ
void Path::PathToData(PathTo *const converter,int part) const
{
	assert(converter!=NULL);

	converter->WriteStartPath(*m_boundRect);
	PathToDataGroup(converter,part);
	converter->WriteEndPath();
}

//f[^\pXɕϊ
void Path::DataToPath(ToPath *const converter)
{
	assert(converter!=NULL);

	converter->ReadStartPath(m_boundRect.get());
	DataToPathGroup(converter);
	converter->ReadEndPath();
}

//`I
void Path::SelectBox(SELECTMODE mode,const FRect &rect)
{
	FRect normalRect(rect);
	normalRect.Normalize();

	SelectBoxGroup(mode,normalRect);
}

//ʂ̒T
bool Path::Find(FindResult *findResult,const Graph *graph,int mask) const
{
	return FindGroup(findResult,graph,mask);
}

//`
void Path::Draw(const Graph *graph) const
{
	if(graph->DoesDraw()) {
		DrawGroup(graph);
	}
}

//pXǉ
void Path::Append(const Path *path)
{
	CopyGroup(path);
}

//SĂ̐ƃTuO[vI
void Path::SelectAll(bool isSelected)
{
	SelectAllCurves(isSelected);
	SelectAllSubGroups(isSelected);
}

//sړ
void Path::Move(double x,double y)
{
	TRANSFORM expression
	(1,0,x			//newx=oldx+argx
	,0,1,y);			//newy=oldy+argy
	Transform(expression);
}

//sړ
void Path::Move(FPoint point)
{
	Move(point.x,point.y);
}

//㉺]
void Path::VReverse()
{
	TRANSFORM expression
		(1,0,0							//x'=x
		,0,-1,m_boundRect->bottom);	//y'=bottom-y
	Transform(expression);
}

//]
void Path::HReverse()
{
	TRANSFORM expression
		(-1,0,m_boundRect->right	//x'=right-x
		,0,1,0);					//y'=y
	Transform(expression);
}

/*pX͂ދ`
partPART_CURVE PART_SUBGROUP PART_SELECTED PART_UNSELECTED
precision͐xprecisionȉ͐mł͂Ȃ*/
void Path::GetOuterRect(RECT *rect,int part,double precision) const
{
	BZP_PARAMETER_ERROR(rect!=NULL,BZPERROR_NULL_ARGUMENT);
	BZP_PARAMETER_ERROR(precision>0,BZPERROR_ZERO_PRECISION);

	GetOuterRect(reinterpret_cast<Rect *>(rect),part,precision);
}

void Path::GetOuterRect(Rect *rect,int part,double precision) const
{
	assert(rect!=NULL);

	FRect frect;
	GetOuterRect(&frect,part,precision);
	rect->SetFRect(frect);
}

void Path::GetOuterRect(FRect *rect,int part,double precision) const
{
	assert(rect!=NULL);
	GetOuterRectGroup(rect,part,precision);
}

BZP_NAMESPACE_END

