#ifndef BZP_PTRDEQUE_H
#define BZP_PTRDEQUE_H

#include "site.h"

#ifdef __cplusplus
#include <deque>
#include <algorithm>

#include "index.h"
#endif

/*************************************************

	|C^z̃ev[gNX̒`

**************************************************/

BZP_NAMESPACE_START

#ifdef __cplusplus

typedef std::deque<void*> PtrDequeBase;

template <class T>
 class PtrDeque
{
public:
	virtual ~PtrDeque() throw()
	{
		Destroy();
	}
	/*vf̐*/
	int Count() const
	{
		return v.size();
	}
	/*ł邩*/
	bool IsEmpty() const
	{
		return v.empty();
	}

	/*indexԖڂ̗vf̃AhX*/
	T *operator [](int index)
	{
		return At(index);
	}

	/*indexԖڂ̗vf̃AhX*/
	const T*operator [](int index) const
	{
		return At(index);
	}

	/*vfSč폜*/
	void Destroy() throw();

	/*Ō̗vfƂĒǉ*/
	int Add(T *item) {
		v.push_back(static_cast<void*>(item));
		return Count()-1;
	}

	/*w肵ԍ̗vfԂ*/
	T *At(int index)
	{
		index=NormalizeIndex(index);
		assert(0<=index&&index<Count());
		return reinterpret_cast<T*>(v[index]);
	}

	/*indexԖڂ̗vf̃AhX*/
	const T *At(int index) const
	{
		index=NormalizeIndex(index);
		assert(0<=index&&index<Count());
		return reinterpret_cast<T*>(v[index]);
	}

	/*w肵ԍ̗vf폜*/
	void Remove(int index);

	/*w肵ԍ̗vf̎ɒǉ*/
	int InsertAfter(T *item,int index);
	
	/*w肵ԍ̂悤̎ɕ̗vfǉ*/
	void InsertAfter(const PtrDeque<T> &items,int index);

	/*tɂ*/
	void Reverse()
	{
		std::reverse(v.begin(),v.end());
	}

	/*IuWFNg͍폜vf*/
	T *ReleaseItem(int index)
	{
		T *obj=At(index);
		v.erase(v.begin()+index);
		return obj;
	}

	/*SĂ̗vfLȂ*/
	void ReleaseAll()
	{
		v.clear();
	}

	/*indexԍɂ*/
	int NormalizeIndex(int index) const;

	/*indexłINDEX_LASTȂǂ̓ʂȒlɂ*/
	int SymbolizeIndex(int index) const;

	/*index擪ɂȂ悤ɗvf̏Ԃ炷*/
	void Rotate(int index);
protected:
	int InsertIndex(int index) const;

	PtrDequeBase::iterator begin()
	{
		return v.begin();
	}

	PtrDequeBase::const_iterator begin() const
	{
		return v.begin();
	}

	PtrDequeBase::iterator end()
	{
		return v.end();
	}

	PtrDequeBase::const_iterator end() const
	{
		return v.end();
	}

private:
	PtrDequeBase v;
};

template <class T>
 int PtrDeque<T>::NormalizeIndex(int index) const
{
	int normal=index;
	switch(index) {
	case INDEX_LAST:
		normal=Count()-1;
		break;
	
	case INDEX_SECOND:
		if(Count()>=2) {
			normal=1;
		} else {
			normal=INDEX_NONE;
		}
		break;

	case INDEX_SECONDLAST:
		if(Count()>=2) {
			normal=Count()-2;
		} else {
			normal=INDEX_NONE;
		}
		break;

	case INDEX_SINGLE:
		assert(Count()==1);
		normal=INDEX_FIRST;
		break;
	}
	return normal;
}

template <class T>
 int PtrDeque<T>::SymbolizeIndex(int index) const
{
	int symbol=index;
	if((index>=0 && index>Count()-1)
		||(Count()==1 && (index==INDEX_SECOND||index==INDEX_SECONDLAST))) {
		symbol=INDEX_NONE;
	} else if(Count()==1 && (index==INDEX_FIRST||index==INDEX_LAST)) {
		symbol=INDEX_SINGLE;
	} else if(Count()>1 && index==INDEX_FIRST) {
		/*̂܂܂̒lԂ*/
	} else if(index==Count()-1) {
		symbol=INDEX_LAST;
	} else if(index==1) {
		symbol=INDEX_SECOND;
	} else if(index==Count()-2) {
		symbol=INDEX_SECONDLAST;
	}
	return symbol;
}

template <class T>
 void PtrDeque<T>::Destroy() throw()
{
	int count=Count();
	FOR(int i=0;i<count;++i) {
		delete reinterpret_cast<T*>(v[i]);
/*vf̃|C^wIuWFNg͎Iɍ폜*/
	}
	v.clear();
}

template <class T>
 void PtrDeque<T>::Remove(int index)
{
	index=NormalizeIndex(index);
	assert(0<=index&&index<Count());
	delete reinterpret_cast<T*>(v[index]);
	v.erase(v.begin()+index);
}

template <class T>
 int PtrDeque<T>::InsertIndex(int index) const
{
	int insertIndex;
	if(index==INDEX_APPEND||index==INDEX_LAST) {
		insertIndex=Count();
	} else if(index==INDEX_PRECEDE) {
		insertIndex=0;
	} else {
		insertIndex=NormalizeIndex(index);
		assert(0<=insertIndex&&insertIndex<Count());
		++insertIndex;
	}
	return insertIndex;
}

template <class T>
 int PtrDeque<T>::InsertAfter(T *item,int index=INDEX_APPEND)
{
	assert(item!=NULL);

	int insertIndex=InsertIndex(index);

	v.insert(v.begin()+insertIndex,static_cast<void*>(item));
	return insertIndex;
}

template <class T>
 void PtrDeque<T>::InsertAfter(const PtrDeque<T> &items,int index)
{
	int insertIndex=InsertIndex(index);

	v.insert(v.begin()+insertIndex,items.begin(),items.end());
}

template <class T>
 void PtrDeque<T>::Rotate(int index)
{
	int normalIndex=NormalizeIndex(index);
	assert(normalIndex!=INDEX_NONE);

	std::rotate(v.begin(),v.begin()+normalIndex,v.end());
}

#endif /*__cplusplus*/

BZP_NAMESPACE_END

#endif
