#ifndef BZP_RECT_H
#define BZP_RECT_H

#include "site.h"

#ifdef __cplusplus
#include "fpoint.h"
#include <windows.h>
#endif

/*************************************************

	`̃NX

**************************************************/

BZP_NAMESPACE_START

/*_̍W̋`̍\*/
struct tagFRECT
{
	double left,top,right,bottom;
};
typedef tagFRECT FRECT;

#ifdef __cplusplus

/*`̌vZev[g
	T@left,top,right,bottom̒l킷^
	BaseT `̍\
	PointT W̌^*/
template <class T,class BaseT,class PointT>
 class RectBase : public BaseT
{
public:
	typedef RectBase<T,BaseT,PointT> THISTYPE;
	RectBase()
	{
	}

	/*`ݒ*/
	void Set(T argLeft,T argTop,T argRight,T argBottom);

	/*Q_w肵ċ`ݒ*/
	void Set(const PointT &cornerA,const PointT &cornerB);

	/*\̂w肵Đݒ*/
	void Set(const BaseT &rect);

	/*left topɂȂ悤ɒl*/
	void Normalize();

	/*S0ɂ*/
	void Clear()
	{
		left=top=right=bottom=0;
	}

	/*镔̍ŏ̋`ݒ肷*/
	bool Intersect(const THISTYPE &rect);

	/*܂ދ`*/
	void Union(const THISTYPE &rect);

	/*`̉*/
	T Width() const
	{
		return Abs(right-left);
	}

	/*`̏c*/
	T Height() const
	{
		return Abs(bottom-top);
	}

	/*W*/
	PointT TopLeft() const
	{
		return PointT(left,top);
	}

	/*EW*/
	PointT TopRight() const
	{
		return PointT(right,top);
	}

	/*W*/
	PointT BottomLeft() const
	{
		return PointT(left,bottom);
	}

	/*EW*/
	PointT BottomRight() const
	{
		return PointT(right,bottom);
	}

	/*center𒆐Sɂsizȇ傫̋`*/
	void SizedRect(const PointT &center,int size);

	/*ړ*/
	void Offset(const PointT &offset);

	/*L*/
	void Inflate(const PointT &size)
	{
		Inflate(size.x,size.y);
	}
	void Inflate(double size)
	{
		Inflate(size,size);
	}
	void Inflate(double sizex,double sizey);

	/*`Ă邩*/
	bool IsInRect(const THISTYPE &rect) const;

	/*W`̒ɂ邩*/
	bool IsPointInRect(const PointT &point) const
	{
		return point.IsInRect(*this);
	}

	/*̋`*/
	bool IsEmpty() const
	{
		return right-left==0 || bottom-top==0;
	}

protected:
	const T Abs(T value) const
	{
		return (value>=0)?value:-value;
	}
};

/*S_w肵Đݒ*/
template <class T,class BaseT,class PointT>
 void RectBase<T,BaseT,PointT>::Set(T argLeft,T argTop,T argRight,T argBottom)
{
	left=argLeft;
	top=argTop;
	right=argRight;
	bottom=argBottom;
}

/*Q_w肵ċ`ݒ*/
template <class T,class BaseT,class PointT>
 void RectBase<T,BaseT,PointT>::Set(const PointT &cornerA,const PointT &cornerB)
{
	left=cornerA.x;
	top=cornerA.y;
	right=cornerB.x;
	bottom=cornerB.y;
}

/*\̂w肵Đݒ*/
template <class T,class BaseT,class PointT>
 void RectBase<T,BaseT,PointT>::Set(const BaseT &rect)
{
	left=rect.left;
	top=rect.top;
	right=rect.right;
	bottom=rect.bottom;
}

/*left topɂȂ悤ɒl*/
template <class T,class BaseT,class PointT>
 void RectBase<T,BaseT,PointT>::Normalize()
{
	if(left>right) {
		T tmp=left;
		left=right;
		right=tmp;
	}
	if(top>bottom) {
		T tmp=top;
		top=bottom;
		bottom=tmp;
	}
}

/*ŏ̋`ݒ肷*/
template <class T,class BaseT,class PointT>
 bool RectBase<T,BaseT,PointT>::Intersect(const THISTYPE &rect)
{
	Normalize();

	THISTYPE b;
	b=rect;
	b.Normalize();

	if(left<b.left) left=b.left;
	if(right>b.right) right=b.right;
	if(top<b.top) top=b.top;
	if(bottom>b.bottom) bottom=b.bottom;

	if(left>right || top>bottom) {
		Clear();
		return false;
	}
	return true;
}

/*܂ލŏ̋`*/
template <class T,class BaseT,class PointT>
 void RectBase<T,BaseT,PointT>::Union(const THISTYPE &rect)
{
	if(rect.IsEmpty()) {
		return;
	}

	if(IsEmpty()) {
		*this=rect;
		return;
	}

	THISTYPE b=rect;
	b.Normalize();
	Normalize();

	if(b.left<left) {
		left=b.left;
	}
	if(b.top<top) {
		top=b.top;
	}
	if(b.right>right) {
		right=b.right;
	}
	if(b.bottom>bottom) {
		bottom=b.bottom;
	}
}

/*center𒆐Sɂsizȇ傫̋`*/
template <class T,class BaseT,class PointT>
 void RectBase<T,BaseT,PointT>::SizedRect(const PointT &center,int size)
{
	left=center.x-size/2;
	top=center.y-size/2;
	right=center.x+(size-size/2);
	bottom=center.y+(size-size/2);
}

/*ړ*/
template <class T,class BaseT,class PointT>
 void RectBase<T,BaseT,PointT>::Offset(const PointT &offset)
{
	left+=offset.x;
	right+=offset.x;
	top+=offset.y;
	bottom+=offset.y;
}

/*`Ă邩*/
template <class T,class BaseT,class PointT>
 bool RectBase<T,BaseT,PointT>::IsInRect(const THISTYPE &rect) const
{
	THISTYPE rectA=*this;
	THISTYPE rectB=rect;

	return (rectA.left<rectB.right)
		&&(rectB.left<rectA.right)
		&&(rectA.top<rectB.bottom)
		&&(rectB.top<rectA.bottom);
}

/*L*/
template <class T,class BaseT,class PointT>
 void RectBase<T,BaseT,PointT>::Inflate(double sizex,double sizey)
{
	int x=sizex;
	if(left>right) {
		x=-x;
	}
	left-=x;
	right+=x;
	
	int y=sizey;
	if(top>bottom) {
		y=-y;
	}
	top-=y;
	bottom+=y;
}

/**************************************************/

/*_̋`NX*/
class FRect : public RectBase<double,FRECT,FPoint>
{
public:
	FRect()
	{
	}

	~FRect() throw()
	{
	}

	FRect(const FRect &src)
	{
		Set(src);
	}

	FRect(double left,double top,double right,double bottom)
	{
		Set(left,top,right,bottom);
	}
};

/*̋`NX*/
class Rect : public RectBase<int,RECT,POINT>
{
public:
	Rect()
	{
	}

	~Rect() throw()
	{
	}

	Rect(const Rect &src)
	{
		Set(src);
	}

	Rect(int left,int top,int right,int bottom)
	{
		Set(left,top,right,bottom);
	}

	void SetFRect(const FRect &frect)
	{
		FRect copyRect(frect);
		copyRect.Normalize();
		Set(static_cast<int>(floor(copyRect.left))
			,static_cast<int>(floor(copyRect.top))
			,static_cast<int>(ceil(copyRect.right))
			,static_cast<int>(ceil(copyRect.bottom)) );
	}
};

#endif /*__cplusplus*/

BZP_NAMESPACE_END
#endif
