#include "prefix.h"

#include <assert.h>

#include "redraw.h"
#include "path.h"
#include "error.h"
#include "graph.h"
#include "exception.h"
#include "toolholder.h"

/*************************************************

	pX̍ĕ\ǗNX

**************************************************/

BZP_NAMESPACE_START

/*
pXύXĕύX\Ȃꍇ
	PrepareBefore		ύXO̕`̏
	DrawBefore ܂ SkipBefore	ύXÕpX̕`
	PrepareAfter		ύX̕`̏
	DrawAfter ܂SkipAfter	ύX̃pX̕`
	End			ύXOƕύX̍ʂɕ\
̏ԂɌĂяoEndȊO͏ȗł
pXύXODrawBefore܂ŌĂяoĂȂ΂Ȃ

pXŜ\ꍇ
	PrepareDraw		`̏
	Draw ܂ SkipDraw	pX̕`
	End		pXʂɕ\
̏ԂɌĂяoEndȊO͏ȗł

ԂGraphIuWFNgGraphIuWFNgofoCXReLXg
΂ĕ`΂̏ɉwȂǂpXƈꏏɕ`Ƃł

EndĂ΂Ȃƕ\͒~

qɂƈԊOLɂȂĕ\
̃pX\ĂŒɂ͕\͂łȂ

Path::Draw͓RedrawNXgĂ̂
ʂɕ\ꍇ͂g
*/

enum PHASE	/**/
{
	PHASE_BEGIN,	/*܂ĂȂ*/
	PHASE_PREPARE_BEFORE,	/*ύXO̕`揀*/
	PHASE_DRAW_BEFORE,	/*ύXÕpX̕`*/
	PHASE_PREPARE_AFTER,	/*ύX̕`揀*/
	PHASE_DRAW_AFTER,	/*ύX̃pX̕`*/
	PHASE_PREPARE_DRAW,	/*SĂ̕̕`揀*/
	PHASE_DRAW,	/*pX̑SĂ̕̕`*/
	PHASE_END	/*\*/
};

Redraw::Redraw()
	:m_phase(PHASE_BEGIN)
	,m_path(NULL)
	,m_graph(NULL)
	,m_tool(NULL)
	,m_canRedraw(true)
{

}

Redraw::Redraw(Path *path,Graph *graph)
	:m_phase(PHASE_BEGIN)
	,m_path(path)
	,m_graph(graph)
	,m_tool(NULL)
	,m_canRedraw(true)
{

}

Redraw::~Redraw() throw()
{
	if(m_phase!=PHASE_BEGIN 
		&& m_phase!=PHASE_END 
		&& m_canRedraw) {
		m_graph->CancelDraw();
	}
}

//\pXݒ
void Redraw::SetPath(Path *path)
{
	BZP_PARAMETER_ERROR(m_phase==PHASE_BEGIN,BZPERROR_CANT_SET_PARAMETER);
	m_path=path;
}

//\GraphIuWFNgݒ
void Redraw::SetGraph(Graph *graph)
{
	BZP_PARAMETER_ERROR(m_phase==PHASE_BEGIN,BZPERROR_CANT_SET_PARAMETER);
	m_graph=graph;
}

void Redraw::SetTool(ToolHolder *tool)
{
	m_tool=tool;
}

//GraphIuWFNg𓾂
const Graph *Redraw::GetGraph() const
{
	assert(m_path!=NULL);
	BZP_PARAMETER_ERROR(m_phase!=PHASE_BEGIN && m_phase!=PHASE_END
		,BZPERROR_CALLED_MISTAKE_ORDER);
	if(!m_canRedraw) {
		return NULL;
	}
	return m_graph;
}

//ύXÕpX̕`̏
void Redraw::PrepareBefore()
{
	BZP_PARAMETER_ERROR(m_phase==PHASE_BEGIN
		,BZPERROR_CALLED_MISTAKE_ORDER);
	BZP_PARAMETER_ERROR(m_graph!=NULL,BZPERROR_NO_GRAPH_OBJECT);

	if(m_graph->IsInUse()) {
		m_canRedraw=false;
	} else {
		m_graph->PrepareBeforeBitmap();
	}
	m_phase=PHASE_PREPARE_BEFORE;
}

//ύXÕpX̕`
void Redraw::DrawBefore()
{
	if(m_phase<PHASE_PREPARE_BEFORE) {
		PrepareBefore();
	}
	BZP_PARAMETER_ERROR(m_phase==PHASE_PREPARE_BEFORE
		,BZPERROR_CALLED_MISTAKE_ORDER);

	if(m_canRedraw) {
		if(m_tool==NULL) {
			m_path->Draw(m_graph);
		} else {
			m_tool->Draw(m_path,m_graph);
		}
	}
	m_phase=PHASE_DRAW_BEFORE;
}

//ύXÕpX\Ȃ
void Redraw::SkipBefore()
{
	if(m_phase<PHASE_PREPARE_BEFORE) {
		PrepareBefore();
	}
	BZP_PARAMETER_ERROR(m_phase==PHASE_PREPARE_BEFORE
		,BZPERROR_CALLED_MISTAKE_ORDER);
	m_phase=PHASE_DRAW_BEFORE;
}

//ύX̃pX̕`̏
void Redraw::PrepareAfter()
{
	if(m_phase<PHASE_DRAW_BEFORE) {
		DrawBefore();
	}
	BZP_PARAMETER_ERROR(m_phase==PHASE_DRAW_BEFORE
		,BZPERROR_CALLED_MISTAKE_ORDER);

	if(m_canRedraw) {
		m_graph->PrepareAfterBitmap();
	}
	m_phase=PHASE_PREPARE_AFTER;
}

//ύX̃pX`悷
void Redraw::DrawAfter()
{
	if(m_phase<PHASE_PREPARE_AFTER) {
		PrepareAfter();
	}
	BZP_PARAMETER_ERROR(m_phase==PHASE_PREPARE_AFTER
		,BZPERROR_CALLED_MISTAKE_ORDER);

	if(m_canRedraw) {
		if(m_tool==NULL) {
			m_path->Draw(m_graph);
		} else {
			m_tool->Draw(m_path,m_graph);
		}
	}
	m_phase=PHASE_DRAW_AFTER;
}

//ύX̃pX`悵Ȃ
void Redraw::SkipAfter()
{
	if(m_phase<PHASE_PREPARE_AFTER) {
		PrepareAfter();
	}
	BZP_PARAMETER_ERROR(m_phase==PHASE_PREPARE_AFTER
		,BZPERROR_CALLED_MISTAKE_ORDER);

	m_phase=PHASE_DRAW_AFTER;
}

//pXŜ\鏀
void Redraw::PrepareDraw()
{
	BZP_PARAMETER_ERROR(m_phase==PHASE_BEGIN
		,BZPERROR_CALLED_MISTAKE_ORDER);

	if(m_graph->IsInUse()) {
		m_canRedraw=false;
	} else {
		m_graph->PrepareBeforeBitmap();
	}
	m_phase=PHASE_PREPARE_DRAW;
}

//pXŜ`悷
void Redraw::Draw()
{
	if(m_phase!=PHASE_PREPARE_DRAW) {
		PrepareDraw();
	}
	if(m_canRedraw) {
		if(m_tool==NULL) {
			m_path->Draw(m_graph);
		} else {
			m_tool->Draw(m_path,m_graph);
		}
	}
	m_phase=PHASE_DRAW;
}

//pX̕\Ȃ
void Redraw::SkipDraw()
{
	if(m_phase!=PHASE_PREPARE_DRAW) {
		PrepareDraw();
	}
	m_phase=PHASE_DRAW;
}

//ĕ\顕KĂ΂Ȃ΂Ȃ
void Redraw::End()
{
	if(m_phase==PHASE_BEGIN || m_phase==PHASE_PREPARE_DRAW) {
		Draw();
	} else if(m_phase<PHASE_DRAW_AFTER) {
		DrawAfter();
	}
	assert(m_phase==PHASE_DRAW || m_phase==PHASE_DRAW_AFTER);

	if(m_canRedraw) {
		if(m_phase==PHASE_DRAW) {
			m_graph->DrawAll();
		} else {
			m_graph->DrawDifference();
		}
	}

	m_phase=PHASE_END;
}

BZP_NAMESPACE_END
