#include "prefix.h"

#include <deque>
#include <new>

#include "steps.h"
#include "curvepointers.h"
#include "path.h"
#include "curve.h"
#include "knot.h"
#include "bezier.h"
#include "fpoint.h"
#include "error.h"
#include "exception.h"
#include "part.h"

/*************************************************

	̊Ԋu̓_

**************************************************/

BZP_NAMESPACE_START

//W߂ĒuNX
class PointQue : public std::deque<FPoint>
{

};

class StepsMember
{
public:
	CurvePointers curvePtrs;	//߂Ēu
	PointQue steps;	//Ԋu̓_߂Ēu
};

Steps::Steps()
	:m_m(new StepsMember)
	,m_option(0)
	,m_polylineResolution(0.25)
	,m_intervalLeft(0)
	,m_knotIndex(0)
	,m_curveIndex(0)
{
	BZP_BAD_ALLOC_TEST(m_m.get());
}

Steps::~Steps() throw()
{

}

/*pX𒊏o
path@Ώۂ̃pX
part  pX̕ PART_SELECTED PART_UNSELECTED PART_SUBGROUP PART_CURVE
*/
void Steps::AddPath(const Path *path,int part)
{
	BZP_PARAMETER_ERROR(path!=NULL,BZPERROR_NULL_ARGUMENT);

	path->ExtractCurve(&(m_m->curvePtrs),part);
}

//܂ɕϊׂO傫łȂƂȂ
void Steps::SetResolution(double resolution)
{
	BZP_PARAMETER_ERROR(resolution>0,BZPERROR_ZERO_RESOLUTION);
	m_polylineResolution=resolution;
}

//ݒ
void Steps::SetOption(int option)
{
	m_option=option;
}

//Ԋuݒ
void Steps::SetInterval(double interval)
{
	BZP_PARAMETER_ERROR(interval>0,BZPERROR_ZERO_INTERVAL);

	m_interval=interval;
}

//_
//߂lfalseȂ_Ȃ̂ŏI
bool Steps::NextPoint(FPoint *point)
{
	BZP_PARAMETER_ERROR(point!=NULL,BZPERROR_NULL_ARGUMENT);

	if(m_m->steps.size()==0) {
		ChargeSteps();
	}
	if(m_m->steps.size()==0) {
		return false;
	}
	*point=m_m->steps.front();
	m_m->steps.pop_front();

	return true;
}

bool Steps::NextPoint(double *x,double *y)
{
	BZP_PARAMETER_ERROR(x!=NULL && y!=NULL
		,BZPERROR_NULL_ARGUMENT);

	FPoint point;
	bool result=NextPoint(&point);
	if(result) {
		*x=point.x;
		*y=point.y;
	}
	return result;
}

//Ȑ_ɂ
void Steps::ChargeSteps()
{
	int curveCount=m_m->curvePtrs.Count();
	FOR(;m_curveIndex<curveCount;++m_curveIndex,m_knotIndex=0) {
		const Curve *curve=m_m->curvePtrs.GetCurve(m_curveIndex);
		assert(curve!=NULL);

		bool isClosed=curve->IsClosed();

		int startSegment=0;
		int knotCount=curve->KnotCount();
		int endSegment=knotCount-1;
		if(!isClosed) {
			--endSegment;
		} else if(DoesAdjustCorner()) {
			startSegment=FirstCorner(curve);
			if(startSegment>0) {
				endSegment=startSegment-1;
			}
		}

		PointQue polyline;


		if(knotCount>1) {
			FOR(;m_knotIndex<knotCount-(isClosed?0:1);) {
				int segment=m_knotIndex+startSegment;
				++m_knotIndex;
				if(segment>=knotCount) {
					segment-=knotCount;
				}

				const Knot *curKnot=curve->GetKnot(segment);
				const Knot *nextKnot=curve->GetNextKnot(segment);
				assert(curKnot!=NULL && nextKnot!=NULL);

				ResolveSegment(&polyline,curKnot,nextKnot);

				if(!isClosed && segment==endSegment) {
					m_m->steps.push_back(curve->GetPoint(PART_ANCHOR,startSegment));
					m_intervalLeft=AssignSteps(&(m_m->steps),polyline,m_intervalLeft,DoesAdjustEnd());
					return;
				} else if(isClosed && segment==endSegment) {
					m_intervalLeft=AssignSteps(&(m_m->steps),polyline,m_intervalLeft,DoesAdjustClosed());
					return;
				} else if(DoesAdjustCorner() && !nextKnot->IsCurved(PART_ANCHOR)) {
					AssignSteps(&(m_m->steps),polyline,m_intervalLeft,true);
					m_intervalLeft=0;
					return;
				} else if(!DoesAdjust()) {
					m_intervalLeft=AssignSteps(&(m_m->steps),polyline,m_intervalLeft,false);
					return;
				} 
			}
		} else if(knotCount==1 && DoesPutSingleKnot()) {
			m_m->steps.push_back(curve->GetPoint(PART_ANCHOR,0));
		}
	}
}

//Ȑ𕪊čׂ܂ɂ
void Steps::ResolveSegment(PointQue *polyline,const Knot *curKnot,const Knot *nextKnot) const
{
	assert(polyline!=NULL);

	Bezier bezier(curKnot,nextKnot);

	FPoint resolutionSize(m_polylineResolution,m_polylineResolution);

	FOR(;;) {
		if(bezier.IsNarrow(resolutionSize) || bezier.IsNotCurve()) {
			polyline->push_back(bezier.GetAnchorA());
			if(!bezier.Next()) break;
		} else {
			bezier.Divide();
		}
	}
	polyline->push_back(bezier.GetAnchorB());
}

//Ԋu̓_蓖Ă		
double Steps::AssignSteps(PointQue *steps,const PointQue &polyline,double intervalLeft,bool doesAdjust) const
{
	if(doesAdjust) {
		AssignStepsAdjust(steps,polyline);
		intervalLeft=0;
	} else {
		intervalLeft=AssignStepsNotAdjust(steps,polyline,intervalLeft);
	}
	return intervalLeft;
}

//_̊Ԋu𒲐Ċ蓖Ă
void Steps::AssignStepsAdjust(PointQue *steps,const PointQue &polyline) const
{
	double length=PolylineLength(polyline);
	int stepCount=ceil(length/m_interval);
	double interval=length/stepCount;

	double intervalLeft=interval;

	steps->push_back(polyline.front());

	int lineCount=polyline.size()-1;
	FOR(int i=0;i<lineCount;++i) {
		FPoint curPoint=polyline[i];
		FPoint nextPoint=polyline[i+1];

		double lineLength=curPoint.Distance(nextPoint);

		FOR(;;) {
			if(lineLength<=intervalLeft) {
				intervalLeft-=lineLength;
				break;
			} else {
				lineLength-=intervalLeft;
				intervalLeft=interval;
				steps->push_back(
					nextPoint.LeavePoint(curPoint,lineLength));
				if(--stepCount==0) break;
			}
		}
	}

	assert(stepCount<=1);

	if(stepCount==1) {
		steps->push_back(polyline.back());
	}
}

//_̊Ԋu𒲐Ɋ蓖Ă
double Steps::AssignStepsNotAdjust(PointQue *steps,const PointQue &polyline,double intervalLeft) const
{
	int lineCount=polyline.size()-1;
	FOR(int i=0;i<lineCount;++i) {
		FPoint curPoint=polyline[i];
		FPoint nextPoint=polyline[i+1];

		double lineLength=curPoint.Distance(nextPoint);

		FOR(;;) {
			if(lineLength<=intervalLeft) {
				intervalLeft-=lineLength;
				break;
			} else {
				lineLength-=intervalLeft;
				intervalLeft=m_interval;
				steps->push_back(
					nextPoint.LeavePoint(curPoint,lineLength));
			}
		}
	}
	return intervalLeft;
}

//܂̒
double Steps::PolylineLength(const PointQue &polyline) const
{
	double length=0;
	int lineCount=polyline.size()-1;
	FOR(int i=0;i<lineCount;++i) {
		FPoint curPoint=polyline[i];
		FPoint nextPoint=polyline[i+1];
		length+=curPoint.Distance(nextPoint);
	}
	return length;
}

//̍ŏ̊p
int Steps::FirstCorner(const Curve *curve) const
{
	int knotCount=curve->KnotCount();
	FOR(int i=0;i<knotCount;++i) {
		if(!curve->IsCurved(PART_ANCHOR,i)) {
			return i;
		}
	}
	return 0;
}

//Ԋu̒ǂł邩
bool Steps::DoesAdjust() const
{
	return DoesAdjustCorner() || DoesAdjustEnd() || DoesAdjustClosed();
}

//pŊԊu̒邩
bool Steps::DoesAdjustCorner() const
{
	return (m_option & STEPS_NOT_ADJUST_CORNER)==0;
}

//̖[ŊԊu̒邩
bool Steps::DoesAdjustEnd() const
{
	return (m_option & STEPS_NOT_ADJUST_END)==0;
}

//ŊԊu̒邩
bool Steps::DoesAdjustClosed() const
{
	return (m_option & STEPS_NOT_ADJUST_CLOSED)==0;
}

//Ǘpڂɂ_u
bool Steps::DoesPutSingleKnot() const
{
	return (m_option & STEPS_NOT_PUT_SINGLEKNOT)==0;
}

BZP_NAMESPACE_END
