#include "prefix.h"

#include "subgroup.h"
#include "curve.h"
#include "findresult.h"
#include "partmask.h"
#include "selectmode.h"
#include "graph.h"
#include "curvepointers.h"
#include "converterbase.h"

/******************************************************

	TuO[vNX̒`
	GrouppĂGroup̊֐g

*******************************************************/

BZP_NAMESPACE_START

BZP_PARTMASK_USING

SubGroup::SubGroup()
	:m_isSelected(true)
{

}

SubGroup::~SubGroup() throw()
{
}

//IĂ邩
bool SubGroup::IsSelected() const
{
	return m_isSelected;
}

//IEI
void SubGroup::Select(bool isSelected)
{
	m_isSelected=isSelected;
}

//O[v`
void SubGroup::DrawSubGroup(const Graph *graph) const
{
	//ԏ̊KwőIĂꍇɑIĂ邱ƂɂȂ
	if(graph->DoesDraw()) {
		if(IsSelected()) {
			DrawSelected(graph);
		} else if(graph->DoesDrawUnselected()) {
			DrawUnselected(graph);
		}
	}
}

//IĂO[v`
void SubGroup::DrawSelected(const Graph *graph) const
{
	int subGroupCount=SubGroupCount();
	FOR(int i=0;i<subGroupCount;++i) {
		GetSubGroup(i)->DrawSelected(graph);
	}

	int curveCount=CurveCount();
	FOR(int i=0;i<curveCount;++i) {
		GetCurve(i)->DrawSelectedSubGroup(graph);
	}
}

//IĂȂO[v`
void SubGroup::DrawUnselected(const Graph *graph) const
{
	if(graph->DoesDrawUnselected()) {
		int subGroupCount=SubGroupCount();
		FOR(int i=0;i<subGroupCount;++i) {
			GetSubGroup(i)->DrawUnselected(graph);
		}

		int curveCount=CurveCount();
		FOR(int i=0;i<curveCount;++i) {
			GetCurve(i)->DrawUnselectedSubGroup(graph);
		}
	}
}

//W̎TuO[ṽpX̕ʂT
bool SubGroup::FindSubGroup(FindResult *findResult,const Graph *graph,int mask) const
{
	if(IsSelectedOpposite(mask,IsSelected())) {
		return false;
	}

	if(!Group::FindGroup(findResult,graph,PART_CURVE|PART_SUBGROUP)) {
		return false;
	}

	findResult->ClearPart();
	findResult->SetPart(PART_SUBGROUP);
	findResult->SetSelect(IsSelected());
	return true;
}

//ΓIɈړ
void SubGroup::MoveRel(const FPoint & rel)
{
	int curveCount=CurveCount();
	FOR(int i=0;i<curveCount;++i) {
		GetCurve(i)->MoveRel(rel);
	}

	int subGroupCount=SubGroupCount();
	FOR(int i=0;i<subGroupCount;++i) {
		GetSubGroup(i)->MoveRel(rel);
	}
}

//`I
void SubGroup::SelectBoxSubGroup(SELECTMODE mode,const FRect &rect)
{
	if(!BoxFind(rect)) {
		return;
	}

	bool isSelected=IsSelected();
	switch(mode) {
	case SELECTMODE_ADD:
	case SELECTMODE_SINGLE:
		isSelected=true;
		break;

	case SELECTMODE_UNSELECT:
		isSelected=false;
		break;

	case SELECTMODE_TOGGLE:
		isSelected=!isSelected;
		break;

	default:
		assert(false);
	}
	Select(isSelected);
}

//`ɃTuO[v̐ꕔłĂ邩
bool SubGroup::BoxFind(const FRect &rect) const
{
	bool isCurveIn=false;

	int curveCount=CurveCount();
	FOR(int i=0;i<curveCount;++i) {
		if(GetCurve(i)->IsInRect(rect)) {
			isCurveIn=true;
			break;
		}
	}

	if(!isCurveIn) {
		int subGroupCount=SubGroupCount();
		FOR(int i=0;i<subGroupCount;++i) {
			if(GetSubGroup(i)->BoxFind(rect)) {
				isCurveIn=true;
				break;
			}
		}
	}
	return isCurveIn;
}

//g̐̃AhXW߂
void SubGroup::ExtractCurve(CurvePointers *curvePtrs,int part) const
{
	int curveCount=CurveCount();
	FOR(int i=0;i<curveCount;++i) {
		const Curve *curve=GetCurve(i);
		curvePtrs->AddCurve(curve);
	}

	int subGroupCount=SubGroupCount();
	FOR(int i=0;i<subGroupCount;++i) {
		GetSubGroup(i)->ExtractCurve(curvePtrs,part);
	}
}

//炩̃f[^\`ɕϊ
void SubGroup::PathToDataSubGroup(PathTo *const converter,int part) const
{
	assert(converter!=NULL);

	if(!IsSelectedOpposite(part,IsSelected())) {
		converter->WriteStartSubGroup(IsSelected());
		PathToDataGroup(converter,PART_ALL);
		converter->WriteEndSubGroup();
	}
}

//炩̃f[^\`pXɕϊ
void SubGroup::DataToPathSubGroup(ToPath *const converter)
{
	assert(converter!=NULL);

	converter->ReadStartSubGroup();
	DataToPathGroup(converter);
	converter->ReadEndSubGroup(&m_isSelected);
}

//TuO[v𕡐
void SubGroup::CopySubGroup(const SubGroup *srcSubGroup)
{
	Select(srcSubGroup->IsSelected());
	CopyGroup(srcSubGroup);
}

BZP_NAMESPACE_END

