#include "prefix.h"

#include "toolbase.h"
#include "runtimeconfig.h"
#include "fpoint.h"
#include "path.h"
#include "event.h"
#include "extractoption.h"
#include "graph.h"
#include "keyboard.h"

/*************************************************

	c[̊{NX

**************************************************/

BZP_NAMESPACE_START

ToolBase::ToolBase()
	:m_path(NULL)
	,m_graph(NULL)
	,m_condition(0)
	,m_outsideWindow(false)
	,m_didModify(false)
{

}

ToolBase::~ToolBase() throw()
{

}

//c[̃IvV̐ݒ
bool ToolBase::SetOption(int option)
{
	if(IsOperating()) return false;

	m_option=option;
	return true;
}

//ݐݒ肳ĂIvV𓾂
int ToolBase::GetOption() const
{
	return m_option;
}

//IvV𕔕IɎo
int ToolBase::ExtractOption(int lowBit,int hiBit) const
{
	return BZP_EXOPTION::ExtractOption(m_option,lowBit,hiBit);
}

//݂̏Ԃ𓾂
int ToolBase::GetCondition() const
{
	return m_condition;
}

//L[ĂԂ
bool ToolBase::IsPushingKey(CONFIG_INT config) const
{
	return BZP_KEYBOARD::IsPushingKey(config);
}

//keyŎw肵L[ĂԂ
bool ToolBase::IsPushedKey(CONFIG_INT config,int key) const
{
	return BZP_KEYBOARD::IsPushedKey(config,key);
}

//Cxg
void ToolBase::Entry(EVENT event,Path *path,Graph *graph,POINT viewPoint,int key)
{
	assert(path!=NULL && graph!=NULL);

	m_path=path;
	m_graph=graph;

	bool isPointInView;
	switch(event) {
	case EVENT_KEYDOWN:
		if(IsPushedKey(KEY_CANCEL,key)) {
			event=EVENT_CANCEL;
		}
		break;

	case EVENT_MOUSEMOVE:
		isPointInView=m_graph->IsPointInView(viewPoint);
		if(isPointInView && m_outsideWindow) {
			m_outsideWindow=false;
			event=EVENT_MOVEINSIDEWINDOW;
		} else if(!isPointInView && !m_outsideWindow) {
			m_outsideWindow=true;
			event=EVENT_MOVEOUTSIDEWINDOW;
		}
		break;
	}	

	FPoint boundPoint=m_graph->ViewToBound(viewPoint);

	//pNX̃CxgĂ
	Event(event,boundPoint,key);
}

//pNXŒ`ĂȂꍇ̊̃Cxg
void ToolBase::Event(EVENT event,const FPoint &,int)
{
	switch(event) {
	case EVENT_DRAW:
		OnDraw();
		break;
	}
}

//ʏ̕\Cxg
void ToolBase::OnDraw()
{
	m_path->Draw(m_graph);
}

//𒆒fđ̓邱Ƃł邩
bool ToolBase::CanStop() const
{
	return !IsOperating();
}

//pXύXꂽ̃tO̐ݒ
void ToolBase::SetModifiedFlag(bool didModify)
{
	m_didModify=didModify;
}

//pXύXꂽ
bool ToolBase::DidModify() const
{
	return m_didModify;
}

BZP_NAMESPACE_END
