#include "prefix.h"

#include "toolchangecontrol.h"
#include "path.h"
#include "findresult.h"
#include "part.h"
#include "fpoint.h"
#include "event.h"
#include "selectmode.h"
#include "partmask.h"

/*************************************************

	_̏Ԃ؂ւ

**************************************************/

BZP_NAMESPACE_START

enum
{
	PHASE_IDLE,	//ĂȂ
	PHASE_GRIPPEDANCHOR,	//ʉߓ_NbN
	PHASE_GRIPPEDCONTROL,	//_NbN
	PHASE_EXTEND,	//ړc[̋@\𗘗pĂ
};

ToolChangeControl::ToolChangeControl()
	:m_phase(PHASE_IDLE)
{
	m_condition=COND_CHANGECONTROL_IDLE;
}

//쒆
bool ToolChangeControl::IsOperating() const
{
	return m_phase!=PHASE_IDLE
		&&(m_phase!=PHASE_EXTEND || ToolExtend::IsOperating());
}

//}EXLv`[邩
bool ToolChangeControl::IsCapturing() const
{
	return m_phase!=PHASE_IDLE
		&&(m_phase!=PHASE_EXTEND || ToolExtend::IsCapturing());
}

//CxgԂƂɐU蕪
void ToolChangeControl::Event(EVENT event,const FPoint & point,int key)
{
	switch(m_phase) {
	case PHASE_IDLE:
		EventIdle(event,point);
		break;

	case PHASE_GRIPPEDANCHOR:
		EventGrippedAnchor(event,point);
		break;

	case PHASE_GRIPPEDCONTROL:
		EventGrippedControl(event,point);
		break;

	case PHASE_EXTEND:
		EventExtend(event,point,key);
		break;
	}
}

//ĂȂԂɂ
void ToolChangeControl::Idle(const FPoint & point)
{
	m_phase=PHASE_IDLE;
	Idle_Move(point);
}

//ĂȂԂł̃Cxg
void ToolChangeControl::EventIdle(EVENT event,const FPoint & point)
{
	switch(event) {
	case EVENT_DRAW:
		OnDraw();
		break;

	case EVENT_LEFTDOWN:
		Start(point);
		break;

	case EVENT_MOUSEMOVE:
		Idle_Move(point);
		break;
	}
}

//ĂȂԂŃJ[\
void ToolChangeControl::Idle_Move(const FPoint & point)
{
	FindResult findResult(point);
	m_path->Find(&findResult,m_graph,PART_KNOT);
	m_condition=COND_CHANGECONTROL_IDLE;
	int part=findResult.GetPart();

	switch(BZP_PARTMASK::GeneralPart(part)) {
	case PART_ANCHOR:
		if(findResult.IsKnotOnSelectedCurve()) {
			if(findResult.IsCurved()) {
				m_condition=COND_CHANGE_STRAIGHT_KNOT;
			} else {
				m_condition=COND_CHANGE_CURVE_KNOT;
			}
		}
		break;

	case PART_CONTROL:
		if(findResult.IsLinked()) {
			m_condition=COND_CHANGE_LINKED_CONTROL;
		} else {
			m_condition=COND_CHANGE_UNLINKED_CONTROL;
		}
		break;

	default:
		if(!ToolExtend::Idle_Move(point)) {
			m_condition=COND_CHANGECONTROL_IDLE;
		}
		break;
	}
}

//ĂȂԂōNbN
void ToolChangeControl::Start(const FPoint & point)
{
	FindResult findResult(point);
	m_path->Find(&findResult,m_graph,PART_KNOT);

	m_grippedKnot=findResult.GetKnotIndex();
	m_grippedCurve=findResult.GetCurveIndex();

	switch(findResult.GeneralPart()) {
	case PART_ANCHOR:
		m_grippedIsSelected=findResult.IsSelected();
		ChangeAnchor(point);
		SetModifiedFlag();
		break;

	case PART_CONTROL:
		m_grippedIsForward=findResult.IsForwardControl();
		ChangeControl(point);
		SetModifiedFlag();
		break;

	default:
		m_phase=PHASE_EXTEND;
		ToolExtend::Start(point);
		break;
	}

}

//ʉߓ_NbN
void ToolChangeControl::ChangeAnchor(const FPoint & point)
{
	if(!m_grippedIsSelected) {
		m_path->SelectKnot(SELECTMODE_SINGLE,m_grippedKnot,m_grippedCurve);
	}

	m_grippedIsCurved=m_path->IsCurved(PART_ANCHOR,m_grippedKnot,m_grippedCurve);
	if(m_grippedIsCurved) {
		m_path->CurveKnot(false,PART_ANCHOR,m_grippedKnot,m_grippedCurve);
	}

	if(MoveMode()!=OPTION_MOVE_NODRAG) {
		m_phase=PHASE_GRIPPEDANCHOR;
	} else {
		if(!m_grippedIsCurved) {
			m_phase=PHASE_EXTEND;
			ToolExtend::MoveWControls(point,true,m_grippedKnot,m_grippedCurve);
		}
	}
}

//ʉߓ_NbNĉ܂
void ToolChangeControl::EventGrippedAnchor(EVENT event,const FPoint & point)
{
	switch(event) {
	case EVENT_DRAW:
		OnDraw();
		break;

	case EVENT_DRAG:
		m_phase=PHASE_EXTEND;
		ToolExtend::DragWControls(point,true,m_grippedKnot,m_grippedCurve);
		break;

	case EVENT_LEFTUP:
		if(!m_grippedIsCurved && MoveMode()==OPTION_MOVE_2WAY) {
			m_phase=PHASE_EXTEND;
			ToolExtend::MoveWControls(point,true,m_grippedKnot,m_grippedCurve);
		} else {
			Idle(point);
		}
		break;

	case EVENT_CANCEL:
		Idle(point);
		break;
	}
}

//_NbN
void ToolChangeControl::ChangeControl(const FPoint & point)
{
	m_path->LinkKnot(false,m_grippedKnot,m_grippedCurve);

	if(MoveMode()!=OPTION_MOVE_NODRAG) {
		m_phase=PHASE_GRIPPEDCONTROL;
	} else {
		m_phase=PHASE_EXTEND;
		ToolExtend::MoveControl(point,m_grippedIsForward,m_grippedKnot,m_grippedCurve);
	}
}

//_NbNĉ܂
void ToolChangeControl::EventGrippedControl(EVENT event,const FPoint & point)
{
	switch(event) {
	case EVENT_DRAW:
		OnDraw();
		break;

	case EVENT_DRAG:
		m_phase=PHASE_EXTEND;
		ToolExtend::DragControl(point,m_grippedIsForward,m_grippedKnot,m_grippedCurve);
		break;

	case EVENT_LEFTUP:
		if(MoveMode()==OPTION_MOVE_2WAY) {
			m_phase=PHASE_EXTEND;
			ToolExtend::MoveControl(point,m_grippedIsForward,m_grippedKnot,m_grippedCurve);
		} else {
			Idle(point);
		}
		break;

	case EVENT_CANCEL:
		Idle(point);
		break;
	}
}

//ړc[̃Cxgɂ܂킷
void ToolChangeControl::EventExtend(EVENT event,const FPoint & point,int key)
{
	ToolExtend::Event(event,point,key);
	if(event==EVENT_CANCEL || !ToolExtend::IsOperating()) {
		Idle(point);
	}
}

BZP_NAMESPACE_END
