#include "prefix.h"

#include <assert.h>
#include <memory>
#include <new>
#include <limits.h>

#include "toolholder.h"
#include "tool.h"
#include "toolbase.h"
#include "runtimeconfig.h"
#include "keyboard.h"
#include "toolnone.h"
#include "event.h"
#include "toolheaders.h"
#include "exception.h"
#include "error.h"

/*************************************************

	c[IuWFNg̏LƐ؂ւ

**************************************************/

BZP_NAMESPACE_START

BZP_KEYBOARD_USING

/*c[ϊ\
ڂ̓L[Ȃԡ̌̓L[ԡ

Ń}Nłꍇ͖̂ʓ|Ȃ̂
Ԃ̃IvVȂǂύXꍇ
AvP[Vŕ\ݒ肷΂悢
}NBZP_DONT_USE_DEFAULT_TOOLTABLE`Ί̕\͎gȂ
*/
#ifndef BZP_DONT_USE_DEFAULT_TOOLTABLE
static TOOLTABLE s_defaultToolTable[TOOLTABLE_COUNT]=
{
	//TOOLTABLE_NONE
	{ TOOL_NONE,0,
		TOOL_NONE,0,
		TOOL_NONE,0,

		TOOL_NONE,0,
		TOOL_NONE,0,
		TOOL_NONE,0,
		TOOL_NONE,0,
	},

	//TOOLTABLE_MOVE
	{ TOOL_MOVE,0,
		TOOL_MOVE,OPTION_SELECT_TOGGLE,
		TOOL_CHANGECONTROL,OPTION_SELECT_ADD,
		TOOL_SELECTKNOTS,0,

		TOOL_MOVE,OPTION_MOVE_NODRAG,
		TOOL_MOVE,OPTION_MOVE_2WAY,
		TOOL_MOVE,OPTION_MOVE_SUBGROUP_WITH_CURVE|OPTION_MOVE_CURVE_JUST|OPTION_SELECT_ADD,
		TOOL_MOVE,OPTION_MOVE_SUBGROUP_WITH_KNOT|OPTION_MOVE_CURVE_SEGMENT,
	},

#ifndef BZP_DONT_USE_PEN
	//TOOLTABLE_PEN
	{ TOOL_PEN,0,
		TOOL_MOVE,0,
		TOOL_INSERTKNOT,0,
		TOOL_DELETEPOINT,0,

		TOOL_PEN,OPTION_PEN_NORUBBER,
		TOOL_NONE,0,
		TOOL_NONE,0,
		TOOL_NONE,0,
	},
#endif

	//TOOLTABLE_INSERTKNOT
	{ TOOL_INSERTKNOT,0,
		TOOL_MOVE,0,
		TOOL_DELETEPOINT,0,
		TOOL_CHANGECONTROL,0,

		TOOL_INSERTKNOT,OPTION_INSERTKNOT_STRAIGHT_CURVE,
		TOOL_INSERTKNOT,OPTION_INSERTKNOT_MOVE_ANCHOR,
		TOOL_NONE,0,
		TOOL_NONE,0,
	},

	//TOOLTABLE_DELETEPOINT
	{ TOOL_DELETEPOINT,0,
		TOOL_MOVE,0,
		TOOL_INSERTKNOT,0,
		TOOL_DELETEPOINT,0,
	
		TOOL_NONE,0,
		TOOL_NONE,0,
		TOOL_NONE,0,
		TOOL_NONE,0,
	},

	//TOOLTABLE_CHANGECONTROL
	{ TOOL_CHANGECONTROL,0,
		TOOL_MOVE,0,
		TOOL_INSERTKNOT,0,
		TOOL_DELETEPOINT,0,

		TOOL_NONE,0,
		TOOL_NONE,0,
		TOOL_NONE,0,
		TOOL_NONE,0,
	},

#ifndef BZP_DONT_USE_CUT
	//TOOLTABLE_CUT
	{ TOOL_CUT,0,
		TOOL_MOVE,0,
		TOOL_INSERTKNOT,0,
		TOOL_DELETEPOINT,0,

		TOOL_CUT,OPTION_CUT_SEGMENT,
		TOOL_NONE,0,
		TOOL_NONE,0,
		TOOL_NONE,0,
	},
#endif

#ifndef BZP_DONT_USE_SHAPE
	//TOOLTABLE_BOX
	{ TOOL_SHAPE,OPTION_SHAPE_BOX,
		TOOL_SHAPE,OPTION_SHAPE_BOX|OPTION_SHAPE_CENTER,
		TOOL_SHAPE,OPTION_SHAPE_BOX|OPTION_SHAPE_SQUARE,
		TOOL_MOVE,0,

		TOOL_SHAPE,OPTION_SHAPE_BOX|OPTION_SHAPE_DRAG,
		TOOL_SHAPE,OPTION_SHAPE_BOX|OPTION_SHAPE_NODRAG,
		TOOL_NONE,0,
		TOOL_NONE,0,
	},

	//TOOLTABLE_CIRCLE
	{ TOOL_SHAPE,OPTION_SHAPE_CIRCLE,
		TOOL_SHAPE,OPTION_SHAPE_CIRCLE|OPTION_SHAPE_CENTER,
		TOOL_SHAPE,OPTION_SHAPE_CIRCLE|OPTION_SHAPE_SQUARE,
		TOOL_MOVE,0,

		TOOL_SHAPE,OPTION_SHAPE_CIRCLE|OPTION_SHAPE_DRAG,
		TOOL_SHAPE,OPTION_SHAPE_CIRCLE|OPTION_SHAPE_NODRAG,
		TOOL_NONE,0,
		TOOL_NONE,0,
	},
#endif

	//TOOLTABLE_SELECTKNOTS
	{ TOOL_SELECTKNOTS,0,
		TOOL_SELECTKNOTS,OPTION_SELECTKNOTS_ALTERNATE,
		TOOL_SELECTKNOTS,OPTION_SELECTKNOTS_SELECT,
		TOOL_SELECTKNOTS,OPTION_SELECTKNOTS_UNSELECT,

		TOOL_NONE,0,
		TOOL_NONE,0,
		TOOL_NONE,0,
		TOOL_NONE,0,
	},

};
#endif //BZP_DONT_USE_DEFAULT_TOOLTABLE

ToolHolder::ToolHolder()
	:m_toolObj(new ToolNone)
#ifndef BZP_DONT_USE_DEFAULT_TOOLTABLE
	,m_toolTable(s_defaultToolTable)
	,m_tableCount(TOOLTABLE_COUNT)
#else
	,m_toolTable(NULL)
#endif
	,m_currentIndex(TOOLTABLE_NONE)
	,m_currentKey(0)
	,m_currentTool(TOOL_NONE)
{
	BZP_BAD_ALLOC_TEST(m_toolObj.get());
}

ToolHolder::ToolHolder(const TOOLTABLE *toolTable,int count)
	:m_toolObj(new ToolNone)
	,m_toolTable(toolTable)
	,m_tableCount(count)
	,m_currentIndex(0)
	,m_currentKey(0)
	,m_currentTool(TOOL_NONE)
{
	BZP_BAD_ALLOC_TEST(m_toolObj.get());
}

ToolHolder::~ToolHolder() throw()
{

}

/*c[ϊ\̐ݒ
toolTableNULLȂ珉ݒɖ߂*/
void ToolHolder::SetToolTable(const TOOLTABLE *toolTable,int count)
{
#ifndef BZP_DONT_USE_DEFAULT_TOOLTABLE
	if(toolTable==NULL) {
		m_toolTable=s_defaultToolTable;
		m_tableCount=TOOLTABLE_COUNT;
	} else {
#else
	BZP_PARAMETER_ERROR(toolTable!=NULL && count>=1
		,BZPERROR_NULL_ARGUMENT);
#endif
		m_toolTable=toolTable;
		m_tableCount=count;
#ifndef BZP_DONT_USE_DEFAULT_TOOLTABLE
	}
#endif

	m_currentIndex=0;
}

/*c[̐؂ւgpc[̕\̔ԍw肷*/
void ToolHolder::SelectTool(int index,Path *path,Graph *graph)
{
	BZP_PARAMETER_ERROR(0<=index && index<m_tableCount
		,BZPERROR_OUT_OF_RANGE);
	BZP_PARAMETER_ERROR(path!=NULL && graph!=NULL
		,BZPERROR_NULL_ARGUMENT);

	BZP_PARAMETER_ERROR(m_toolTable!=NULL
		,BZPERROR_NO_TOOL_TABLE);

	if(m_toolObj->IsOperating()) {
		m_toolObj->Entry(EVENT_CANCEL,path,graph);
	}
	
	SetTool(index,0);
}

//ƃL[̏Ԃɂăc[IuWFNgƃIvV؂ւ
void ToolHolder::SetTool(int index,int key)
{
	assert(m_toolTable!=NULL);	/*̕\gȂ̂
			AvP[VŒ`Ȃ΂Ȃ*/

	TOOL newTool=(TOOL)(m_toolTable[index].keyset[key].tool);
	BZP_PARAMETER_ERROR(0<=newTool && newTool<TOOL_MAX
		,BZPERROR_BAD_TOOL_TABLE);

	LONG newOption=m_toolTable[index].keyset[key].option;

	if(newTool!=m_currentTool) {
		BZP_AUTOPTR_RESET(m_toolObj,ToolBase
			,CreateToolObject(newTool));
		BZP_BAD_ALLOC_TEST(m_toolObj.get());
		m_currentTool=newTool;
	}
	m_toolObj->SetOption(newOption);
	m_currentIndex=index;
	m_currentKey=key;
}

//c[IuWFNg̃Cxg
void ToolHolder::Entry(EVENT event,Path *path,Graph *graph,POINT point,int key)
{
	BZP_PARAMETER_ERROR(path!=NULL && graph!=NULL
		,BZPERROR_NULL_ARGUMENT);
	BZP_PARAMETER_ERROR(0<=event && event<EVENT_LIMIT
		,BZPERROR_OUT_OF_RANGE);

	m_toolObj->Entry(event,path,graph,point,key);

	if(!IsOperating()) {
		AlternateTool(path,graph,point);
	}
}

void ToolHolder::Entry(EVENT event,Path *path,Graph *graph)
{
	POINT point={-3000,-3000};	/*KȍW*/
	Entry(event,path,graph,point);
}

//c[oRĕ\
void ToolHolder::Draw(Path *path,Graph *graph)
{
	BZP_PARAMETER_ERROR(path!=NULL && graph!=NULL
		,BZPERROR_NULL_ARGUMENT);

	m_toolObj->Entry(EVENT_DRAW,path,graph);
}

//c[̓𒆎~
void ToolHolder::Cancel(Path *path,Graph *graph)
{
	BZP_PARAMETER_ERROR(path!=NULL && graph!=NULL
		,BZPERROR_NULL_ARGUMENT);

	m_toolObj->Entry(EVENT_CANCEL,path,graph);
}

const enum CONFIG_INT s_toolHolderKey[TOOLHOLDER_KEY_COUNT]=
{
	KEY_ALTERNATETOOL1,
	KEY_ALTERNATETOOL2,
	KEY_ALTERNATETOOL3,

	KEY_OPTIONTEST1,
	KEY_OPTIONTEST2,
	KEY_OPTIONTEST3,
	KEY_OPTIONTEST4,
};

//ݒ肳ĂL[ꂽꍇ̓c[؂ւ
void ToolHolder::AlternateTool(Path *path,Graph *graph,POINT point)
{
	int key=0;
	FOR(int i=0;i<TOOLHOLDER_KEY_COUNT;++i) {
		if(IsPushingKey(s_toolHolderKey[i])) {
			key=i+1;
			break;
		}
	}

	if(m_currentKey!=key) {
		SetTool(m_currentIndex,key);
		m_toolObj->Entry(EVENT_MOUSEMOVE,path,graph,point);
	}
}

//c[IuWFNgs
bool ToolHolder::IsOperating() const
{
	assert(m_toolObj.get()!=NULL);
	return m_toolObj->IsOperating();
}

//c[IuWFNg}EXLv`[KvƂ邩
bool ToolHolder::IsCapturing() const
{
	assert(m_toolObj.get()!=NULL);
	return m_toolObj->IsCapturing();
}

//c[IuWFNg𒆒fł邩
bool ToolHolder::CanStop() const
{
	assert(m_toolObj.get()!=NULL);
	return m_toolObj->CanStop();
}

//ݑIĂc[\̍
int ToolHolder::GetCurrentIndex() const
{
	return m_currentIndex;
}

//c[̏
int ToolHolder::GetCondition() const
{
	assert(m_toolObj.get()!=NULL);
	return m_toolObj->GetCondition();
}

//c[IuWFNg쐬
ToolBase *ToolHolder::CreateToolObject(TOOL tool)
{
	ToolBase *toolObj=NULL;

	switch(tool) {
	case TOOL_NONE:
		toolObj=static_cast<ToolBase*>(new ToolNone);
		break;

#ifndef BZP_DONT_USE_PEN
	case TOOL_PEN:
		toolObj=static_cast<ToolBase*>(new ToolPen);
		break;
#endif

#ifndef BZP_DONT_USE_MOVE
	case TOOL_MOVE:
		toolObj=static_cast<ToolBase*>(new ToolMove);
		break;
#endif

#ifndef BZP_DONT_USE_INSERTKNOT
	case TOOL_INSERTKNOT:
		toolObj=static_cast<ToolBase*>(new ToolInsertKnot);
		break;
#endif

#ifndef BZP_DONT_USE_DELETEPOINT
	case TOOL_DELETEPOINT:
		toolObj=static_cast<ToolBase*>(new ToolDeletePoint);
		break;
#endif

#ifndef BZP_DONT_USE_CHANGECONTROL
	case TOOL_CHANGECONTROL:
		toolObj=static_cast<ToolBase*>(new ToolChangeControl);
		break;
#endif

#ifndef BZP_DONT_USE_CUT
	case TOOL_CUT:
		toolObj=static_cast<ToolBase*>(new ToolCut);
		break;
#endif

#ifndef BZP_DONT_USE_SHAPE
	case TOOL_SHAPE:
		toolObj=static_cast<ToolBase*>(new ToolShape);
		break;
#endif

#ifndef BZP_DONT_USE_SELECTKNOTS
	case TOOL_SELECTKNOTS:
		toolObj=static_cast<ToolBase*>(new ToolSelectKnots);
		break;
#endif

	default:
		assert(false);
		break;
	}

	return toolObj;
}

//pX̕ύX̃tOɖ߂
void ToolHolder::ClearModifiedFlag()
{
	m_toolObj->SetModifiedFlag(false);
}

//pXύXꂽ\邩
bool ToolHolder::DidModify() const
{
	return m_toolObj->DidModify();
}

//c[̐ݒ𓾂
int ToolHolder::GetToolOption(int lowBit,int hiBit) const
{
	if(lowBit<0) {
		lowBit=0;
	}
	if(hiBit<0) {
		hiBit=32;	//int^̃rbg
	}

	BZP_PARAMETER_ERROR(
		(0<=lowBit)
		&& (lowBit<hiBit) 
		&& ((1<<hiBit)-1<=INT_MAX)
		,BZPERROR_OUT_OF_RANGE);

	return m_toolObj->ExtractOption(lowBit,hiBit);
}

BZP_NAMESPACE_END
