#include "prefix.h"

#include <new>

#include "toolmove.h"
#include "selectmode.h"
#include "path.h"
#include "part.h"
#include "runtimeconfig.h"
#include "findresult.h"
#include "rect.h"
#include "movecurvemode.h"
#include "event.h"
#include "partmask.h"
#include "knot.h"
#include "graph.h"
#include "fpoint.h"

/*************************************************

	ړc[

**************************************************/

BZP_NAMESPACE_START

BZP_PARTMASK_USING

enum
{
	PHASE_IDLE,	//ĂȂ
	PHASE_GRIPPED,	//iNbNĂ
	PHASE_SELECT_AND_GRIPPED,	//IiNbNĂ顃hbOȂΗ
	PHASE_DRAG,	//ihbOĂ
	PHASE_MOVE,	//ĩNbN߂Ĉړ

	//`őI
	PHASE_PUTSELECTBOX,	//In߂
	PHASE_DRAGSELECTBOX,	//hbOőIL
	PHASE_MOVESELECTBOX,	//xNbN܂őIL
};

class ToolMoveMember
{
public:
	FPoint grippedPoint;	//񂾏ꏊ
	FPoint endCorner;	//`ȈI_
	Knot originKnot;	//[񂾏ꍇ̌̌pڂ̏
};


ToolMove::ToolMove()
	:m_m(new ToolMoveMember)
	,m_phase(PHASE_IDLE)
{
	BZP_BAD_ALLOC_TEST(m_m.get());

	m_condition=COND_MOVE_IDLE;
}

ToolMove::~ToolMove() throw()
{

}

//쒆
bool ToolMove::IsOperating() const
{
	return m_phase!=PHASE_IDLE;
}

//}EXLv`[邩
bool ToolMove::IsCapturing() const
{
	return m_phase!=PHASE_IDLE;
}

//ԂƂɃCxgU蕪
void ToolMove::Event(EVENT event,const FPoint & point,int)
{
	switch(m_phase) {
	case PHASE_IDLE:	//ĂȂ
		EventIdle(event,point);
		break;

	case PHASE_GRIPPED:	//iNbNĂ
		EventGripped(event,point);
		break;

	case PHASE_SELECT_AND_GRIPPED:	//IiNbNĂ
		EventSelectAndGripped(event,point);
		break;

	case PHASE_DRAG:	//ihbOĂ
		EventDrag(event,point);
		break;

	case PHASE_MOVE:	//ĩNbN߂Ĉړ
		EventMove(event,point);
		break;

	case PHASE_PUTSELECTBOX:	//In߂
		EventPutSelectBox(event,point);
		break;

	case PHASE_DRAGSELECTBOX:	//hbOőIL
		EventDragSelectBox(event,point);
		break;

	case PHASE_MOVESELECTBOX:	//xNbN܂őIL
		EventMoveSelectBox(event,point);
		break;
	}
}

//ĂȂԂɂ
void ToolMove::Idle(const FPoint & point)
{
	m_phase=PHASE_IDLE;
	Idle_Move(point);
}

//ĂȂԂ̃Cxg
void ToolMove::EventIdle(EVENT event,const FPoint & point)
{
	switch(event) {
	case EVENT_DRAW:
		OnDraw();
		break;

	case EVENT_LEFTDOWN:
		Start(point);
		break;

	case EVENT_MOUSEMOVE:
		Idle_Move(point);
		break;
	}
}

//ĂȂԂŃJ[\
bool ToolMove::Idle_Move(const FPoint & point)
{
	FindResult findResult(point);
	m_path->Find(&findResult,m_graph,PART_ALL);

	switch(findResult.GeneralPart()) {
	case PART_ANCHOR:
		m_condition=COND_MOVE_FIND_ANCHOR;
		break;

	case PART_CONTROL:
		m_condition=COND_MOVE_FIND_CONTROL;
		break;

	case PART_CURVE:
		m_condition=COND_MOVE_FIND_CURVE;
		break;

	case PART_SUBGROUP:
		m_condition=COND_MOVE_FIND_SUBGROUP;
		break;

	default:
		m_condition=COND_MOVE_IDLE;
		return false;
	}
	return true;
}

//c[̓n߂
void ToolMove::Start(const FPoint & point)
{
	m_m->grippedPoint=point;

	FindResult findResult(point);
	if(m_path->Find(&findResult,m_graph,PART_ALL)) {
		m_grippedPart=findResult.GetPart();
		m_grippedKnot=findResult.GetKnotIndex();
		m_grippedCurve=findResult.GetCurveIndex();
		m_grippedSubGroup=findResult.GetSubGroupIndex();

		Grip(point,findResult.IsSelected());
		switch(GeneralPart(m_grippedPart)) {
		case PART_ANCHOR:
			m_condition=COND_MOVE_KNOT;
			break;

		case PART_CONTROL:
			m_condition=COND_MOVE_CONTROL;
			break;

		case PART_CURVE:
			m_condition=COND_MOVE_CURVE;
			break;

		case PART_SUBGROUP:
			m_condition=COND_MOVE_SUBGROUP;
			break;
		}
		SetModifiedFlag();
	} else {
		PutSelectBox();
	}
}

//i
void ToolMove::Grip(const FPoint & point,bool grippedIsSelected)
{
	bool isSelectedBefore=grippedIsSelected;
	if(ContainControl(m_grippedPart)) {
		isSelectedBefore=true;
	}
	bool isSelectedAfter=true;
	SELECTMODE selectMode=SelectMode();

	if(!isSelectedBefore || selectMode==SELECTMODE_TOGGLE) {
		switch(GeneralPart(m_grippedPart)) {
		case PART_ANCHOR:
			if(ContainEndAnchor(m_grippedPart)
				&& !m_path->IsSelectedCurve(m_grippedCurve)) {

				m_path->SelectCurve(selectMode,m_grippedCurve);
			}
			m_path->SelectKnot(selectMode,m_grippedKnot,m_grippedCurve);
			isSelectedAfter=m_path->IsSelectedKnot(m_grippedKnot,m_grippedCurve);
			break;

		case PART_CURVE:
			m_path->SelectCurve(selectMode,m_grippedCurve);
			isSelectedAfter=m_path->IsSelectedCurve(m_grippedCurve);
			break;

		case PART_SUBGROUP:
			m_path->SelectSubGroup(selectMode,m_grippedSubGroup);
			isSelectedAfter=m_path->IsSelectedSubGroup(m_grippedSubGroup);
			break;
		}
	}

	if(ContainEndAnchor(m_grippedPart)) {
		m_path->CopyKnotTo(&(m_m->originKnot),m_grippedKnot,m_grippedCurve);
		m_m->originKnot.Select(true);
	}

	if(MoveMode()==OPTION_MOVE_DRAG && isSelectedAfter) {
		Drag(point);
	} else if(MoveMode()==OPTION_MOVE_NODRAG && isSelectedBefore && isSelectedAfter) {
		Move(point);
	} else if(MoveMode()==OPTION_MOVE_2WAY && isSelectedAfter) {
		if(isSelectedBefore) {
			m_phase=PHASE_GRIPPED;
		} else {
			SelectAndGrip();
		}
	} else {
		Idle(point);
	}
}

//i񂾏Ԃ̃Cxg
void ToolMove::EventGripped(EVENT event,const FPoint & point)
{
	switch(event) {
	case EVENT_DRAW:
		OnDraw();
		break;

	case EVENT_LEFTUP:
		if(MoveMode()!=OPTION_MOVE_DRAG) {
			Move(point);
		}
		break;

	case EVENT_DRAG:
		if(MoveMode()!=OPTION_MOVE_NODRAG) {
			Drag(point);
		}
		break;

	case EVENT_CANCEL:
		Idle(point);
		break;
	}
}

//IĂȂiőI
void ToolMove::SelectAndGrip()
{
	m_phase=PHASE_SELECT_AND_GRIPPED;
}

//IĂȂiőIԂ̃Cxg
void ToolMove::EventSelectAndGripped(EVENT event,const FPoint & point)
{
	switch(event) {
	case EVENT_DRAW:
		OnDraw();
		break;

	case EVENT_DRAG:
		Drag(point);
		break;

	case EVENT_LEFTUP:
	case EVENT_CANCEL:
		Idle(point);
		break;
	}
}

//xNbN܂ňړ
void ToolMove::Move(const FPoint & point)
{
	m_phase=PHASE_MOVE;
	MoveParts(point);
}

//ړ̃Cxg
void ToolMove::EventMove(EVENT event,const FPoint & point)
{
	switch(event) {
	case EVENT_DRAW:
		OnDraw();
		break;

	case EVENT_MOUSEMOVE:
		MoveParts(point);
		break;
	
	case EVENT_LEFTDOWN:
		Release(point);
		break;

	case EVENT_CANCEL:
		Idle(point);
		break;
	}
}

//i𗣂
void ToolMove::Release(const FPoint & point)
{
	if(ContainEndAnchor(m_grippedPart)) {
		FindResult findResult(point);
		SetIgnoreKnot(&findResult);
		if(m_path->Find(&findResult,m_graph,PART_ENDANCHOR|PART_UNSELECTED)) {
			ChainFindKnotAndGrippedKnot(findResult);
		}
	}
	Idle(point);
}

//iړ
void ToolMove::MoveParts(const FPoint & point)
{
	FPoint point_=point;
	int subGroupMode=MoveSubGroupMode();

	if(ContainAnchor(m_grippedPart) 
		|| (ContainSubGroup(m_grippedPart) && subGroupMode==OPTION_MOVE_SUBGROUP_WITH_KNOT)) {
		MoveAnchor(&point_);
	}

	if(ContainControl(m_grippedPart)) {
		MoveControl(&point_);
	}

	if(ContainCurve(m_grippedPart)
		|| (ContainSubGroup(m_grippedPart) && subGroupMode==OPTION_MOVE_SUBGROUP_WITH_CURVE && MoveCurveMode()==OPTION_MOVE_CURVE_ALLSELECTED)) {
		MoveCurve(point_);
	}

	if(ContainSubGroup(m_grippedPart)
		|| (ContainAnchor(m_grippedPart) && subGroupMode==OPTION_MOVE_SUBGROUP_WITH_KNOT)
		|| (ContainCurve(m_grippedPart) && subGroupMode==OPTION_MOVE_SUBGROUP_WITH_CURVE)) {
		MoveSubGroup(point_);
	}
	m_m->grippedPoint=point_;
}

//hbOňړ
void ToolMove::Drag(const FPoint & point)
{
	m_phase=PHASE_DRAG;
	MoveParts(point);
}

//hbOňړԂ̃Cxg
void ToolMove::EventDrag(EVENT event,const FPoint & point)
{
	switch(event) {
	case EVENT_DRAW:
		OnDraw();
		break;

	case EVENT_MOUSEMOVE:
		Drag(point);
		break;

	case EVENT_LEFTUP:
		Release(point);
		break;

	case EVENT_CANCEL:
		Idle(point);
		break;
	}
}

//`In߂
void ToolMove::PutSelectBox()
{
	m_condition=COND_SELECTBOX;
	if(MoveMode()==OPTION_MOVE_DRAG) {
		m_phase=PHASE_DRAGSELECTBOX;
	} else if(MoveMode()==OPTION_MOVE_NODRAG) {
		m_phase=PHASE_MOVESELECTBOX;
	} else {
		m_phase=PHASE_PUTSELECTBOX;
	}

	m_m->endCorner=m_m->grippedPoint;

	DrawSelectBox();
}

//`In߂Ԃ̃Cxg
void ToolMove::EventPutSelectBox(EVENT event,const FPoint & point)
{
	switch(event) {
	case EVENT_DRAW:
		OnDraw();
		break;

	case EVENT_LEFTUP:
		if(MoveMode()!=OPTION_MOVE_DRAG) {
			MoveSelectBox(point);
		} 
		break;

	case EVENT_DRAG:
		if(MoveMode()!=OPTION_MOVE_NODRAG) {
			DragSelectBox(point);
		}
		break;

	case EVENT_CANCEL:
		Idle(point);
		break;
	}
}

//xNbN܂őILԂɂ
void ToolMove::MoveSelectBox(const FPoint & point)
{
	m_phase=PHASE_MOVESELECTBOX;	
	SelectBox_Move(point);
}

//ǂxNbN܂őILԂ̃Cxg
void ToolMove::EventMoveSelectBox(EVENT event,const FPoint & point)
{
	switch(event) {
	case EVENT_DRAW:
		OnDraw();
		break;

	case EVENT_MOUSEMOVE:
		SelectBox_Move(point);
		break;

	case EVENT_LEFTDOWN:
		EndSelectBox(point);
		break;

	case EVENT_CANCEL:
		DrawSelectBox();
		Idle(point);
		break;
	}
}

//{^𗣂܂őIL悤ɂ
void ToolMove::DragSelectBox(const FPoint & point)
{
	m_phase=PHASE_DRAGSELECTBOX;
	SelectBox_Move(point);
}

//{^𗣂܂őILԂ̃Cxg
void ToolMove::EventDragSelectBox(EVENT event,const FPoint & point)
{
	switch(event) {
	case EVENT_DRAW:
		OnDraw();
		break;

	case EVENT_MOUSEMOVE:
		SelectBox_Move(point);
		break;

	case EVENT_LEFTUP:
		EndSelectBox(point);
		break;

	case EVENT_CANCEL:
		DrawSelectBox();
		Idle(point);
		break;
	}
}

//IL
void ToolMove::SelectBox_Move(const FPoint & point)
{
	DrawSelectBox();
	m_m->endCorner=point;
	DrawSelectBox();
}

//II
void ToolMove::EndSelectBox(const FPoint & point)
{
	DrawSelectBox();

	FRect rect;
	rect.Set(m_m->grippedPoint,m_m->endCorner);
	m_path->SelectBox(SelectMode(),rect);
	Idle(point);
	SetModifiedFlag();
}

//`I\
void ToolMove::DrawSelectBox()
{
	if(m_phase==PHASE_MOVESELECTBOX || m_phase==PHASE_DRAGSELECTBOX) {
		FRect rect;
		rect.Set(m_m->grippedPoint,m_m->endCorner);
		m_graph->DrawSelectShape(rect,false,true);
	}
}

//IvV́AhbO邩Ȃ̐ݒԂ
int ToolMove::MoveMode() const
{
	return ExtractOption(MOVE_MOVE_LOW,MOVE_MOVE_HI);
}

//IvṼTuO[v̈ړ@Ԃ
int ToolMove::MoveSubGroupMode() const
{
	return ExtractOption(MOVE_SUBGROUP_LOW,MOVE_SUBGROUP_HI);
}

//IvV̐̈ړ@Ԃ
MOVECURVEMODE ToolMove::MoveCurveMode() const
{
	return static_cast<MOVECURVEMODE>(ExtractOption(MOVE_CURVE_LOW,MOVE_CURVE_HI)>>MOVE_CURVE_LOW);
}

//IvV̑IɊւ镔Ԃ
SELECTMODE ToolMove::SelectMode() const
{
	int option=ExtractOption(MOVE_SELECT_LOW,MOVE_SELECT_HI);

	return static_cast<SELECTMODE>(option>>MOVE_SELECT_LOW);
}

//TOpڂ̐ݒ
void ToolMove::SetIgnoreKnot(FindResult *findResult) const
{
	findResult->SetIgnoreKnot(m_grippedKnot,m_grippedCurve);
}

//̖[𑼂̖[ɂȂ
void ToolMove::ChainFindKnotAndGrippedKnot(const FindResult &findResult)
{
	int findKnot=findResult.GetKnotIndex();
	int findCurve=findResult.GetCurveIndex();

	bool grippedForward=ContainLastAnchor(m_grippedPart);

	m_path->Chain(&m_grippedKnot,&m_grippedCurve,grippedForward,findKnot,findCurve,true);
}

//ʉߓ_ňړ
void ToolMove::MoveAnchor(FPoint *point)
{
	if(ContainEndAnchor(m_grippedPart)) {
		if(IsPushingKey(KEY_MOVE_FIXANGLE)) {
			int symbol=m_path->SymbolizeKnotIndex(m_grippedKnot,m_grippedCurve);
			int secondEndKnot;
			if(symbol==INDEX_FIRST) {
				secondEndKnot=INDEX_SECOND;
			} else if(symbol==INDEX_LAST) {
				secondEndKnot=INDEX_SECONDLAST;
			} else {
				secondEndKnot=INDEX_NONE;
			}
			if(secondEndKnot!=INDEX_NONE
				&& !m_path->IsSelectedKnot(secondEndKnot,m_grippedCurve)) {

				FPoint prePoint=m_path->GetPoint(PART_ANCHOR,secondEndKnot,m_grippedCurve);
				point->FixAngle(prePoint);
			}
		}

		m_path->MoveSelectedKnots((*point) -m_m->grippedPoint);

		FindResult findResult(*point);
		SetIgnoreKnot(&findResult);
		if(m_path->Find(&findResult,m_graph,PART_ENDANCHOR|PART_UNSELECTED)) {
			Knot mimicKnot;
			m_path->CopyKnotTo(&mimicKnot,findResult.GetKnotIndex(),findResult.GetCurveIndex());
			if((findResult.IsLastAnchor() && ContainLastAnchor(m_grippedPart))
				|| (!findResult.IsLastAnchor() && !ContainLastAnchor(m_grippedPart))) {
				mimicKnot.Reverse();
			}
			m_path->CopyKnot(&mimicKnot,m_grippedKnot,m_grippedCurve);

			if(m_path->IsSameCurve(findResult.GetCurveIndex(),m_grippedCurve)) {
				m_condition=COND_MOVE_CLOSE;
			} else {
				m_condition=COND_MOVE_CHAIN;
			}
		} else {
			m_m->originKnot.MovePoint(*point,PART_ANCHOR);
			m_path->CopyKnot(&(m_m->originKnot),m_grippedKnot,m_grippedCurve);

			m_condition=COND_MOVE_KNOT;
		}
	} else {
		m_path->MoveSelectedKnots((*point)-m_m->grippedPoint);
	}
}

//_ړ
void ToolMove::MoveControl(FPoint *point)
{
	if(IsPushingKey(KEY_MOVE_FIXANGLE)) {
		FPoint anchorPoint=m_path->GetPoint(PART_ANCHOR,m_grippedKnot,m_grippedCurve);
		point->FixAngle(anchorPoint);
	}
	m_path->MovePoint(*point,m_grippedPart,m_grippedKnot,m_grippedCurve);
}

//ړ
void ToolMove::MoveCurve(const FPoint &point)
{
	m_path->MoveCurve(point,m_m->grippedPoint,MoveCurveMode(),m_grippedKnot,m_grippedCurve);
}

//TuO[vړ
void ToolMove::MoveSubGroup(const FPoint &point)
{
	m_path->MoveSelectedSubGroups(point-m_m->grippedPoint);
}

//ToolExtendNXĂяoł镔ȉݒ肷
void ToolMove::SetGripPart(const FPoint & point,int part,int knotIndex,int curveIndex)
{
	m_m->grippedPoint=point;
	m_grippedPart=part;
	m_grippedKnot=knotIndex;
	m_grippedCurve=curveIndex;
}

BZP_NAMESPACE_END
