#include "prefix.h"

#include <new>

#include "toolpen.h"
#include "selectmode.h"
#include "path.h"
#include "part.h"
#include "runtimeconfig.h"
#include "graph.h"
#include "findresult.h"
#include "event.h"
#include "fpoint.h"

/*************************************************

	yc[

**************************************************/

BZP_NAMESPACE_START

enum
{
	PHASE_IDLE,	//ĂȂ
	PHASE_PUTKNOT,	//pڂuĂ̂܂
	PHASE_CHAINKNOT,	//̐ƂȂ
	PHASE_MOVE,	//ړ
	PHASE_MOVEWCONTROLS,	//L΂Ă顃NbNƏI
	PHASE_DRAGWCONTROLS,	//L΂Ă顃{^𗣂ƏI
	PHASE_CHAINDRAGWCONTROLS,	//̐ƂȂĂL΂Ă
	PHASE_TERMINATE,	//NbNƏI
	PHASE_OUTSIDE,	//EChE̊Oɏo
	PHASE_EXTEND,	//ړc[̋@\gĂ
};

class ToolPenMember
{
public:
	FPoint rubberAnchorA;	//o[oh̍W
	FPoint rubberForwardA;
	FPoint rubberBackB;
	FPoint rubberAnchorB;
};

ToolPen::ToolPen()
	:m_m(new ToolPenMember)
	,m_phase(PHASE_IDLE)
{
	BZP_BAD_ALLOC_TEST(m_m.get());
	m_condition=COND_PEN_IDLE;
}

//쒆
bool ToolPen::IsOperating() const
{
	return m_phase!=PHASE_IDLE;
}

//}EXLv`[邩
bool ToolPen::IsCapturing() const
{
	return IsOperating() && m_phase!=PHASE_OUTSIDE && m_phase!=PHASE_TERMINATE;
}

//𒆒fĂ؂̏ԂŁȂȂǂł邩
bool ToolPen::CanStop() const
{
	return !IsCapturing() || m_phase==PHASE_MOVE;
}

//ԂƂɃCxgU蕪
void ToolPen::Event(EVENT event,const FPoint & point,int key)
{
	switch(m_phase) {
	case PHASE_IDLE:
		EventIdle(event,point);
		break;

	case PHASE_PUTKNOT:
		EventPutKnot(event,point);
		break;

	case PHASE_CHAINKNOT:
		EventChainKnot(event,point);
		break;

	case PHASE_MOVE:
		EventMove(event,point,key);
		break;

	case PHASE_MOVEWCONTROLS:
	case PHASE_DRAGWCONTROLS:
	case PHASE_CHAINDRAGWCONTROLS:
		EventMoveWControls(event,point,key);
		break;

	case PHASE_TERMINATE:
		EventTerminate(event,point,key);
		break;

	case PHASE_OUTSIDE:
		EventOutside(event,point);
		break;

	case PHASE_EXTEND:
		EventExtend(event,point,key);
		break;
	}
}

//ĂȂԂɂ
void ToolPen::Idle(const FPoint & point)
{
	m_phase=PHASE_IDLE;
	Idle_Move(point);
}

//ĂȂԂ̃Cxg
void ToolPen::EventIdle(EVENT event,const FPoint & point)
{
	switch(event) {
	case EVENT_DRAW:
		OnDraw();
		break;

	case EVENT_LEFTDOWN:
		Start(point);
		break;

	case EVENT_MOUSEMOVE:
		Idle_Move(point);
		break;
	}
}

//ĂȂԂŃ}EX𓮂
void ToolPen::Idle_Move(const FPoint & point)
{
	m_lastKnot=INDEX_NONE;
	m_lastCurve=INDEX_NONE;
	if(FindEndAnchor(point)) {
		m_condition=COND_PEN_CHAIN;
	} else {
		m_condition=COND_PEN_IDLE;
	}
}

//c[̓n߂
void ToolPen::Start(const FPoint & point)
{
	FindResult findResult(point);
	m_path->Find(&findResult,m_graph,PART_ENDANCHOR);

	SetFirstKnot(findResult,point);
	SetModifiedFlag();
}

//ŏ̌pڂݒ
void ToolPen::SetFirstKnot(const FindResult &findResult,const FPoint & point)
{
	m_phase=PHASE_PUTKNOT;
	m_addCount=0;
	if(findResult.IsEndAnchor()) {
		SetFindKnotToLastKnot(findResult);
	} else {
		SetNewCurveToLastKnot(point);
	}	
}

//pڂǉ
void ToolPen::PutKnot(const FPoint & point)
{
	FPoint point_=point;
	if(IsPushingKey(KEY_PEN_FIXANGLE)) {
		FPoint lastPoint=m_path->GetPoint(PART_ANCHOR,m_lastKnot,m_lastCurve);
		point_.FixAngle(lastPoint);
	}

	m_phase=PHASE_PUTKNOT;

	FindResult findResult(point_);
	FindEndAnchor(&findResult);

	if(findResult.IsEndAnchor()) {
		ChainKnot(findResult);
	} else {
		AddLastKnot(point_);
	}
	SetModifiedFlag();
}

//pڂuԂ̃Cxg
void ToolPen::EventPutKnot(EVENT event,const FPoint & point)
{
	switch(event) {
	case EVENT_DRAW:
		OnDraw();
		break;

	case EVENT_DRAG:
		DragWControls(point);
		break;

	case EVENT_LEFTUP:
		Move(point);
		break;

	case EVENT_CANCEL:
		Idle(point);
		break;
	}
}

//̐ƂȂ
void ToolPen::ChainKnot(const FindResult &findResult)
{
	m_phase=PHASE_CHAINKNOT;
	ChainFindKnotAndLastKnot(findResult);
}

//̐ƂȂԂ̃Cxg
void ToolPen::EventChainKnot(EVENT event,const FPoint & point)
{
	switch(event) {
	case EVENT_DRAW:
		OnDraw();
		break;

	case EVENT_MOUSEMOVE:
		ChainKnot_Move(point);
		break;

	case EVENT_LEFTUP:
	case EVENT_CANCEL:
		Idle(point);
		break;
	}
}

//̐ƂȂĂhbOĕL΂
void ToolPen::ChainKnot_Move(const FPoint & point)
{
	m_phase=PHASE_CHAINDRAGWCONTROLS;
	ToolExtend::DragWControls(point,m_grippedIsForward,m_lastKnot,m_lastCurve);
}

//pڂũhbOŕL΂
void ToolPen::DragWControls(const FPoint & point)
{
	m_phase=PHASE_DRAGWCONTROLS;
	ToolExtend::DragWControls(point,m_grippedIsForward,m_lastKnot,m_lastCurve);
}

//L[͂ɂ؂ւŕL΂
void ToolPen::MoveWControls(const FPoint & point)
{
	m_phase=PHASE_MOVEWCONTROLS;
	ToolExtend::MoveWControls(point,m_grippedIsForward,m_lastKnot,m_lastCurve);
}

//𗼕ɐL΂ĂԂ̃Cxg
void ToolPen::EventMoveWControls(EVENT event,const FPoint & point,int key)
{
	ToolExtend::Event(event,point,key);

	switch(event) {
	case EVENT_KEYUP:
		if(m_phase==PHASE_MOVEWCONTROLS && IsPushedKey(KEY_PEN_MOVEWCONTROLS,key)) {
			Move(point);
		}
		break;

	case EVENT_CANCEL:
		Idle(point);
		break;

	default:
		if(!ToolExtend::IsOperating()) {
			if(m_phase==PHASE_CHAINDRAGWCONTROLS) {
				Idle(point);
			} else {
				Move(point);
			}
		}
		break;
	}
}

//̌pڂł߂Ƀo[oh\Ȃړ
void ToolPen::Move(const FPoint & point)
{
	if(IsPushingKey(KEY_PEN_TERMINATE)) {
		Terminate();
		return;
	}

	if(m_outsideWindow) {
		Outside();
		return;
	}

	FPoint point_=point;
	if(IsPushingKey(KEY_PEN_FIXANGLE)) {
		FPoint lastPoint=m_path->GetPoint(PART_ANCHOR,m_lastKnot,m_lastCurve);
		point_.FixAngle(lastPoint);
	}

	FindResult findResult(point_);
	FindEndAnchor(&findResult);

	if(findResult.IsEndAnchor()) {
		if(m_path->IsSameCurve(m_lastCurve,findResult.GetCurveIndex())) {
			m_condition=COND_PEN_CLOSE;
		} else {
			m_condition=COND_PEN_CHAIN;
		}
	} else {
		m_condition=COND_PEN_MOVE;
	}

	m_phase=PHASE_MOVE;
	SetRubberBand(findResult,point_);
}

//o[oh\ȂړĂԂ̃Cxg
void ToolPen::EventMove(EVENT event,const FPoint & point,int key)
{
	switch(event) {
	case EVENT_DRAW:
		OnDraw();
		break;

	case EVENT_MOUSEMOVE:
		Move(point);
		break;

	case EVENT_LEFTDOWN:
		PutKnot(point);
		break;

	case EVENT_KEYDOWN:
		Move_KeyDown(point,key);
		break;

	case EVENT_KEYUP:
		Move_KeyUp(point,key);
		break;

	case EVENT_MOVEOUTSIDEWINDOW:
		Outside();
		break;

	case EVENT_UNDO:
		Move_Undo(point);
		break;

	case EVENT_REDO:
		Move_Redo(point);
		break;

	case EVENT_RIGHTDOWN:
	case EVENT_CANCEL:
		Idle(point);
		break;
	}
}

//ړĂԂŃL[ꂽ
void ToolPen::Move_KeyDown(const FPoint & point,int key)
{
	if(IsPushedKey(KEY_PEN_TERMINATE,key)) {
		Terminate();
	} else if(IsPushedKey(KEY_PEN_FIXANGLE,key)) {
		Move(point);
	}
}

//ړĂԂŃL[ꂽ
void ToolPen::Move_KeyUp(const FPoint & point,int key)
{
	if(IsPushedKey(KEY_PEN_MOVEWCONTROLS,key)) {
		MoveWControls(point);
	} else {
		Move(point);
	}
}

//AvP[VŌɖ߂삪sꂽ
void ToolPen::Move_Undo(const FPoint & point)
{
	//ǉpڂ̍폜͗̊Ǘ̂قł
	assert(m_addCount>0);

	if(m_lastKnot!=INDEX_FIRST) {
		m_lastKnot=INDEX_LAST;
	}

	m_addCount--;
	if(m_addCount==0) {
		Idle(point);
	} else {
		Move(point);
	}
}

//AvP[Vł蒼삪sꂽ
void ToolPen::Move_Redo(const FPoint & point)
{
	//pڂ̒ǉ̂蒼͗̊Ǘ̂قł
	if(m_lastKnot!=INDEX_FIRST) {
		m_lastKnot=INDEX_LAST;
	}

	++m_addCount;
	Move(point);
}

//NbNƏIԂɂ
void ToolPen::Terminate()
{
	m_phase=PHASE_TERMINATE;
	m_condition=COND_PEN_TERMINATE;
}

//NbNƏIɂȂԂ̃Cxg
void ToolPen::EventTerminate(EVENT event,const FPoint & point,int key)
{
	switch(event) {
	case EVENT_DRAW:
		OnDraw();
		break;

	case EVENT_MOUSEMOVE:
		if(!ToolExtend::Idle_Move(point)) {
			m_condition=COND_PEN_TERMINATE;
		}
		break;

	case EVENT_KEYUP:
		if(IsPushedKey(KEY_PEN_TERMINATE,key)) {
			if(ToolExtend::IsOperating()) {
				ToolExtend::Idle(point);
			}
			Move(point);
		}
		break;

	case EVENT_LEFTDOWN:
		Terminate_LeftDown(point);
		break;

	case EVENT_UNDO:
		Move_Undo(point);
		break;

	case EVENT_REDO:
		Move_Redo(point);
		break;

	case EVENT_CANCEL:
		if(ToolExtend::IsOperating()) {
			ToolExtend::Idle(point);
		}
		Idle(point);
		break;
	}
}

//NbNƏIԂōNbN
void ToolPen::Terminate_LeftDown(const FPoint & point)
{
	FindResult findResult(point);
	if(!m_path->Find(&findResult,m_graph,PART_ALL)) {
		Idle(point);
	} else {
		m_terminate_curveCount=m_path->CurveCount();
		ToolExtend::Start(point);
		m_phase=PHASE_EXTEND;
	}
}

//ړc[̃Cxg
void ToolPen::EventExtend(EVENT event,const FPoint & point,int key)
{
	ToolExtend::Event(event,point,key);
	if(event==EVENT_CANCEL
		|| m_terminate_curveCount!=m_path->CurveCount()
		|| m_path->IsClosedCurve(m_lastCurve)) {
		Idle(point);
	} else if(!ToolExtend::IsOperating()) {
		Terminate();
	}
}

//ړ̏ԂŃJ[\EChE̊Oɏo
void ToolPen::Outside()
{
	m_phase=PHASE_OUTSIDE;
	m_condition=COND_PEN_OUTSIDE;
}

//J[\OɏoĂԂ̃Cxg
void ToolPen::EventOutside(EVENT event,const FPoint & point)
{
	switch(event) {
	case EVENT_DRAW:
		OnDraw();
		break;

	case EVENT_MOVEINSIDEWINDOW:
		Move(point);
		break;

	case EVENT_UNDO:
		Move_Undo(point);
		break;

	case EVENT_REDO:
		Move_Redo(point);
		break;

	case EVENT_CANCEL:
		Idle(point);
		break;
	}
}

//\
void ToolPen::OnDraw()
{
	ToolBase::OnDraw();
	DrawRubberBand();
}

////////////////////////////////////////

//̐[T
bool ToolPen::FindEndAnchor(const FPoint & point) const
{
	FindResult findResult(point);
	SetIgnoreKnot(&findResult);
	return m_path->Find(&findResult,m_graph,PART_ENDANCHOR);
}

//̐[T
bool ToolPen::FindEndAnchor(FindResult *findResult) const
{
	SetIgnoreKnot(findResult);
	return m_path->Find(findResult,m_graph,PART_ENDANCHOR);
}

//Ȃpڂݒ肷
void ToolPen::SetIgnoreKnot(FindResult *findResult) const
{
	findResult->SetIgnoreKnot(m_lastKnot,m_lastCurve);
}


//ق̒[_pڂɐݒ肷
void ToolPen::SetFindKnotToLastKnot(const FindResult &findResult)
{
	m_lastCurve=findResult.GetCurveIndex();
	m_lastKnot=findResult.GetKnotIndex();
	m_path->SelectCurve(SELECTMODE_SINGLE,m_lastCurve);
	m_path->SelectKnot(SELECTMODE_SINGLE,m_lastKnot,m_lastCurve);
	if(m_path->SymbolizeKnotIndex(m_lastKnot,m_lastCurve)==INDEX_FIRST) {
		m_grippedIsForward=false;
	} else {
		m_grippedIsForward=true;
	}
}

//VɐƌpڂǉĂpڂɐݒ肷
void ToolPen::SetNewCurveToLastKnot(const FPoint & point)
{
	m_addCount=1;
	m_lastCurve=m_path->NewCurve(INDEX_LAST);
	m_path->SelectCurve(SELECTMODE_SINGLE,INDEX_LAST);
	m_lastKnot=m_path->AddEndKnot(point,INDEX_LAST,INDEX_LAST);
	m_grippedIsForward=true;
}

//[ƂȂ
void ToolPen::ChainFindKnotAndLastKnot(const FindResult &findResult)
{
	m_path->Chain(&m_lastKnot,&m_lastCurve,m_grippedIsForward
		,findResult.GetKnotIndex()
		,findResult.GetCurveIndex(),false);
	m_path->SelectKnot(SELECTMODE_SINGLE,m_lastKnot,m_lastCurve);
}

//Čpڂǉ
void ToolPen::AddLastKnot(const FPoint & point)
{
	++m_addCount;
	m_lastKnot=m_path->AddEndKnot(point,m_lastKnot,m_lastCurve);
	m_path->SelectKnot(SELECTMODE_SINGLE,m_lastKnot,m_lastCurve);
}

//o[oh̏Ԃݒ
void ToolPen::SetRubberBand(const FindResult &findResult,const FPoint & point)
{
	m_m->rubberAnchorA=m_path->GetPoint(PART_ANCHOR,m_lastKnot,m_lastCurve);
	int controlPart=PART_FORWARDCONTROL;
	if(m_path->SymbolizeKnotIndex(m_lastKnot,m_lastCurve)==INDEX_FIRST) {
		controlPart=PART_BACKCONTROL;
	}
	m_m->rubberForwardA=m_path->GetPoint(controlPart,m_lastKnot,m_lastCurve);

	if(findResult.IsEndAnchor()) {
		RubberChainFindKnot(findResult);
	} else {
		RubberNoChain(point);
	}
}

//pڂɃo[ohȂ
void ToolPen::RubberChainFindKnot(const FindResult &findResult)
{
	int findKnot=findResult.GetKnotIndex();
	int findCurve=findResult.GetCurveIndex();

	int controlPart=PART_FORWARDCONTROL;
	if(m_path->SymbolizeKnotIndex(findKnot,findCurve)==INDEX_FIRST) {
		controlPart=PART_BACKCONTROL;
	}

	m_m->rubberBackB=m_path->GetPoint(controlPart,findKnot,findCurve);
	m_m->rubberAnchorB=m_path->GetPoint(PART_ANCHOR,findKnot,findCurve);
}

//o[oh̐[𑼂̌pڂɂȂȂ
void ToolPen::RubberNoChain(const FPoint & point)
{
	m_m->rubberBackB=point;
	m_m->rubberAnchorB=point;
}

//o[oh\
void ToolPen::DrawRubberBand() const
{
	if(m_phase==PHASE_MOVE && DoesDrawRubberBand()) {
		m_graph->DrawCurve(m_m->rubberAnchorA,m_m->rubberForwardA
			,m_m->rubberBackB,m_m->rubberAnchorB);
		m_graph->DrawAnchor(m_m->rubberAnchorB,false);
	}
}

//o[oh\邩
bool ToolPen::DoesDrawRubberBand() const
{
	return ExtractOption(PEN_RUBBER_LOW,PEN_RUBBER_HI)==OPTION_PEN_RUBBER;
}

BZP_NAMESPACE_END
