#include "prefix.h"

#include "toolselectknots.h"
#include "fpoint.h"
#include "path.h"
#include "event.h"
#include "graph.h"
#include "index.h"
#include "selectmode.h"
#include "findresult.h"

/*************************************************

	̑SĂ̌pڂ̑I̐؂ւ

**************************************************/

BZP_NAMESPACE_START

enum
{
	PHASE_IDLE,	//ĂȂ
	PHASE_EXTEND,	//ړc[̏Ă
};

ToolSelectKnots::ToolSelectKnots()
	:m_phase(PHASE_IDLE)
{
	m_condition=COND_SELECTKNOTS_IDLE;
}

ToolSelectKnots::~ToolSelectKnots() throw()
{

}

//쒆
bool ToolSelectKnots::IsOperating() const
{
	return m_phase==PHASE_EXTEND && ToolExtend::IsOperating();
}

//}EXLv`[邩
bool ToolSelectKnots::IsCapturing() const
{
	return m_phase==PHASE_EXTEND && ToolExtend::IsCapturing();
}

//CxgU蕪
void ToolSelectKnots::Event(EVENT event,const FPoint & point,int key)
{
	switch(m_phase) {
	case PHASE_IDLE:
		EventIdle(event,point);
		break;

	case PHASE_EXTEND:
		EventExtend(event,point,key);
		break;
	}
}

void ToolSelectKnots::Idle(const FPoint &point)
{
	m_phase=PHASE_IDLE;
	Idle_Move(point);
}

void ToolSelectKnots::EventIdle(EVENT event,const FPoint &point)
{
	switch(event) {
	case EVENT_DRAW:
		OnDraw();
		break;

	case EVENT_MOUSEMOVE:
		Idle_Move(point);
		break;

	case EVENT_LEFTDOWN:
		Start(point);
		break;
	}
}

//}EẌړ
void ToolSelectKnots::Idle_Move(const FPoint &point)
{
	int findPart=PART_CURVE;
	if(DoesMoveAnchor()) {
		findPart=PART_CURVE|PART_KNOT;
	}

	FindResult findResult(point);
	m_path->Find(&findResult,m_graph,findPart);

	if(findResult.GeneralPart()==PART_CURVE) {
		m_condition=SelectionCondition(findResult);
	} else if(!ToolExtend::Idle_Move(point)) {
		m_condition=COND_SELECTKNOTS_IDLE;
	}
}

//̃J[\̈ʒuŃNbNꍇ̏
int ToolSelectKnots::SelectionCondition(const FindResult &findResult)
{
	bool isFirstKnotSelected
		=m_path->IsSelectedKnot(INDEX_FIRST,findResult.GetCurveIndex());

	int condition;
	switch(SelectMode()) {
	case OPTION_SELECTKNOTS_TOGGLE:
		if(isFirstKnotSelected) {
			condition=COND_SELECTKNOTS_UNSELECT;
		} else {
			condition=COND_SELECTKNOTS_SELECT;
		}
		break;

	case OPTION_SELECTKNOTS_SELECT:
		condition=COND_SELECTKNOTS_SELECT;
		break;

	case OPTION_SELECTKNOTS_UNSELECT:
		condition=COND_SELECTKNOTS_UNSELECT;
		break;

	case OPTION_SELECTKNOTS_ALTERNATE:
		condition=COND_SELECTKNOTS_ALTERNATE;
		break;

	default:
		assert(false);
	}
	return condition;
}

//}EX̍{^
void ToolSelectKnots::Start(const FPoint & point)
{
	int findPart=PART_CURVE;
	if(DoesMoveAnchor()) {
		findPart=PART_CURVE|PART_KNOT;
	}
	FindResult findResult(point);
	m_path->Find(&findResult,m_graph,findPart);

	if(findResult.IsCurve()) {
		int condition=SelectionCondition(findResult);

		int curveIndex=findResult.GetCurveIndex();
		m_path->SelectCurve(SELECTMODE_ADD,curveIndex);

		switch(condition) {
		case COND_SELECTKNOTS_SELECT:
			m_path->SelectKnot(SELECTMODE_ALL,INDEX_NONE,curveIndex);
			break;
			
		case COND_SELECTKNOTS_UNSELECT:
			m_path->SelectKnot(SELECTMODE_UNSELECTALL,INDEX_NONE,curveIndex);
			break;

		case COND_SELECTKNOTS_ALTERNATE:
			AlternativeSelect(curveIndex);
			break;

		default:
			assert(false);
		}
		SetModifiedFlag();
	} else {
		ToolExtend::Start(point);
		m_phase=PHASE_EXTEND;
	}
}

//PɑI
void ToolSelectKnots::AlternativeSelect(int curveIndex)
{
	int knotCount=m_path->KnotCount(curveIndex);	
	bool select=!m_path->IsSelectedKnot(INDEX_FIRST,curveIndex);
	FOR(int i=0;i<knotCount;++i) {
		SELECTMODE selectMode;
		if(select) {
			selectMode=SELECTMODE_ADD;
		} else {
			selectMode=SELECTMODE_UNSELECT;
		}
		m_path->SelectKnot(selectMode,i,curveIndex);
		select=!select;
	}
}

//ړc[̃Cxg
void ToolSelectKnots::EventExtend(EVENT event,const FPoint &point,int key)
{
	ToolExtend::Event(event,point,key);
	if(event==EVENT_CANCEL || !ToolExtend::IsOperating()) {
		Idle(point);
	}
}

//ʉߓ_FĈړ邩
bool ToolSelectKnots::DoesMoveAnchor() const
{
	int option=ExtractOption(SELECTKNOTS_ANCHOR_LOW,SELECTKNOTS_ANCHOR_HI);
	return option==OPTION_SELECTKNOTS_MOVE_ANCHOR;
}

//I̐ݒ𓾂
int ToolSelectKnots::SelectMode() const
{
	return ExtractOption(SELECTKNOTS_SELECT_LOW,SELECTKNOTS_SELECT_HI);
}


BZP_NAMESPACE_END


