#include "prefix.h"

#include <new>

#include "toolshape.h"
#include "fpoint.h"
#include "path.h"
#include "event.h"
#include "graph.h"
#include "index.h"
#include "selectmode.h"
#include "rect.h"
#include "mathdefine.h"
#include "runtimeconfig.h"

/*************************************************

	lpc[

**************************************************/

BZP_NAMESPACE_START

enum
{
	PHASE_IDLE,	//ĂȂ
	PHASE_PUT,	//n_NbN
	PHASE_DRAG,	//hbOňړ
	PHASE_MOVE,	//{^𗣂Ĉړ
};

class ToolShapeMember
{
public:
	FPoint startCorner;	/*n_*/
	FPoint endCorner;	/*I_*/
	FRect drewRect;
};

ToolShape::ToolShape()
	:m_m(new ToolShapeMember)
	,m_phase(PHASE_IDLE)
{
	BZP_BAD_ALLOC_TEST(m_m.get());

	m_condition=COND_SHAPE_IDLE;
}

ToolShape::~ToolShape() throw()
{

}

//쒆
bool ToolShape::IsOperating() const
{
	return m_phase!=PHASE_IDLE;
}

//}EXLv`[邩
bool ToolShape::IsCapturing() const
{
	return m_phase!=PHASE_IDLE;
}

//ԂƂɃCxgU蕪
void ToolShape::Event(EVENT event,const FPoint & point,int)
{
	switch(m_phase) {
	case PHASE_IDLE:
		EventIdle(event,point);
		break;

	case PHASE_PUT:
		EventPut(event);
		break;

	case PHASE_DRAG:
		EventDrag(event,point);
		break;

	case PHASE_MOVE:
		EventMove(event,point);
	}
}

//ĂȂԂɂ
void ToolShape::Idle()
{
	m_phase=PHASE_IDLE;
	m_condition=COND_SHAPE_IDLE;
}

//ĂȂԂ̃Cxg
void ToolShape::EventIdle(EVENT event,const FPoint & point)
{
	switch(event) {
	case EVENT_DRAW:
		OnDraw();
		break;

	case EVENT_LEFTDOWN:
		Start(point);
		break;
	}
}

//ĂȂԂŃ}EX̍{^
void ToolShape::Start(const FPoint & point)
{
	m_condition=COND_SHAPE_MOVE;

	m_m->startCorner=point;
	m_m->endCorner=point;

	switch(MoveMode()) {
	case OPTION_SHAPE_2WAY:
		m_phase=PHASE_PUT;
		break;

	case OPTION_SHAPE_DRAG:
		m_phase=PHASE_DRAG;
		break;

	case OPTION_SHAPE_NODRAG:
		m_phase=PHASE_MOVE;
		break;
	}

	DrawShape();
}

//n_߂Ă̂܂ܓĂȂ
void ToolShape::EventPut(EVENT event)
{
	switch(event) {
	case EVENT_DRAW:
		OnDraw();
		break;

	case EVENT_DRAG:
		m_phase=PHASE_DRAG;
		break;

	case EVENT_LEFTUP:
		m_phase=PHASE_MOVE;
		break;

	case EVENT_CANCEL:
		Idle();
		break;
	}
}

//hbOŏI_߂
void ToolShape::EventDrag(EVENT event,const FPoint &point)
{
	switch(event) {
	case EVENT_DRAW:
		OnDraw();
		break;

	case EVENT_LEFTUP:
		EraseShape();
		End(point);
		break;

	case EVENT_MOUSEMOVE:
		EraseShape();
		m_m->endCorner=point;
		DrawShape();
		break;

	case EVENT_CANCEL:
		EraseShape();
		Idle();
		break;
	}
}

//NbNďI_߂
void ToolShape::EventMove(EVENT event,const FPoint &point)
{
	switch(event) {
	case EVENT_DRAW:
		OnDraw();
		break;

	case EVENT_LEFTDOWN:
		EraseShape();
		End(point);
		break;

	case EVENT_MOUSEMOVE:
		EraseShape();
		m_m->endCorner=point;
		DrawShape();
		break;

	case EVENT_RIGHTDOWN:
	case EVENT_CANCEL:
		EraseShape();
		Idle();
		break;
	}
}

void ToolShape::End(const FPoint &point)
{
	FRect rect;
	CalcRect(&rect,m_m->startCorner,point
		,DoesStartCorner()
		,IsSquare() || IsPushingKey(KEY_SHAPE_SQUARE));

	m_path->NewCurve(INDEX_LAST);
	m_path->SelectCurve(SELECTMODE_SINGLE,INDEX_LAST);

	switch(ShapeMode()) {
	case OPTION_SHAPE_BOX:
		PutBox(rect);
		break;

	case OPTION_SHAPE_CIRCLE:
		PutCircle(rect);
		break;

	default:
		assert(false);
	}

	Idle();

	SetModifiedFlag();
}

void ToolShape::PutBox(const FRect &rect)
{
	m_path->AddEndKnot(rect.TopLeft(),INDEX_LAST,INDEX_LAST);
	m_path->AddEndKnot(rect.BottomLeft(),INDEX_LAST,INDEX_LAST);
	m_path->AddEndKnot(rect.BottomRight(),INDEX_LAST,INDEX_LAST);
	m_path->AddEndKnot(rect.TopRight(),INDEX_LAST,INDEX_LAST);
	m_path->CloseCurve(true,INDEX_LAST);
}

void ToolShape::PutCircle(const FRect &rect)
{
	double horzCenter=(rect.left+rect.right)/2;

	FPoint top(horzCenter,rect.top);
	m_path->AddEndKnot(top,INDEX_LAST,INDEX_LAST);
	double controlWidth=rect.Width()*BZP_BOX_TO_ROUND/2;
	FPoint topControl(horzCenter-controlWidth,rect.top);
	m_path->MoveWControls(topControl,true,INDEX_LAST,INDEX_LAST);

	double vertCenter=(rect.top+rect.bottom)/2;

	FPoint left(rect.left,vertCenter);
	m_path->AddEndKnot(left,INDEX_LAST,INDEX_LAST);
	double controlHeight=rect.Height()*BZP_BOX_TO_ROUND/2;
	FPoint leftControl(rect.left,vertCenter+controlHeight);
	m_path->MoveWControls(leftControl,true,INDEX_LAST,INDEX_LAST);

	FPoint bottom(horzCenter,rect.bottom);
	m_path->AddEndKnot(bottom,INDEX_LAST,INDEX_LAST);
	FPoint bottomControl(horzCenter+controlWidth,rect.bottom);
	m_path->MoveWControls(bottomControl,true,INDEX_LAST,INDEX_LAST);

	FPoint right(rect.right,vertCenter);
	m_path->AddEndKnot(right,INDEX_LAST,INDEX_LAST);
	FPoint rightControl(rect.right,vertCenter-controlHeight);
	m_path->MoveWControls(rightControl,true,INDEX_LAST,INDEX_LAST);

	m_path->CloseCurve(true,INDEX_LAST);
}

//OČ`킷߂ɐ}``
void ToolShape::DrawShape()
{
	CalcRect(&m_m->drewRect,m_m->startCorner,m_m->endCorner
		,DoesStartCorner()
		,IsSquare() || IsPushingKey(KEY_SHAPE_SQUARE));
	switch(ShapeMode()) {
	case OPTION_SHAPE_BOX:
		m_graph->DrawSelectShape(m_m->drewRect,false,false);
		break;

	case OPTION_SHAPE_CIRCLE:
		m_graph->DrawSelectShape(m_m->drewRect,true,false);
		break;

	default:
		assert(false);
	}
}

//DrawShapeŏ}`
void ToolShape::EraseShape()
{
	switch(ShapeMode()) {
	case OPTION_SHAPE_BOX:
		m_graph->DrawSelectShape(m_m->drewRect,false,false);
		break;

	case OPTION_SHAPE_CIRCLE:
		m_graph->DrawSelectShape(m_m->drewRect,true,false);
		break;

	default:
		assert(false);
	}
}

//ړ̐ݒ𓾂
int ToolShape::MoveMode() const
{
	return ExtractOption(SHAPE_MOVE_LOW,SHAPE_MOVE_HI);
}

//n_pɂ邩
bool ToolShape::DoesStartCorner() const
{
	return ExtractOption(SHAPE_START_LOW,SHAPE_START_HI)
		==OPTION_SHAPE_CORNER;
}

//`ɂ邩
bool ToolShape::IsSquare() const
{
	return ExtractOption(SHAPE_SQUARE_LOW,SHAPE_SQUARE_HI)
		==OPTION_SHAPE_SQUARE;
}

int ToolShape::ShapeMode() const
{
	return ExtractOption(SHAPE_SHAPE_LOW,SHAPE_SHAPE_HI);
}

//lp̍WvZ
void ToolShape::CalcRect(FRect *rect,const FPoint &start,const FPoint &end,bool doesStartCorner,bool isSquare) const
{
	FPoint cornerB=end;
	if(isSquare) {
		double width=abs(start.x-end.x);
		double height=abs(start.y-end.y);
		if(width>=height) {
			if(start.y<=cornerB.y) {
				cornerB.y=start.y+width;
			} else {
				cornerB.y=start.y-width;
			}
		} else {
			if(start.x<=cornerB.x) {
				cornerB.x=start.x+height;
			} else {
				cornerB.x=start.x-height;
			}
		}
	}

	FPoint cornerA=start;
	
	if(!doesStartCorner) {
		cornerA-=cornerB-start;
	}

	rect->Set(cornerA,cornerB);
	rect->Normalize();
}

BZP_NAMESPACE_END


