#include "prefix.h"

#include "winman.h"
#include "frame.h"
#include "document.h"
#include "view.h"
#include "toolbar.h"
#include "resource.h"
#include "command.h"
#include "winmessage.h"
#include "main.h"

/*************************************************

	OgEBhENX

**************************************************/

SAMPLE_NAMESPACE_START

#define STATUSBARID 0xe801
#define TOOLBARID	0xe800

//c[o[̐ݒ衊̃t[[N𗘗pĂꍇ͂\[XȂǂłł傤A
//܂CɂɂȂłB
struct ToolbarButton s_toolbarButtons[]={
	{ID_PAINT_STEPS,IDB_STEPS,0,TBSTYLE_BUTTON},
	{ID_PAINT_FILL,IDB_FILL,0,TBSTYLE_BUTTON},
	{0,0,0,TBSTYLE_SEP},
	{ID_TOOL_MOVE,IDB_TOOL_MOVE,0,TBSTYLE_CHECKGROUP},
	{ID_TOOL_PEN,IDB_TOOL_PEN,TBSTATE_CHECKED,TBSTYLE_GROUP},
	{ID_TOOL_INSERTKNOT,IDB_TOOL_INSERTKNOT,0,TBSTYLE_GROUP},
	{ID_TOOL_DELETEPOINT,IDB_TOOL_DELETEPOINT,0,TBSTYLE_GROUP},
	{ID_TOOL_CHANGECONTROL,IDB_TOOL_CHANGECONTROL,0,TBSTYLE_GROUP},
	{ID_TOOL_CUT,IDB_TOOL_CUT,0,TBSTYLE_GROUP},
	{ID_TOOL_RECTANGLE,IDB_TOOL_RECTANGLE,0,TBSTYLE_GROUP},
	{ID_TOOL_CIRCLE,IDB_TOOL_CIRCLE,0,TBSTYLE_GROUP},
	{ID_TOOL_SELECTKNOTS,IDB_TOOL_SELECTKNOTS,0,TBSTYLE_GROUP},
	{0,0,0,TBSTYLE_SEP},
	{ID_PATH_GROUP,IDB_GROUP,0,TBSTYLE_BUTTON},
	{ID_PATH_UNGROUP,IDB_UNGROUP,0,TBSTYLE_BUTTON},
};
#define TOOLBARBUTTONCOUNT (sizeof(s_toolbarButtons)/sizeof(struct ToolbarButton))

Frame::Frame()
:m_view(NULL)
,m_statusBar(NULL)
,m_toolBar(NULL)
{
}

Frame::~Frame() throw()
{
	delete m_view;
}

void Frame::SetView(View *view)
{
	assert(view!=NULL && m_view==NULL);
	m_view=view;
}

bool Frame::Create()
{
	const char *title=GetView()->GetDocument()->GetTitle();
	if(!Window::Create(title,WS_OVERLAPPEDWINDOW,NULL)) {
		return false;
	}

	if(!m_view->Create(this)) {
		return false;
	}
	::InitCommonControls();

	m_statusBar=::CreateStatusWindow(
		WS_CHILD|WS_VISIBLE
		,NULL,GetHWND(),STATUSBARID);
	if(m_statusBar==NULL) {
		return false;
	}

	m_toolBar=CreateToolbar(s_toolbarButtons
		,TOOLBARBUTTONCOUNT,TOOLBARID,GetHWND());

	return true;
}

bool Frame::WindowProc(UINT msg,WPARAM wparam,LPARAM lparam)
{
	switch(msg) {
	case WM_COMMAND:
		OnCommand(wparam,lparam);
		break;

	case WM_SIZE:
		OnSize((UINT)wparam,LOWORD(lparam),HIWORD(lparam));
		break;

	case WM_NCDESTROY:
		OnNcDestroy();
		break;

	case WM_INITMENUPOPUP:
		OnInitMenuPopup(reinterpret_cast<HMENU>(wparam)
			,LOWORD(lparam)
			,HIWORD(lparam)!=FALSE);
		break;

	case WM_MENUSELECT:
		OnMenuSelect(LOWORD(wparam),HIWORD(wparam),HMENU(lparam));
		break;

	case WM_KEYDOWN:
		OnKeyDown(wparam,lparam);
		break;

	case WM_KEYUP:
		OnKeyUp(wparam,lparam);
		break;

	case SAMPLE_WM_UPDATE_WINDOW:
		OnUpdateWindow();
		break;

	default:
		return false;
	}
	return true;
}

void Frame::OnNcDestroy()
{
	UnregisterWindow(this);
	Window::OnNcDestroy();
}

void Frame::OnSize(UINT type,WORD width,WORD height)
{
	::SendMessage(m_statusBar,WM_SIZE,(WPARAM)type,MAKELONG(width,height));
	::SendMessage(m_toolBar,WM_SIZE,(WPARAM)type,MAKELONG(width,height));

	RECT rect;
	GetClientRect(&rect);

	RECT statusBarRect;
	::GetWindowRect(m_statusBar,&statusBarRect);
	rect.bottom-=Height(statusBarRect);

	RECT toolBarRect;
	::GetWindowRect(m_toolBar,&toolBarRect);
	rect.top+=Height(toolBarRect);

	m_view->SetWindowPos(0,rect.left,rect.top,Width(rect),Height(rect),0);
}

View *Frame::GetView()
{
	return m_view;
}

void Frame::OnCommand(WPARAM wparam,LPARAM lparam)
{
	int id=LOWORD(wparam);
	Frame *frame;
	switch(id) {
	case ID_FILE_NEWDOCUMENT:
		frame=CreateDocumentWindow();
		frame->ShowWindow(SW_SHOW);
		break;

	case ID_FILE_CLOSEWINDOW:
		DestroyWindow();
		break;

	case ID_FILE_APPEXIT:
		DestroyAllWindows();
		break;

	default:
		m_view->SendMessage(WM_COMMAND,wparam,lparam);
		break;
	}
}

void Frame::OnInitMenuPopup(HMENU hmenu,WORD,bool isSystemMenu)
{
	if(isSystemMenu) return;

	int menuCount;
	menuCount=::GetMenuItemCount(hmenu);

	for(int i=0;i<menuCount;++i) {
		Command command;
		command.SetMenu(hmenu,i);
		UINT commandID=::GetMenuItemID(hmenu,i);
		OnUpdateCommand(commandID,&command);
	}
}

void Frame::OnUpdateWindow()
{
	HWND toolbar=NULL;
	for(HWND hwnd=::GetWindow(GetHWND(),GW_CHILD)
		;hwnd!=NULL;hwnd=::GetWindow(hwnd,GW_HWNDNEXT)) {

		UINT id=::GetDlgCtrlID(hwnd);
		if(id==TOOLBARID) {
			toolbar=hwnd;
			break;
		}
	}

	int count=::SendMessage(toolbar,TB_BUTTONCOUNT,0,0);
	for(int i=0;i<count;++i) {
		TBBUTTON tbbutton;
		::SendMessage(toolbar,TB_GETBUTTON,i,reinterpret_cast<LPARAM>(&tbbutton));
		WORD id=tbbutton.idCommand;
		Command command;
		command.SetToolbar(toolbar,id);
		OnUpdateCommand(id,&command);
	}
}

void Frame::OnUpdateCommand(UINT id,Command *command)
{
	GetView()->OnUpdateCommand(id,command);
}

void Frame::SetStatusText(UINT textID)
{
	static UINT lastID=0;
	if(textID==lastID) return;
	lastID=textID;

	char text[255];
	if(textID!=0) {
		if(::LoadString(g_hInstance,textID,text,255)==0) {
			text[0]=0;
		} else {
			char *ret=strchr(text,'\n');
			if(ret!=NULL) {
				ret[0]=0;
			}
		}
	} else {
		text[0]=0;
	}
	::SendMessage(m_statusBar,WM_SETTEXT,0,(LPARAM)text);
}

void Frame::OnMenuSelect(UINT item,UINT flag,HMENU)
{
	if((flag&MF_POPUP)==0) {
		SetStatusText(item);
	}
}

void Frame::OnKeyDown(WPARAM wparam,LPARAM lparam)
{
	GetView()->SendMessage(WM_KEYDOWN,wparam,lparam);
}

void Frame::OnKeyUp(WPARAM wparam,LPARAM lparam)
{
	GetView()->SendMessage(WM_KEYUP,wparam,lparam);
}

SAMPLE_NAMESPACE_END
