#include "prefix.h"

#include "history.h"

#include <assert.h>
#include <deque>

/*************************************************

	̊Ǘ Ȃ蔲

**************************************************/

HIS_NAMESPACE_START

/////////////////////History///////////////////////

typedef std::deque<Record *> Records;
typedef Records::iterator RecIterator;

class HistoryMember
{
public:
	HistoryMember()
	{
	}
	~HistoryMember() throw()
	{
	}

	Records records;
};

History::History()
 :m_m(new HistoryMember)
 ,m_undoRedoNow(false)
 ,m_recordedCount(0)
 ,m_modifiedCount(0)
 ,m_index(0)
 ,m_nestCount(0)
{
	SAMPLE_BAD_ALLOC_TEST(m_m.get());
}

History::~History() throw()
{


}

//A̗̋L^Jn
void History::BeginRecord()
{
	if(m_nestCount==0 && !m_undoRedoNow) {
		m_pushedFirstRecord=false;
	}
	++m_nestCount;
}

//̋L^I
void History::EndRecord()
{
	assert(m_nestCount>0);

	if(m_pushedFirstRecord && m_index>0) {
		m_m->records[m_index-1]->SetLast();
	}

	--m_nestCount;
}

//L^
void History::Push(Record *record)
{
	if(m_undoRedoNow) return;

	if(!m_pushedFirstRecord) {
		record->SetFirst();
		m_pushedFirstRecord=true;
	}
	RemoveFollowRecords(m_index);

	m_m->records.push_back(record);

	if(record->DoesModified()) {
		++m_modifiedCount;
	}
	++m_recordedCount;
	++m_index;
}

//L^𒆎~
void History::Cancel() throw()
{
	if(m_nestCount>0 && m_pushedFirstRecord) {
		Seek(true);
		m_recordedCount=m_index;
	} else if(m_undoRedoNow) {
		Seek(!m_doesUndo);
	}
	m_nestCount=0;
}

void History::Seek(bool undo) throw()
{
	for(const Record *record=UndoRedoFirst(undo);record!=NULL;record=UndoRedoNext());
}

void History::RemoveFollowRecords(int index)
{
	if(index<m_recordedCount) {
		for(int i=index;i<m_m->records.size();++i) {
			Record *record=m_m->records[i];
			delete record;
			m_m->records[i]=NULL;
		}

		m_m->records.erase(m_m->records.begin()+index
			,m_m->records.end());
		m_recordedCount=index;
	}
}


//ɖ߂邩
bool History::CanUndoRedo(bool undo) const
{
	bool result;
	if(undo) {
		result=m_index>0;
	} else {
		result=m_index<m_recordedCount;
	}
	return result;
}

Record *History::GetRecord(bool undo)
{
	if(!CanUndoRedo(undo)) {
		m_undoRedoNow=false;
		return NULL;
	}

	if(undo) {
		--m_index;
	}
	Record *record=m_m->records[m_index];
	record->Rewind();
	if(record->DoesModified()) {
		if(undo) {
			--m_modifiedCount;
		} else {
			++m_modifiedCount;
		}
	}
	if(!undo) {
		++m_index;
	}

	if(m_doesUndo) {
		if(record->IsFirst()) {
			m_undoRedoNow=false;
		}
	} else {
		if(record->IsLast()) {
			m_undoRedoNow=false;
		}
	}
	return record;
}

Record *History::UndoRedoFirst(bool undo)
{
	m_doesUndo=undo;
	m_undoRedoNow=true;
	Record *record=GetRecord(undo);
	if(record==NULL) {
		m_undoRedoNow=false;
		return NULL;
	}
	return record;
}

Record *History::UndoRedoNext()
{
	if(!m_undoRedoNow) {
		return NULL;
	}

	return GetRecord(m_doesUndo);
}

////////////////////Record//////////////////

enum
{
	FLAGS_NONE,
	FLAGS_FIRST,
	FLAGS_LAST,
};

enum
{
	TYPE_BUFFER,
};


Record::Record(int key,int subKey,bool doesModified)
	:m_key(key)
	,m_subKey(subKey)
	,m_doesModified(doesModified)
	,m_flags(FLAGS_NONE)
	,m_buffer(NULL)
	,m_size(0)
	,m_pointer(0)
{

}

Record::~Record() throw()
{
	free(m_buffer);
}

void Record::Rewind()
{
	m_pointer=0;
}

void Record::SetFirst()
{
	m_flags|=FLAGS_FIRST;
}
void Record::SetLast()
{
	m_flags|=FLAGS_LAST;
}

bool Record::IsFirst() const
{
	return (m_flags&FLAGS_FIRST)!=0;
}

bool Record::IsLast() const
{
	return (m_flags&FLAGS_LAST)!=0;
}

bool Record::DoesModified() const
{
	return m_doesModified;
}

void Record::WriteBuffer(const char *buffer,int size)
{
	SubWriteInt(TYPE_BUFFER);
	SubWriteInt(size);
	SubWriteData(buffer,size);
}

void Record::SubWriteInt(int value)
{
	SubWriteData(&value,sizeof(int));
}

void Record::SubWriteData(const void *data,int size)
{
	void *newBuffer=realloc(m_buffer,m_size+size);
	if(newBuffer==NULL) {
		throw Exception(HISERROR_OUT_OF_MEMORY);
	}

	m_buffer=reinterpret_cast<char*>(newBuffer);
	memcpy(m_buffer+m_size,data,size);
	m_size+=size;
}

const char *Record::ReadBuffer(int *size)
{
	int type=SubReadInt();
	assert(type==TYPE_BUFFER);
	*size=SubReadInt();

	char *buffer=m_buffer+m_pointer;
	m_pointer+=*size;
	return buffer;
}

int Record::SubReadInt()
{
	int value;
	SubReadData(&value,sizeof(int));
	return value;
}

void Record::SubReadData(void *data,int size)
{
	memcpy(data,m_buffer+m_pointer,size);
	m_pointer+=size;
}

int Record::GetKey() const
{
	return m_key;
}

int Record::GetSubKey() const
{
	return m_subKey;
}

HIS_NAMESPACE_END
