#include "prefix.h"

#include <windows.h>
#include "image.h"

/*************************************************

	摜obt@

**************************************************/

#define PIXELBYTES 3

SAMPLE_NAMESPACE_START

Image::Image()
	:m_buffer(NULL)
{
	const BITMAPINFOHEADER bihInit={
		sizeof(BITMAPINFOHEADER)
		,0,0	//width,height
		,1	//planes
		,PIXELBYTES*8	//bitCount
		,BI_RGB	//compression
		,0	//size
		,0,0	//PelsPerMeter
		,0	//clrUsed
		,0	//clrImportant
	};

	m_bitmapInfo.bmiHeader=bihInit;
}

Image::~Image() throw()
{
	Destroy();
}

void Image::Destroy() throw()
{
	if(m_buffer!=NULL) {
		::GlobalFree(m_buffer);
		m_buffer=NULL;
	}
}

int Image::Width() const
{
	return m_size.cx;
}

int Image::Height() const
{
	return m_size.cy;
}

int Image::RowTotalBytes() const
{
	return RowBytes()+PadBytes();
}

int Image::PadBytes() const
{
	return RowBytes()&(sizeof(LONG)-1);
}

int Image::RowBytes() const
{
	return Width()*PIXELBYTES;
}

int Image::TotalBytes() const
{
	return RowTotalBytes()*Height();
}

bool Image::Create(SIZE size)
{
	Destroy();

	if(size.cx<=0 || size.cy<=0) {
		return false;
	}

	m_size=size;
	m_buffer=static_cast<unsigned char*>(::GlobalAlloc(GMEM_FIXED,TotalBytes()));

	for(int y=0;y<Height();++y) {
		assert(RowBytes()>0);
		memset(m_buffer+y*RowTotalBytes(),0xff,RowBytes());
		if(PadBytes()>0) {
			memset(m_buffer+y*RowTotalBytes()+RowBytes(),0,PadBytes());
		}
	}

	m_bitmapInfo.bmiHeader.biWidth=size.cx;
	m_bitmapInfo.bmiHeader.biHeight=-size.cy;

	return (m_buffer!=NULL);
}

unsigned char *Image::GetBuffer()
{
	return m_buffer;
}

unsigned char *Image::GetRow(int y)
{
	return m_buffer+RowTotalBytes()*y;
}

const BITMAPINFO *Image::BitmapInfo() const
{
	return &m_bitmapInfo;
}

int Image::PixelBytes() const
{
	return PIXELBYTES;
}

SAMPLE_NAMESPACE_END
