#include "prefix.h"

#include <windows.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "main.h"
#include "sampleview.h"
#include "sampledocument.h"
#include "resource.h"
#include "dc.h"
#include "clipboard.h"
#include "hglobal.h"
#include "winmessage.h"
#include "command.h"
#include "dimptr.h"
#include "history.h"
#include "image.h"
#include "filedialog.h"
#include "messagebox.h"
#include "frame.h"

#include <bzpath.h>

/*************************************************

	Tvr[NX

**************************************************/

SAMPLE_NAMESPACE_START

using namespace bzp;

//̎ʔԍ
#define HISTORY_PATH 'path'

class SampleMember
{
public:
	bzp::Path clipboard;
	bzp::Graph graph;
	bzp::Path previousPath;
	bzp::ToolHolder tool;
};

SampleView::SampleView()
	:m_m(new SampleMember)
	,m_magnification(100)
	,m_showPath(true)
	,m_drawMode(DRAWMODE_VIEW)
	,m_cutKnot(false)
	,m_isToolOperating(false)
{
	SAMPLE_BAD_ALLOC_TEST(m_m.get());

	m_margin.cx=30;
	m_margin.cy=40;
}

SampleView::~SampleView() throw()
{

}

void SampleView::OnInitialUpdate()
{
	View::OnInitialUpdate();

	SIZE docSize=GetDocument()->GetSize();
	SetScrollSize(docSize);

	//\̋`̍Wݒ
	RECT viewRect;
	CalcViewRect(&viewRect);
	m_m->graph.SetViewRect(viewRect);

	//W̋`̐ݒ
	RECT boundRect;
	::SetRect(&boundRect,0,0,docSize.cx,docSize.cy);
	m_m->graph.SetBoundRect(boundRect);
	//\@̐ݒ
	m_m->graph.SetDrawMode(m_drawMode);

	//ŏ̃c[I
	SelectTool(TOOLTABLE_PEN);

	//ύXƔr褖߂肷邽߂ɁA
	//݂̃pX̏ԂĂ
	SetPreviousPath();

	SetWindowTitle();
}

//EChȆ薼ݒ
void SampleView::SetWindowTitle()
{
	const char *title=GetDocument()->GetTitle();

	char buffer[255];
	snprintf(buffer,255,"%s:%d (%d%%)",title,GetViewNumber(),m_magnification);

	GetFrame()->SetWindowText(buffer);
}

SampleDocument *SampleView::GetDocument()
{
	return reinterpret_cast<SampleDocument*>(View::GetDocument());
}

const SampleDocument *SampleView::GetDocument() const
{
	return reinterpret_cast<const SampleDocument*>(View::GetDocument());
}

bzp::Path *SampleView::GetPath()
{
	return GetDocument()->GetPath();
}


#define MAGNIFY(a) ((a*m_magnification+99)/100)

//r[̃hLg̕\̑傫w肵
//XN[ݒ肷
void SampleView::SetScrollSize(SIZE docSize)
{
	SIZE scrollSize;
	scrollSize.cx=MAGNIFY(docSize.cx)+m_margin.cx*2;
	scrollSize.cy=MAGNIFY(docSize.cy)+m_margin.cy*2;
	View::SetScrollSize(scrollSize);
}

//EBhEbZ[W
bool SampleView::WindowProc(UINT msg,WPARAM wparam,LPARAM lparam)
{
	switch(msg) {
	case WM_PAINT:
		OnPaint();
		break;

	case WM_LBUTTONDOWN:
		OnLButtonDown((short)LOWORD(lparam),(short)HIWORD(lparam),(int)wparam);
		break;

	case WM_LBUTTONDBLCLK:
		OnLButtonDblClk((short)LOWORD(lparam),(short)HIWORD(lparam),(int)wparam);
		break;

	case WM_LBUTTONUP:
		OnLButtonUp((short)LOWORD(lparam),(short)HIWORD(lparam),(int)wparam);
		break;

	case WM_RBUTTONDOWN:
		OnRButtonDown((short)LOWORD(lparam),(short)HIWORD(lparam),(int)wparam);
		break;

	case WM_MOUSEMOVE:
		OnMouseMove((short)LOWORD(lparam),(short)HIWORD(lparam),(int)wparam);
		break;

	case WM_KEYDOWN:
		OnKeyDown((int)wparam,(UINT)lparam);
		break;

	case WM_KEYUP:
		OnKeyUp((int)wparam,(UINT)lparam);
		break;

	case WM_SETCURSOR:
		if(OnSetCursor((HWND)wparam,LOWORD(lparam),HIWORD(wparam))) {
			break;
		}
		return Window::WindowProc(msg,wparam,lparam);

	case WM_ACTIVATE:
		OnActivate(LOWORD(wparam),HIWORD(wparam)!=FALSE,(HWND)lparam);
		break;

	case WM_RENDERALLFORMATS:
		OnRenderAllFormats();
		break;

	case WM_RENDERFORMAT:
		OnRenderFormat((UINT)wparam);
		break;

	case WM_DESTROYCLIPBOARD:
		OnDestroyClipboard();
		break;

		//pX
	case SAMPLE_WM_HIDE_PATH:
		OnHidePath((HWND)wparam,(SampleDocument *)lparam);
		break;

		//pX\
	case SAMPLE_WM_SHOW_PATH:
		OnShowPath((HWND)wparam,(SampleDocument *)lparam);
		break;

		//c[؂ւ
	case SAMPLE_WM_SET_TOOL:
		OnSelectTool((WORD)wparam);
		break;

		//\XV
	case SAMPLE_WM_REDRAW_VIEW:
		OnRedrawView((SampleDocument*)wparam,(const RECT*)lparam);
		break;

	default:
		return View::WindowProc(msg,wparam,lparam);
	}
	return true;
}

void SampleView::OnCommand(WPARAM wparam,LPARAM lparam)
{
	WORD id=LOWORD(wparam);

	switch(id) {
	case ID_FILE_IMPORT:
		ModifyPath(&SampleView::OnFileImport);
		break;

	case ID_FILE_EXPORT:
		UsePath(&SampleView::OnFileExport);
		break;

	case ID_EDIT_UNDO:
		ModifyPath(&SampleView::OnUndo,true);
		break;

	case ID_EDIT_REDO:
		ModifyPath(&SampleView::OnRedo,true);
		break;

	case ID_EDIT_CUT:
		ModifyPath(&SampleView::OnCut);
		break;

	case ID_EDIT_COPY:
		ModifyPath(&SampleView::OnCopy);
		break;

	case ID_EDIT_PASTE:
		ModifyPath(&SampleView::OnPaste);
		break;

	case ID_EDIT_SELECT_ALL:
		ModifyPath(&SampleView::OnSelectAll);
		break;

	case ID_EDIT_UNSELECT_ALL:
		ModifyPath(&SampleView::OnUnselectAll);
		break;

	case ID_TOOL_PEN:
	case ID_TOOL_MOVE:
	case ID_TOOL_INSERTKNOT:
	case ID_TOOL_DELETEPOINT:
	case ID_TOOL_CHANGECONTROL:
	case ID_TOOL_CUT:
	case ID_TOOL_RECTANGLE:
	case ID_TOOL_CIRCLE:
	case ID_TOOL_SELECTKNOTS:
		OnToolButton(id);
		break;

	case ID_VIEW_PER25:
	case ID_VIEW_PER50:
	case ID_VIEW_PER75:
	case ID_VIEW_PER100:
	case ID_VIEW_PER200:
	case ID_VIEW_PER400:
		OnViewMagnification(id);
		break;

	case ID_DRAWMODE_HIDE_MARGIN:
	case ID_DRAWMODE_HIDE_UNSELECTED:
	case ID_DRAWMODE_ARROW_FORWARDCONTROL:
		OnDrawMode(id);
		break;

	case ID_PATH_HREVERSE:
		ModifyPath(&SampleView::OnHReverse);
		break;

	case ID_PATH_VREVERSE:
		ModifyPath(&SampleView::OnVReverse);
		break;

	case ID_PATH_GROUP:
		ModifyPath(&SampleView::OnGroup);
		break;

	case ID_PATH_UNGROUP:
		ModifyPath(&SampleView::OnUngroup);
		break;

	case ID_PAINT_STEPS:
		UsePath(&SampleView::OnSteps);
		break;

	case ID_PAINT_FILL:
		UsePath(&SampleView::OnFill);
		break;

	case ID_OPTION_CUT_KNOT:
		OnCutKnot();
		break;

	default:
		View::OnCommand(wparam,lparam);
		break;
	}
}

//GUI̍XV
void SampleView::OnUpdateCommand(UINT id,Command *command)
{
	View::OnUpdateCommand(id,command);

	switch(id) {
	case ID_FILE_IMPORT:
		OnUpdateImport(command);
		break;

	case ID_EDIT_UNDO:
		OnUpdateUndo(command);
		break;
	case ID_EDIT_REDO:
		OnUpdateRedo(command);
		break;

	case ID_EDIT_CUT:
	case ID_EDIT_COPY:
		OnUpdateCopy(command);
		break;

	case ID_EDIT_PASTE:
		OnUpdatePaste(command);
		break;

	case ID_TOOL_PEN:
	case ID_TOOL_MOVE:
	case ID_TOOL_INSERTKNOT:
	case ID_TOOL_DELETEPOINT:
	case ID_TOOL_CHANGECONTROL:
	case ID_TOOL_CUT:
	case ID_TOOL_RECTANGLE:
	case ID_TOOL_CIRCLE:
	case ID_TOOL_SELECTKNOTS:
		OnUpdateTool(id,command);
		break;

	case ID_VIEW_PER25:
	case ID_VIEW_PER50:
	case ID_VIEW_PER75:
	case ID_VIEW_PER100:
	case ID_VIEW_PER200:
	case ID_VIEW_PER400:
		OnUpdateViewMagnification(id,command);
		break;

	case ID_DRAWMODE_HIDE_MARGIN:
	case ID_DRAWMODE_HIDE_UNSELECTED:
	case ID_DRAWMODE_ARROW_FORWARDCONTROL:
		OnUpdateDrawMode(id,command);
		break;

	case ID_PATH_GROUP:
		OnUpdateGroup(command);
		break;

	case ID_PATH_UNGROUP:
		OnUpdateUngroup(command);
		break;

	case ID_FILE_EXPORT:
	case ID_PAINT_STEPS:
	case ID_PAINT_FILL:
	case ID_EDIT_SELECT_ALL:
	case ID_EDIT_UNSELECT_ALL:
	case ID_PATH_HREVERSE:
	case ID_PATH_VREVERSE:
		OnUpdateNotEmpty(command);
		break;

	case ID_OPTION_CUT_KNOT:
		OnUpdateCutKnot(command);
		break;
	}
}

//WɃXN[l
POINT SampleView::GetViewPoint(int x,int y) const
{
	POINT scroll=GetScrollPos();
	POINT point;
	point.x=x+scroll.x;
	point.y=y+scroll.y;
	return point;
}

//EBhEbZ[WƂ̍W
POINT SampleView::GetMessagePoint() const
{
	DWORD pos=::GetMessagePos();
	POINTS points=MAKEPOINTS(pos);
	POINT point;
	point.x=points.x;
	point.y=points.y;
	ScreenToClient(&point);
	return GetViewPoint(point.x,point.y);
}

//hLg\`vZ
void SampleView::CalcViewRect(RECT *rect)
{
	SIZE imageSize=GetDocument()->GetSize();
	RECT boundRect;
	::SetRect(&boundRect,0,0,imageSize.cx,imageSize.cy);
	BoundToView(rect,&boundRect);
}

//pX̃CxgĂ
void SampleView::CallEventProcess(EVENT event,POINT point,int key)
{
	//pXύXꍇɁAɏ͕̂s\Ȃ̂ŁA
	//̃EBhEœpXgĂΕ\Ȃ悤ɂ
	if(event==EVENT_LEFTDOWN && !m_isToolOperating) {
		SendMessageToAllViews(SAMPLE_WM_HIDE_PATH
			,(WPARAM)GetHWND(),(LPARAM)GetDocument());
		m_isToolOperating=true;
	}

	//pXύXOrĕ\
	ClientDC dc(GetHWND());
	bzp::Redraw redraw;
	//ύXȌԂrbg}bvɕ`
	BeginRedraw(&redraw,&dc);

	bzp::Path *path=GetPath();

	//c[̃CxgĂяo
	m_m->tool.Entry(event,path,&(m_m->graph),point,key);

	//ύX̏Ԃrbg}bvɕ`悵
	//O̔rĉʂɕ\
	EndRedraw(&redraw);	
	dc.Delete();

	//̍쐬

	//c[𒆒fłԂŉ炩̕ύXȂĂꍇ
	if(m_m->tool.CanStop() && m_m->tool.DidModify()) {
		//ύXĂƂtO̓AvŖ߂
		m_m->tool.ClearModifiedFlag();
		PushHistory();
		SetPreviousPath();
	}

	//J[\̕\ς
	SetPathCursor();

	//Xe[^Xo[̕\
	SetPathStatusText();

	//}EXLv`[؂ւ
	if(m_m->tool.IsCapturing()) {
		if(::GetCapture()!=GetHWND()) {
			SetCapture();
		}
	} else {
		if(::GetCapture()==GetHWND()) {
			::ReleaseCapture();
		}
	}

	//pX̕ύX삪I瑼̃EBhEŔ\
	//ȂĂpX\
	if(m_isToolOperating && !m_m->tool.IsOperating()) {
		SendMessageToAllViews(SAMPLE_WM_SHOW_PATH
			,(WPARAM)GetHWND(),(LPARAM)GetDocument());
		m_isToolOperating=false;
	}
}

//////////////////Cxg//////////////////

//̂Ƃꕔ̃EBhEbZ[WgĂȂ
//łǂȂ邩͕Ȃ

//\
void SampleView::OnPaint()
{
	PaintDC dc(GetHWND());
	HDC hDC=dc.GetDC();
	OnPrepareDC(hDC);

	RECT viewRect;
	CalcViewRect(&viewRect);
	::Rectangle(hDC,viewRect.left-1,viewRect.top-1,viewRect.right+1,viewRect.bottom+1);

	Image *image=GetDocument()->GetImage();
	SIZE size=GetDocument()->GetSize();

	::StretchDIBits(hDC
		,viewRect.left,viewRect.top
		,viewRect.right-viewRect.left,viewRect.bottom-viewRect.top
		,0,0
		,size.cx,size.cy
		,image->GetBuffer()
		,image->BitmapInfo()
		,DIB_RGB_COLORS
		,SRCCOPY);

	bzp::Redraw redraw;
	SetRedraw(&redraw,&dc);
	redraw.End();
}

//{^
void SampleView::OnLButtonDown(int x,int y,int)
{
	//bZ[W̍WɃEBhẼXN[̒l
	//Wn
	POINT viewPoint=GetViewPoint(x,y);
	CallEventProcess(EVENT_LEFTDOWN,viewPoint,0);

	//hbO邩𒲂ׂ
	POINT screenPoint;
	screenPoint.x=x;
	screenPoint.y=y;
	ClientToScreen(&screenPoint);
	if(DragDetect(screenPoint)) {
		//hbOn߂bZ[W
		CallEventProcess(EVENT_DRAG,viewPoint,0);
	} else {
		CallEventProcess(EVENT_LEFTUP,viewPoint,0);
	}
}

//{^𗣂
void SampleView::OnLButtonUp(int x,int y,int)
{
	POINT point=GetViewPoint(x,y);
	CallEventProcess(EVENT_LEFTUP,point,0);
}

//{^̂Qx
void SampleView::OnLButtonDblClk(int x,int y,int)
{
	POINT point=GetViewPoint(x,y);
	CallEventProcess(EVENT_LEFTDOUBLECLICK,point,0);
}

//E{^
void SampleView::OnRButtonDown(int x,int y,int)
{
	POINT point=GetViewPoint(x,y);
	CallEventProcess(EVENT_RIGHTDOWN,point,0);
}

//J[\̈ړ
void SampleView::OnMouseMove(int x,int y,int)
{
	POINT point=GetViewPoint(x,y);
	CallEventProcess(EVENT_MOUSEMOVE,point,0);
}

//L[
void SampleView::OnKeyDown(int key,UINT)
{
	//L[͂łWnȂƂȂ
	POINT point=GetMessagePoint();
	CallEventProcess(EVENT_KEYDOWN
		,point,key);
}

//L[𗣂
void SampleView::OnKeyUp(int key,UINT)
{
	POINT point=GetMessagePoint();
	CallEventProcess(EVENT_KEYUP,point
		,key);
}
//////////////////////////////////////////////////////

//J[\̕\
bool SampleView::OnSetCursor(HWND hwnd,WORD hitTest,WORD)
{
	if(hwnd==GetHWND() && hitTest==HTCLIENT) {
		//}EXLv`[ĂԂŃJ[\ύXȂ
		//Ȃ̂ŃpX̃J[\ݒ蕔͕܂
		if(SetPathCursor()) {
			return true;
		}
	}
	return false;
}

//̃EBhEANeBuɂȂ
//t[[Nɂăr[ƃEBhEʂȂA
//r[ANeBuɂȂꍇ̏
void SampleView::OnActivate(WORD activeMode,bool,HWND)
{
	if(activeMode!=WA_INACTIVE) {
		//̃EBhẼpX\ɂȂĂ\
		ShowPath(true);
	}
}

//EBhȆ傫ύXꂽ
void SampleView::OnSize(UINT type,WORD cx,WORD cy)
{
	View::OnSize(type,cx,cy);

	//r[̐^񒆂ɃhLg\悤ɂĂꍇɂ
	//EBhȆ傫ɂăhLg̕\ʒuς̂
	//ČvZKv
	RECT viewRect;
	CalcViewRect(&viewRect);
	m_m->graph.SetViewRect(viewRect);
}

//pXɊւJ[\̕\
bool SampleView::SetPathCursor()
{
/*J[\ɂ邩̓AvP[Vɂ܂܂
ƑGcɂĂ悢ASĂ̏Ԃɕʂ̃J[\蓖ĂĂǂ
Ɏ`咣ȂΉ̂悤ɂĂ*/

	UINT cursorID=0;
	LPCTSTR standardCursor=NULL;

	//c[̏Ԃ擾
	int condition=m_m->tool.GetCondition();

	//c[̏ԂɂăJ[\
	switch(condition) {
	case COND_MOVE_IDLE:
	case COND_SELECTBOX:
	case COND_INSERTKNOT_IDLE:
	case COND_CHANGECONTROL_IDLE:
	case COND_DELETEPOINT_IDLE:
	case COND_SELECTKNOTS_IDLE:
		standardCursor=IDC_ARROW;
		break;

	case COND_PEN_IDLE:
	case COND_PEN_MOVE:
		cursorID=IDC_CURSORPEN;
		break;

	case COND_PEN_CHAIN:
	case COND_PEN_CLOSE:
		cursorID=IDC_CURSORPENCHAIN;
		break;

	case COND_MOVE_CURVE:
	case COND_MOVE_SUBGROUP:
	case COND_MOVE_KNOT:
	case COND_MOVE_FIND_ANCHOR:
	case COND_MOVE_FIND_CURVE:
	case COND_MOVE_FIND_SUBGROUP:
		cursorID=IDC_CURSORMOVE;
		break;

	case COND_MOVE_CHAIN:
	case COND_MOVE_CLOSE:
		cursorID=IDC_CURSORMOVECHAIN;
		break;

	case COND_MOVE_CONTROL:
	case COND_MOVE_FIND_CONTROL:
	case COND_MOVE_WCONTROLS:
	case COND_CHANGE_UNLINKED_CONTROL:
		cursorID=IDC_CURSORMOVECONTROL;
		break;

	case COND_CHANGE_LINKED_CONTROL:
		cursorID=IDC_CURSORUNLINKCONTROL;
		break;

	case COND_PEN_TERMINATE:
		cursorID=IDC_CURSORTERMINATE;
		break;

	case COND_NONE_IDLE:
	case COND_PEN_OUTSIDE:	
		break;

	case COND_INSERT_KNOT:
		cursorID=IDC_CURSORINSERT;
		break;

	case COND_CHANGE_STRAIGHT_KNOT:
	case COND_CHANGE_CURVE_KNOT:
		cursorID=IDC_CURSORCHANGE;
		break;

	case COND_DELETE_KNOT:
	case COND_DELETE_CONTROL:
		cursorID=IDC_CURSORDELETE;
		break;

	case COND_SHAPE_IDLE:
	case COND_SHAPE_MOVE:
		standardCursor=IDC_CROSS;
		break;

	case COND_CUT_IDLE:
	case COND_CUT_FIND_CURVE:
		cursorID=IDC_CURSORCUT;
		break;

	case COND_SELECTKNOTS_SELECT:
		cursorID=IDC_CURSOR_SELECT_KNOTS;
		break;

	case COND_SELECTKNOTS_UNSELECT:
		cursorID=IDC_CURSOR_UNSELECT_KNOTS;
		break;

	case COND_SELECTKNOTS_ALTERNATE:
		cursorID=IDC_CURSOR_ALTERNATE_SELECT_KNOTS;
		break;

	default:
		assert(false);
	}

	HCURSOR cursor=NULL;
	if(cursorID!=0) {
		cursor=::LoadCursor(g_hInstance,MAKEINTRESOURCE(cursorID));
	} else if(standardCursor!=NULL) {
		cursor=::LoadCursor(NULL,MAKEINTRESOURCE(standardCursor));
	}

	if(cursor!=NULL) {
		::SetCursor(cursor);
		return true;
	}
	return false;
}

//pXɊւXe[^Xo[͕̕\
void SampleView::SetPathStatusText()
{
	/*ǂ\邩̓IvVL[̐ݒɂ܂
	Ⴆ΃Tvł́uhbOāEEEvƕ\ꍇ܂
	hbOȂݒɂꍇ͂̕\͂
	L[ƂǂȂ邩Ƃ̂ݒɂ܂
	*/

	/* vύX@ԂɂăXe[^Xo[̕\Ȃق
  ǂ̂܂񂪖ʓ|
  p҂ݒύXł悤ɂA
	̐؂ւȂĂ炮ႮɂȂł傤
	*/

	//c[̏Ԃ擾
	int condition=m_m->tool.GetCondition();
	UINT textID=0;	//\[XID

	//Iꍇɐ؂ւ邩
	bool toggleMode=(m_m->tool.GetToolOption(MOVE_SELECT_LOW,MOVE_SELECT_HI)==OPTION_SELECT_TOGGLE);

	switch(condition) {
	case COND_CHANGECONTROL_IDLE:
		textID=IDS_CHANGECONTROL;
		break;

	case COND_CHANGE_STRAIGHT_KNOT:
		textID=IDS_CHANGE_STRAIGHT_KNOT;
		break;

	case COND_CHANGE_CURVE_KNOT:
		textID=IDS_CHANGE_CURVE_KNOT;
		break;

	case COND_CHANGE_LINKED_CONTROL:	
		textID=IDS_CHANGE_LINKED_CONTROL;
		break;

	case COND_CUT_IDLE:	
	case COND_CUT_FIND_CURVE:	
		textID=IDS_CUT;
		break;

	case COND_DELETEPOINT_IDLE:
		textID=IDS_DELETEPOINT;
		break;

	case COND_DELETE_KNOT:	
		textID=IDS_DELETE_KNOT;
		break;

	case COND_DELETE_CONTROL:	
		textID=IDS_DELETE_CONTROL;
		break;

	case COND_INSERTKNOT_IDLE:	
	case COND_INSERT_KNOT:
		textID=IDS_INSERT_KNOT;
		break;

	case COND_MOVE_IDLE:
		if(toggleMode) {
			textID=IDS_SELECT_TOGGLE;
		} else {
			textID=IDS_MOVE;
		}
		break;

	case COND_MOVE_KNOT:
	case COND_MOVE_FIND_ANCHOR:
		if(toggleMode) {
			textID=IDS_SELECT_TOGGLE;
		} else {
			textID=IDS_MOVE_KNOT;
		}
		break;

	case COND_MOVE_CHAIN:	
	case COND_PEN_CHAIN:	
		textID=IDS_CHAIN;
		break;

	case COND_MOVE_CLOSE:	
	case COND_PEN_CLOSE:	
		textID=IDS_CLOSE;
		break;

	case COND_MOVE_CONTROL:
	case COND_MOVE_WCONTROLS:
	case COND_CHANGE_UNLINKED_CONTROL:	
	case COND_MOVE_FIND_CONTROL:	
		textID=IDS_MOVE_CONTROL;
		break;

	case COND_MOVE_CURVE:
	case COND_MOVE_FIND_CURVE:	
		if(toggleMode) {
			textID=IDS_SELECT_TOGGLE;
		} else {
			textID=IDS_MOVE_CURVE;
		}
		break;

	case COND_MOVE_SUBGROUP:	
	case COND_MOVE_FIND_SUBGROUP:	
		if(toggleMode) {
			textID=IDS_SELECT_TOGGLE;
		} else {
			textID=IDS_MOVE_GROUP;
		}
		break;

	case COND_SELECTBOX:
		textID=IDS_SELECTBOX;
		break;

	case COND_PEN_OUTSIDE:
	case COND_NONE_IDLE:
		textID=0;
		break;

	case COND_PEN_IDLE:
		textID=IDS_PEN;
		break;

	case COND_PEN_MOVE:	
		textID=IDS_PEN_MOVE;
		break;

	case COND_PEN_TERMINATE:
		textID=IDS_PEN_TERMINATE;
		break;

	case COND_SELECTKNOTS_IDLE:
		switch(m_m->tool.GetToolOption(SELECTKNOTS_SELECT_LOW,SELECTKNOTS_SELECT_HI)) {
		case OPTION_SELECTKNOTS_TOGGLE:
			textID=IDS_SELECTKNOTS_TOGGLE;
			break;

		case OPTION_SELECTKNOTS_SELECT:
			textID=IDS_SELECTKNOTS_SELECT;
			break;

		case OPTION_SELECTKNOTS_UNSELECT:
			textID=IDS_SELECTKNOTS_UNSELECT;
			break;

		case OPTION_SELECTKNOTS_ALTERNATE:
			textID=IDS_SELECTKNOTS_ALTERNATE;
			break;
		}
		break;

	case COND_SELECTKNOTS_SELECT:
		textID=IDS_SELECTKNOTS_SELECT;
		break;

	case COND_SELECTKNOTS_UNSELECT:	
		textID=IDS_SELECTKNOTS_UNSELECT;
		break;

	case COND_SELECTKNOTS_ALTERNATE:	
		textID=IDS_SELECTKNOTS_ALTERNATE;
		break;

	case COND_SHAPE_IDLE:
		switch(m_m->tool.GetToolOption(SHAPE_SHAPE_LOW,SHAPE_SHAPE_HI)) {
		case OPTION_SHAPE_BOX:
			textID=IDS_BOX;
			break;

		case OPTION_SHAPE_CIRCLE:
			textID=IDS_CIRCLE;
			break;
		}
		break;

	case COND_SHAPE_MOVE:
		switch(m_m->tool.GetToolOption(SHAPE_SHAPE_LOW,SHAPE_SHAPE_HI)) {
		case OPTION_SHAPE_BOX:
			textID=IDS_BOX_END;
			break;

		case OPTION_SHAPE_CIRCLE:
			textID=IDS_CIRCLE_END;
			break;
		}
		break;

	default:
		assert(false);
	}

	SetStatusText(textID);
}

/*̃EBhEœpX𑀍쒆̏ꍇɔ\
sendhwnd bZ[W𑗂n߂EBhE
hideDocument \ɂpXhLg*/
void SampleView::OnHidePath(HWND sendhwnd,SampleDocument *hideDocument)
{
	/*̃Cu[ł͕ύXꍇɂ͕ύX̂
	̂ŁApX𔽓]ŕ\Ăꍇɂ
	̃EBhEŃpXύXOɏĂȂ
	ƕ\
	TlC\Ȃǂ̂܂ܑSĂ܂ꍇ
	Kv͂Ȃ
	ʓ|ȂANeBuEBhEȊO̓pX\Ȃ悤ɂĂǂ*/

	if(sendhwnd!=GetHWND() && hideDocument==GetDocument()) {
		ShowPath(false);
	}

	/*pXʂ̃hLgAvP[Vł
	hLgł͂ȂpX̃AhXłn΂悢ł傤*/
}

//\ĂȂpX\
void SampleView::OnShowPath(HWND sendhwnd,SampleDocument *hideDocument)
{
	if(sendhwnd!=GetHWND() && hideDocument==GetDocument()) {
		ShowPath(true);
	}
}

//pX̕\ؑւ
void SampleView::ShowPath(bool show)
{
	if((show && !m_showPath)
		|| (!show && m_showPath) ) {

		ClientDC dc(GetHWND());
		bzp::Redraw redraw;
		BeginRedraw(&redraw,&dc);
		m_showPath=show;
		EndRedraw(&redraw);
	}
}

//c[I
void SampleView::OnToolButton(WORD id)
{
	//c[؂ւꍇSẴr[Ƀc[̐؂ւ
	//`ȂƂ΂΂ɂȂĂ܂
	//΂΂ŗǂAvP[VȂ`Ȃėǂ

	//r[ȊȌꏊŕʂ̖ړIŎgĂꍇ̓c[ݒ
	//ʂɂ邱Ƃł
	SendMessageToAllViews(SAMPLE_WM_SET_TOOL,(WPARAM)id,0);
}

//r[̊g嗦̕\
struct MagnificationTable
{
	WORD id;
	int magnification;
};
struct MagnificationTable s_magnificationTable[]={
	ID_VIEW_PER25,25,
	ID_VIEW_PER50,50,
	ID_VIEW_PER75,75,
	ID_VIEW_PER100,100,
	ID_VIEW_PER200,200,
	ID_VIEW_PER400,400,
};
#define MAGNIFICATIONCOUNT (sizeof(s_magnificationTable)/sizeof(struct MagnificationTable))

int CommandIdToMagnification(WORD id)
{
	int magnification=0;
	FOR(int i=0;i<MAGNIFICATIONCOUNT;i++) {
		if(s_magnificationTable[i].id==id) {
			magnification=s_magnificationTable[i].magnification;
			break;
		}
	}
	assert(magnification!=0);
	return magnification;
}

//r[̊g嗦ύX
void SampleView::OnViewMagnification(WORD id)
{
	m_magnification=CommandIdToMagnification(id);

	RECT viewRect;
	CalcViewRect(&viewRect);
	m_m->graph.SetViewRect(viewRect);

	SIZE docSize=GetDocument()->GetSize();
	SetScrollSize(docSize);
	InvalidateRect();

	SetWindowTitle();
}

void SampleView::OnUpdateViewMagnification(WORD id,Command *command)
{
	int magnification=CommandIdToMagnification(id);

	command->SetCheck(m_magnification==magnification);
}

//\@̕\
struct DrawModeTable
{
	WORD id;
	int mode;
};
struct DrawModeTable s_drawModeTable[]={
	ID_DRAWMODE_HIDE_MARGIN,DRAWMODE_HIDE_MARGIN,
	ID_DRAWMODE_HIDE_UNSELECTED,DRAWMODE_HIDE_UNSELECTED,
	ID_DRAWMODE_ARROW_FORWARDCONTROL,DRAWMODE_ARROW_FORWARDCONTROL,
};
#define DRAWMODECOUNT (sizeof(s_drawModeTable)/sizeof(struct DrawModeTable))

int CommandIdToDrawMode(WORD id)
{
	int mode=0;
	FOR(int i=0;i<DRAWMODECOUNT;i++) {
		if(s_drawModeTable[i].id==id) {
			mode=s_drawModeTable[i].mode;
		}
	}
	assert(mode!=0);
	return mode;
}

//\@ύX
void SampleView::OnDrawMode(WORD id)
{
	int mode=CommandIdToDrawMode(id);

	if((m_drawMode&mode)!=0) {
		m_drawMode&=~mode;
	} else {
		m_drawMode|=mode;
	}

	InvalidateRect();
}

//pX̕\@GUI̍XV
void SampleView::OnUpdateDrawMode(WORD id,Command *command)
{
	int mode=CommandIdToDrawMode(id);

	command->SetCheck((m_drawMode&mode)!=0);
}

//R}hIDc[̍̔ԍo
struct IDtoToolTable
{
	WORD id;
	int toolIndex;
};
struct IDtoToolTable s_IDtoToolTable[]={
	{ID_TOOL_MOVE,TOOLTABLE_MOVE},
	{ID_TOOL_PEN,TOOLTABLE_PEN},
	{ID_TOOL_INSERTKNOT,TOOLTABLE_INSERTKNOT},
	{ID_TOOL_DELETEPOINT,TOOLTABLE_DELETEPOINT},
	{ID_TOOL_CHANGECONTROL,TOOLTABLE_CHANGECONTROL},
	{ID_TOOL_CUT,TOOLTABLE_CUT},
	{ID_TOOL_RECTANGLE,TOOLTABLE_BOX},
	{ID_TOOL_CIRCLE,TOOLTABLE_CIRCLE},
	{ID_TOOL_SELECTKNOTS,TOOLTABLE_SELECTKNOTS},
};
#define TOOLCOUNT (sizeof(s_IDtoToolTable)/sizeof(struct IDtoToolTable))

int CommandIdToToolTable(WORD id)
{
	int toolIndex=-1;
	FOR(int i=0;i<TOOLCOUNT;i++) {
		if(s_IDtoToolTable[i].id==id) {
			toolIndex=s_IDtoToolTable[i].toolIndex;
			break;
		}
	}
	assert(toolIndex>=0);
	return toolIndex;
}

//[U[c[̑I
//̏ꍇ͑SẴEBhEŋʂ̃c[ĝ
//SẴpXɓc[ݒ肷
//̖ړIŃpX𗘗pĂꍇɕʁX̃c[g悤ɂ邱Ƃł
void SampleView::OnSelectTool(WORD id)
{
	int toolIndex=CommandIdToToolTable(id);
	SelectTool(toolIndex);
}

void SampleView::OnUpdateTool(WORD id,Command *command)
{
	int toolIndex=CommandIdToToolTable(id);

	command->SetCheck(m_m->tool.GetCurrentIndex()==toolIndex);
}

//c[؂ւ
void SampleView::SelectTool(int toolIndex)
{
	ClientDC dc(GetHWND());
	bzp::Redraw redraw;
	BeginRedraw(&redraw,&dc);

	if(m_m->tool.GetCurrentIndex()!=toolIndex) {
		//r[ƂɃc[Ⴄ悤ɂȂĂ̂
		//ʂ̃c[gꍇ͑SẴc[ݒ肵Ȃ
		POINT point=GetMessagePoint();
		m_m->tool.SelectTool(toolIndex,GetPath(),&(m_m->graph));
	}

	EndRedraw(&redraw);

	if(GetCapture()==GetHWND() && !m_m->tool.IsCapturing()) {
		::ReleaseCapture();
	}
}

//pXƂɑOƌ㏈
//ԂhandlerĂяo
void SampleView::ModifyPath(void (SampleView::*handler)(),bool doesUndoRedo)
{
	try
	{
		if(!m_m->tool.CanStop()) return;

		//̃EBhE̕\~߂
		SendMessageToAllViews(SAMPLE_WM_HIDE_PATH
			,(WPARAM)GetHWND()
			,(LPARAM)GetDocument());

		ClientDC dc(GetHWND());
		bzp::Redraw redraw;
		BeginRedraw(&redraw,&dc);

		if(!doesUndoRedo && m_m->tool.IsOperating()) {
			m_m->tool.Cancel(GetPath(),&(m_m->graph));
			m_isToolOperating=false;
		}

		(this->*handler)();

		EndRedraw(&redraw);
		dc.Delete();

		//̃EBhE̕\
		SendMessageToAllViews(SAMPLE_WM_SHOW_PATH
			,(WPARAM)GetHWND()
			,(LPARAM)GetDocument());

		//ɖ߂蒼Ă̂łȂΗ
		if(!doesUndoRedo) {
			PushHistory();
		}
		SetPreviousPath();
	}
	catch(bzp::Exception &e)
	{
		RecoverFailureModifyPath(e.GetErrorCode());
	}
	catch(std::bad_alloc)
	{
		RecoverFailureModifyPath(BZPERROR_OUT_OF_MEMORY);
	}
	catch(...)
	{
		RecoverFailureModifyPath(BZPERROR_UNKNOWN);
	}
}

//pXύXĂŒ̗ȌC
void SampleView::RecoverFailureModifyPath(int error)
{
	GetDocument()->GetHistory()->Cancel();
	GetPath()->Copy(&(m_m->previousPath));

	ShowError(error);
	InvalidateRect();
}

//pX֐̑O㏈
void SampleView::UsePath(void (SampleView::*handler)())
{
	try
	{
		(this->*handler)();
	}
	catch(bzp::Exception &e)
	{
		RecoverFailureUsePath(e.GetErrorCode());
	}
	catch(std::bad_alloc)
	{
		RecoverFailureUsePath(BZPERROR_OUT_OF_MEMORY);
	}
	catch(...)
	{
		RecoverFailureUsePath(BZPERROR_UNKNOWN);
	}
}

//pX𗘗pĂŒ̗ȌC
void SampleView::RecoverFailureUsePath(int error)
{
	ShowError(error);
	InvalidateRect();
}

struct ErrorMessage {
	int errorCode;
	UINT id;
};

//G[ԍ當͂ɕϊe[u
static const struct ErrorMessage s_errorMessageID[]={
	{BZPERROR_OUT_OF_MEMORY,IDS_ERR_OUT_OF_MEMORY},
	{BZPERROR_NOT_ENOUGH_BUFFER_SIZE,IDS_ERR_NOT_ENOUGH_BUFFER_SIZE},
	{BZPERROR_READ_BEYOND_BUFFER,IDS_ERR_ILLEGAL_DATA_FORMAT},
	{BZPERROR_ILLEGAL_DATA_FORMAT,IDS_ERR_ILLEGAL_DATA_FORMAT},
	{BZPERROR_CANT_READ_NEW_FORMAT,IDS_ERR_CANT_READ_NEW_FORMAT},
	{BZPERROR_CANT_CREATE_DC,IDS_ERR_CANT_CREATE_DC},
	{BZPERROR_CANT_CREATE_BITMAP,IDS_ERR_CANT_CREATE_BITMAP},
};

#define ERRORMESCOUNT (sizeof(s_errorMessageID)/sizeof(ErrorMessage))

//G[͂̕\
void SampleView::ShowError(int error) const
{
	UINT messageID=IDS_ERR_UNKNOWN;
	for(int i=0;i<ERRORMESCOUNT;++i) {
		if(s_errorMessageID[i].errorCode==error) {
			messageID=s_errorMessageID[i].id;
			break;
		}
	}

	MessageBox(IDS_CAPTION,messageID,GetHWND());
}

//E]
void SampleView::OnHReverse()
{
	GetPath()->HReverse();
}

//㉺]
void SampleView::OnVReverse()
{
	GetPath()->VReverse();
}

//IĂƃTuO[v̃TuO[vɂ܂Ƃ߂
void SampleView::OnGroup()
{
	GetPath()->PutInGroup();
}

void SampleView::OnUpdateGroup(Command *command)
{
	command->SetEnable(bzp::CanGroup(GetPath()) && m_m->tool.CanStop() );
}

//IĂTuO[v΂炷
void SampleView::OnUngroup()
{
	GetPath()->Ungroup();
}

void SampleView::OnUpdateUngroup(Command *command)
{
	command->SetEnable(bzp::CanUngroup(GetPath()) && m_m->tool.CanStop() );
}

#define PENSIZE 32

//pXɂēԊu̓_`
void SampleView::OnSteps()
{
	bzp::Steps steps;
	steps.AddPath(GetPath(),PART_ALL);
	steps.SetInterval(15);	//ߑłĂ邯ǖ{͊Ԋu͉ł

	RECT modifiedRect;
	::SetRect(&modifiedRect,0,0,0,0);

	double x,y;
	while(steps.NextPoint(&x,&y)) {
		PutPen(x,y);

		POINT pointi;
		pointi.x=(int)x;
		pointi.y=(int)y;

		RECT penRect;
		::SetRect(&penRect
			,pointi.x-PENSIZE/2
			,pointi.y-PENSIZE/2
			,pointi.x-PENSIZE/2+PENSIZE
			,pointi.y-PENSIZE/2+PENSIZE);
		::UnionRect(&modifiedRect,&modifiedRect,&penRect);
	}

	RedrawAllViews(&modifiedRect);
}

#define DRAW_OPAQUE 0.5

//yp^[`
void SampleView::PutPen(double x,double y)
{
	/*ʂ͐̂ӂ炩ɂ邽߂1/4ʂ炵
	p^[pӂď肷̂ŁA͏_̍W
	Ă	
	Tvł͂̂܂܏炱̏ꍇ͈ӖȂ*/

	SampleDocument *doc=GetDocument();

	SIZE imageSize=doc->GetSize();

	Image *image=doc->GetImage();

	int left=int(x)-PENSIZE/2;
	int width=PENSIZE;
	int penOffsetX=0;
	if(left<0) {
		penOffsetX=-left;
		width+=left;
		left=0;
	}
	if(left+width>imageSize.cx) {
		width=imageSize.cx-left;
	}
	if(width<=0) return;
	
	int top=int(y)-PENSIZE/2;
	int height=PENSIZE;
	int penOffsetY=0;
	if(top<0) {
		penOffsetY=-top;
		height+=top;
		top=0;
	}
	if(top+height>imageSize.cy) {
		height=imageSize.cy-top;
	}
	if(height<0) return;

	int pixelBytes=image->PixelBytes();
	int rowBytes=image->RowTotalBytes();
	unsigned char *buffer=image->GetRow(top)+left*pixelBytes;

	unsigned char color[3];
	assert(pixelBytes<=3);
	FOR(int i=0;i<pixelBytes;++i) {
		int random=rand();
		color[i]=static_cast<unsigned char>(random);
	}

	FOR(int py=0;py<height;++py) {
		FOR(int px=0;px<width;++px) {
			double distance=hypot(px+penOffsetX-PENSIZE/2,py+penOffsetY-PENSIZE/2)/(PENSIZE/2);
			double depth;
			if(distance<=0.9) {
				depth=1.0*DRAW_OPAQUE;
			} else if(distance<1.0) {
				depth=(1.0-distance)*10*DRAW_OPAQUE;
			} else {
				depth=0;
			}
			FOR(int channel=0;channel<pixelBytes;++channel) {
				unsigned char oldColor=*buffer;
				double newColor=oldColor*(1.0-depth)+color[channel]*depth;
				*buffer++=(unsigned char)newColor;
			}
		}
		buffer+=rowBytes-width*pixelBytes;
	}
}

//pX̓hԂ
void SampleView::OnFill()
{
	bzp::Fill fill;
	fill.AddPath(GetPath(),PART_ALL);

	SampleDocument *doc=GetDocument();

	RECT rect;
	GetPath()->GetOuterRect(&rect,PART_CURVE|PART_SUBGROUP);

	SIZE imageSize=doc->GetSize();
	RECT imageRect;
	::SetRect(&imageRect,0,0,imageSize.cx,imageSize.cy);
	::IntersectRect(&rect,&rect,&imageRect);
	SIZE size;
	size.cx=rect.right-rect.left;
	size.cy=rect.bottom-rect.top;

	Image *image=doc->GetImage();
	const int pixelBytes=image->PixelBytes();

	DimPtr <double> buffer(new double[size.cx]);
	SAMPLE_BAD_ALLOC_TEST(buffer.get());

	double *bufferPtr=buffer.get();

	unsigned char color[3];
	assert(pixelBytes<=3);
	FOR(int i=0;i<pixelBytes;++i) {
		color[i]=(unsigned char)(rand()&0xff);
	}

	FOR(int y=0;y<size.cy;++y) {
		fill.ScanLine(rect.left,rect.top+y,buffer.get(),size.cx);

		unsigned char *row=image->GetRow(rect.top+y);

		FOR(int x=0;x<size.cx;++x) {
			double depth=bufferPtr[x]*DRAW_OPAQUE;
			FOR(int channel=0;channel<pixelBytes;++channel) {

				unsigned char oldColor=row[(rect.left+x)*pixelBytes+channel];
				double newColor=oldColor*(1.0-depth)+color[channel]*depth;
				row[(rect.left+x)*pixelBytes+channel]
					=static_cast<unsigned char>(newColor);
			}
		}
	}

	RedrawAllViews(&rect);
}

//摜̋`\̋`ɕϊ
void SampleView::BoundToView(RECT *viewRect,const RECT *boundRect) const
{
	RECT clientRect;
	GetClientRect(&clientRect);
	int clientWidth=clientRect.right-clientRect.left;
	int clientHeight=clientRect.bottom-clientRect.top;

	SIZE imageSize=GetDocument()->GetSize();

	int left;
	if(MAGNIFY(imageSize.cx)+m_margin.cx*2>clientWidth) {
		left=m_margin.cx+clientRect.left;
	} else {
		left=(clientWidth-MAGNIFY(imageSize.cx))/2+clientRect.left;
	}

	int top;
	if(MAGNIFY(imageSize.cy)+m_margin.cy*2>clientHeight) {
		top=m_margin.cy+clientRect.top;
	} else {
		top=(clientHeight-MAGNIFY(imageSize.cy))/2+clientRect.top;
	}

	viewRect->left=left+MAGNIFY(boundRect->left);
	viewRect->top=top+MAGNIFY(boundRect->top);
	viewRect->right=left+MAGNIFY(boundRect->right);
	viewRect->bottom=top+MAGNIFY(boundRect->bottom);


/*]蒆ɕ\肵Ȃ̂Ȃ
PɃhLg̑傫΂悢*/
}

//pXΎgGUI̍XV
void SampleView::OnUpdateNotEmpty(Command *command)
{
	command->SetEnable(!GetPath()->IsEmpty() && m_m->tool.CanStop());
}

//؂
void SampleView::OnCut()
{
	OnCopy();

	bzp::Cut(GetPath(),GetPath(),m_cutKnot);
}

#define AI_CLIPBOARD_FORMAT "ADOBE AI3"

//
void SampleView::OnCopy()
{
	Clipboard clip;
	if(!clip.Open(GetHWND())) return;
	clip.Clear();

	bzp::Copy(&(m_m->clipboard),GetPath(),m_cutKnot);

	UINT tagFormat=::RegisterClipboardFormat(BZP_CLIPBOARD_FORMAT);
	clip.SetData(tagFormat,NULL);
	UINT aiFormat=::RegisterClipboardFormat(AI_CLIPBOARD_FORMAT);
	clip.SetData(aiFormat,NULL);
}

//\GUI̍XV
void SampleView::OnUpdateCopy(Command *command)
{
	command->SetEnable(bzp::CanCopy(GetPath()) && m_m->tool.CanStop());
}

//\t
void SampleView::OnPaste()
{
	Clipboard clip;
	if(!clip.Open(GetHWND())) return;

	UINT tagFormat=::RegisterClipboardFormat(BZP_CLIPBOARD_FORMAT);
	UINT aiFormat=::RegisterClipboardFormat(AI_CLIPBOARD_FORMAT);

	UINT format;
	if(::IsClipboardFormatAvailable(tagFormat)) {
		format=tagFormat;
	} else if(::IsClipboardFormatAvailable(aiFormat)) {
		format=aiFormat;
	} else {
		return;
	}

	HGLOBAL data=clip.GetData(format);
	char *dataptr=(char*)::GlobalLock(data);
	if(dataptr==NULL) return;
	bzp::Path clipboardPath;
	int size=::GlobalSize(data);
	try {
		if(format==tagFormat) {
			TagToPath(&clipboardPath,dataptr,size);
		} else {
			AIToPath(&clipboardPath,dataptr,size);
		}
	}
	catch(...) {
		::GlobalUnlock(data);
		throw;
	}
	::GlobalUnlock(data);
	clip.Close();

	bzp::Paste(GetPath(),&clipboardPath
		,GetDocument()->GetSize());
}

//\t邩GUI̍XV
void SampleView::OnUpdatePaste(Command *command)
{
	bool isEnabled=false;

	UINT tagFormat=::RegisterClipboardFormat(BZP_CLIPBOARD_FORMAT);
	UINT aiFormat=::RegisterClipboardFormat(AI_CLIPBOARD_FORMAT);
	if(::IsClipboardFormatAvailable(tagFormat)
		||::IsClipboardFormatAvailable(aiFormat)) {
		isEnabled=true;
	}
	command->SetEnable(isEnabled && m_m->tool.CanStop());
}

//Nbv{[h̓eSĂ̌`̃[nhɕϊēo^
void SampleView::OnRenderAllFormats()
{
	RenderTagFormat();	
	RenderAIFormat();
}

//Nbv{[h̓ew肵`ɕϊēo^
void SampleView::OnRenderFormat(UINT format)
{
	UINT tagFormat=::RegisterClipboardFormat(BZP_CLIPBOARD_FORMAT);
	UINT aiFormat=::RegisterClipboardFormat(AI_CLIPBOARD_FORMAT);
	if(format==tagFormat) {
		RenderTagFormat();
	} else if(format==aiFormat) {
		RenderAIFormat();
	}
}

//Nbv{[hɃ^O`œo^
void SampleView::RenderTagFormat()
{
	int size;	//obt@̕Kv

	//pX^O`ɕϊNX
	size=PathToTagSize(&(m_m->clipboard),PART_ALL);

	HGlobal clipData(::GlobalAlloc(GMEM_DDESHARE,size));

	char *clipPtr=(char*)clipData.Lock();
	if(clipPtr==NULL) return;

	PathToTag(&(m_m->clipboard),PART_ALL,clipPtr,size);

	UINT tagFormat=::RegisterClipboardFormat(BZP_CLIPBOARD_FORMAT);
	::SetClipboardData(tagFormat,clipData.Release());
}

//Nbv{[hAI`œo^
void SampleView::RenderAIFormat()
{
	int size;	//obt@̕Kv

	//pX^O`ɕϊNX
	size=PathToAISize(&(m_m->clipboard),PART_ALL);

	HGlobal clipData(::GlobalAlloc(GMEM_DDESHARE,size));

	char *clipPtr=(char*)clipData.Lock();
	if(clipPtr==NULL) return;

	PathToAI(&(m_m->clipboard),PART_ALL,clipPtr,size);

	UINT aiFormat=::RegisterClipboardFormat(AI_CLIPBOARD_FORMAT);
	::SetClipboardData(aiFormat,clipData.Release());
}

//Nbv{[h̓ej
void SampleView::OnDestroyClipboard()
{
	m_m->clipboard.Destroy();	
}

//SĂ̐ƃTuO[vI
void SampleView::OnSelectAll()
{
	GetPath()->SelectAll(true);	
}

//SĂ̐ƃTuO[vIɂ
void SampleView::OnUnselectAll()
{
	GetPath()->SelectAll(false);
}

//Nbv{[h֌pڂŐ؂邩Ő؂邩
void SampleView::OnCutKnot()
{
	m_cutKnot=!m_cutKnot;
}

void SampleView::OnUpdateCutKnot(Command *command)
{
	command->SetCheck(m_cutKnot);
}

//\XV̏
void SampleView::SetRedraw(bzp::Redraw *redraw,DC *dc)
{
	HDC hDC=dc->GetDC();
	OnPrepareDC(hDC);	//foCXReLXgXN[ꍇɃr[̍̈ʒuݒ肵Ă
	m_m->graph.SetViewDC(hDC);
	m_m->graph.SetDrawMode(m_showPath?m_drawMode:DRAWMODE_HIDE_ALL);
	redraw->SetGraph(&(m_m->graph));
	redraw->SetTool(&(m_m->tool));
	redraw->SetPath(GetPath());
}

//ύXn
void SampleView::BeginRedraw(bzp::Redraw *redraw,DC *dc)
{
	SetRedraw(redraw,dc);
	redraw->DrawBefore();
}

//ύXI
void SampleView::EndRedraw(bzp::Redraw *redraw)
{
	m_m->graph.SetDrawMode(m_showPath?m_drawMode:DRAWMODE_HIDE_ALL);
	redraw->End();
}

//p̍ėXgɒǉ
void SampleView::PushHistory()
{
	int size=bzp::DifferenceSize(&(m_m->previousPath)
		,GetPath());
	DimPtr<char> buffer(new char[size]);	//vύX@̃NXgĂ
	SAMPLE_BAD_ALLOC_TEST(buffer.get());

	size=bzp::CreateDifference(&(m_m->previousPath)
		,GetPath(),buffer.get(),size);

	//ύXȂꍇ
	if(size==0) return;

	history::History *history=GetDocument()->GetHistory();

	history->BeginRecord();

	std::auto_ptr<history::Record> record(
		new history::Record(HISTORY_PATH,0,true) );
	SAMPLE_BAD_ALLOC_TEST(record.get());

	record->WriteBuffer(buffer.get(),size);
	buffer.reset(NULL);

	history->Push(record.get());
	record.release();

	history->EndRecord();
}

//ƏĈ߂ɕύXÕpX𑼂̏ꏊɎĂ
void SampleView::SetPreviousPath()
{
	m_m->previousPath.Copy(GetPath());
}

//ɖ߂
void SampleView::OnUndo()
{
	UndoRedo(true);
	POINT point=GetMessagePoint();
	CallEventProcess(EVENT_UNDO,point,0);
}

void SampleView::OnUpdateUndo(Command *command)
{
	command->SetEnable(GetDocument()->GetHistory()->CanUndoRedo(true) && m_m->tool.CanStop());
}

//蒼
void SampleView::OnRedo()
{
	UndoRedo(false);
	POINT point=GetMessagePoint();
	CallEventProcess(EVENT_REDO,point,0);
}

void SampleView::OnUpdateRedo(Command *command)
{
	command->SetEnable(GetDocument()->GetHistory()->CanUndoRedo(false) && m_m->tool.CanStop());
}

//ɖ߂Ƃ蒼̐U蕪
void SampleView::UndoRedo(bool undo)
{
	history::History *history=GetDocument()->GetHistory();
	history::Record *record=history->UndoRedoFirst(undo);

	while(record!=NULL) {
		switch(record->GetKey()) {
		case HISTORY_PATH:
			UndoRedoPath(record,undo);
			break;

		default:
			assert(false);
		}
		record=history->UndoRedoNext();
	}
}

//pX̌ɖ߂Ƃ蒼
void SampleView::UndoRedoPath(history::Record *record,bool undo)
{
	int size;
	const char *diff=record->ReadBuffer(&size);

	bzp::Path *path=GetPath();
	bzp::UndoRedoPath(undo,path,path,diff,size);
}


//hLgSẴr[̍ĕ\
//rect͎W
void SampleView::RedrawAllViews(const RECT *rect)
{
	SendMessageToAllViews(SAMPLE_WM_REDRAW_VIEW
		,(WPARAM)GetDocument(),(LPARAM)rect);
}

//hLgr[̍ĕ\
//rect͎W
void SampleView::OnRedrawView(SampleDocument *document,const RECT *rect)
{
	if(document==GetDocument()) {
		RECT viewRect;
		BoundToView(&viewRect,rect);
		POINT scroll=GetScrollPos();
		::OffsetRect(&viewRect,-scroll.x,-scroll.y);
		InvalidateRect(&viewRect);
	}
}

//t@C̓ǂݍ
void SampleView::OnFileImport()
{
	//t@CI
	FileDialog fileDialog(IDS_OPEN_CAPTION
		,IDS_OPEN_FILTER,GetHWND());
	if(!fileDialog.OpenFileName()) {
		return;
	}

	//t@CJ
	HANDLE file=::CreateFile(fileDialog.GetFileName()
		,GENERIC_READ,FILE_SHARE_READ
		,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);

	if(file==INVALID_HANDLE_VALUE) {
		MessageBox(IDS_CAPTION,IDS_ERR_CANT_OPEN_FILE
			,GetHWND());
		return;
	}

	DWORD size=::GetFileSize(file,NULL);

	char *buffer=(char*)malloc(size);
	if(buffer==NULL) {
		::CloseHandle(file);
		MessageBox(IDS_CAPTION,IDS_ERR_OUT_OF_MEMORY
			,GetHWND());
		return;
	}

	//t@C̓ǂݍ
	//f[^\ǂݍނƂ܂ł̓AvP[Vł
	DWORD readSize;
	if(!::ReadFile(file,buffer,size,&readSize,NULL)) {
		readSize=0;
	}
	::CloseHandle(file);

	if(readSize<size) {
		free(buffer);
		MessageBox(IDS_CAPTION,IDS_ERR_CANT_READ_FILE
			,GetHWND());
		return;
	}

	//pX֕ϊ
	try {
		bzp::Path path;
/*̂Ƃ뗬ʂĂt@C`őΉĂ̂*.aiȂ̂AIToPathg
^O`̓hLgt@CɖߍނƂɎgĂ
*/
		AIToPath(&path,buffer,size);
		Paste(GetPath(),&path,GetDocument()->GetSize());
/*
̃Tv̓yCg\tgz肵Ă̂ŁA
pX̂hLgƂĂ͈킸ɁA炩̉摜tH[}bg
̈ꕔƂĈ߁AJł͂Ȃēǂݍ݂ɂĂ܂
ǂłǂłǡ
ǉÃ݂pX폜ēǂݍނȂ
  AIToPath(GetPath(),buffer,size);
  */
	}
	catch(...) {
		free(buffer);
		throw;
	}
}

//ǂݍ݂̃j[̍XV
void SampleView::OnUpdateImport(Command *command)
{
	command->SetEnable(true);
}

//t@Cɏo
void SampleView::OnFileExport()
{
	FileDialog fileDialog(IDS_SAVE_CAPTION,IDS_SAVE_FILTER
		,GetHWND());
	if(!fileDialog.SaveFileName()) {
		return;
	}

	int size=PathToAISize(GetPath(),PART_ALL);
	DimPtr<char> buffer(new char[size]);
	SAMPLE_BAD_ALLOC_TEST(buffer.get());

	PathToAI(GetPath(),PART_ALL,buffer.get(),size);

	HANDLE file=::CreateFile(fileDialog.GetFileName()
		,GENERIC_WRITE,0
		,NULL,CREATE_ALWAYS
		,FILE_ATTRIBUTE_NORMAL,NULL);

	if(file==INVALID_HANDLE_VALUE) {
		MessageBox(IDS_CAPTION,IDS_ERR_CANT_CREATE_FILE
			,GetHWND());
		return;
	}

	DWORD writeSize;
	if(!::WriteFile(file,buffer.get(),size,&writeSize,NULL)) {
		writeSize=0;
	}
	::CloseHandle(file);

	if(writeSize<size) {
		MessageBox(IDS_CAPTION,IDS_ERR_CANT_WRITE_FILE
			,GetHWND());
	}
}


SAMPLE_NAMESPACE_END
