#include "prefix.h"

#include "winman.h"
#include "frame.h"
#include "document.h"
#include "view.h"
#include "resource.h"

/*************************************************

	r[EBhE{NX

**************************************************/

SAMPLE_NAMESPACE_START

View::View()
:m_document(NULL)
{

}

View::~View() throw()
{
	if(m_document!=NULL) {
		m_document->Unlock();
	}
}

void View::SetDocument(Document *document)
{
	assert(document!=NULL && m_document==NULL);
	m_document=document;
	m_document->Lock();
	m_viewNumber=document->GetViewNumber();
}

bool View::Create(Frame *frame)
{
	if(!Window::Create(NULL,WS_VISIBLE|WS_CHILD|WS_HSCROLL|WS_VSCROLL,frame)) {
		return false;
	}
	OnInitialUpdate();
	return true;
}

bool View::WindowProc(UINT msg,WPARAM wparam,LPARAM lparam)
{
	switch(msg) {
	case WM_HSCROLL:
		OnHScroll((int)LOWORD(wparam),(short int)HIWORD(wparam),HWND(wparam));
		break;

	case WM_VSCROLL:
		OnVScroll((int)LOWORD(wparam),(short int)HIWORD(wparam),HWND(wparam));
		break;

	case WM_SIZE:
		OnSize((UINT)wparam,LOWORD(lparam),HIWORD(lparam));
		break;

	case WM_COMMAND:
		OnCommand(wparam,lparam);
		break;

	default:
		return false;
	}
	return true;
}

Document *View::GetDocument()
{
	return m_document;
}

const Document *View::GetDocument() const
{
	return m_document;
}

Frame *View::GetFrame()
{
	return m_frame;
}

void View::SetFrame(Frame *frame)
{
	m_frame=frame;
}

#define SCROLLLINESIZE 16
#define SCROLLPAGESIZE 64

void View::SetScrollSize(SIZE size)
{
	m_scrollSize=size;

	SCROLLINFO info;
	info.cbSize=sizeof(info);
	info.fMask=SIF_POS|SIF_PAGE|SIF_RANGE;
	info.nPage=SCROLLPAGESIZE;
	info.nPos=0;
	info.nMin=0;
	info.nMax=0;
	SetScrollInfo(SB_HORZ,&info,FALSE);
	SetScrollInfo(SB_VERT,&info,FALSE);

	OnSize(0,0,0);
}

void View::OnHScroll(int scrollCode,short int pos,HWND)
{
	OnScroll(scrollCode,pos,SB_HORZ);
}

void View::OnVScroll(int scrollCode,short int pos,HWND)
{
	OnScroll(scrollCode,pos,SB_VERT);
}

void View::OnScroll(int scrollCode,short int pos,int bar)
{
	SCROLLINFO scrollInfo;
	scrollInfo.cbSize=sizeof(SCROLLINFO);
	scrollInfo.fMask=SIF_POS|SIF_RANGE|SIF_PAGE;
	GetScrollInfo(bar,&scrollInfo);

	int scrollMax=scrollInfo.nMax;

	int oldPos=scrollInfo.nPos;
	int newPos;
	switch(scrollCode) {
	default:
		return;

	case SB_BOTTOM:
		newPos=scrollMax;
		break;

	case SB_LINEDOWN:
		newPos=oldPos+SCROLLLINESIZE;
		break;

	case SB_LINEUP:
		newPos=oldPos-SCROLLLINESIZE;
		break;

	case SB_PAGEDOWN:
		newPos=oldPos+scrollInfo.nPage;
		break;

	case SB_PAGEUP:
		newPos=oldPos-scrollInfo.nPage;
		break;
		
	case SB_THUMBPOSITION:
	case SB_THUMBTRACK:
		newPos=pos;
		break;

	case SB_TOP:
		newPos=0;
		break;
	}
	if(newPos<0) {
		newPos=0;
	} else if(newPos>scrollMax) {
		newPos=scrollMax;
	}

	int dx,dy;
	if(bar==SB_HORZ) {
		dx=oldPos-newPos;
		dy=0;
	} else {
		dx=0;
		dy=oldPos-newPos;
	}

	RECT clientRect;
	GetClientRect(&clientRect);

	ScrollWindowEx(dx,dy,&clientRect,NULL,NULL,NULL,SW_INVALIDATE|SW_ERASE);

	SCROLLINFO newScrollInfo;
	newScrollInfo.cbSize=sizeof(SCROLLINFO);
	newScrollInfo.fMask=SIF_POS;
	newScrollInfo.nPos=newPos;
	SetScrollInfo(bar,&newScrollInfo,TRUE);
	UpdateWindow();
}

void View::OnSize(UINT,WORD,WORD)
{
	RECT clientRect;
	GetClientRect(&clientRect);

	OnSizeSub(SB_HORZ,Width(clientRect),m_scrollSize.cx);
	OnSizeSub(SB_VERT,Height(clientRect),m_scrollSize.cy);
}

void View::OnSizeSub(int bar,int clientSize,int scrollSize)
{
	SCROLLINFO info;
	info.cbSize=sizeof(SCROLLINFO);
	info.fMask=SIF_POS;
	GetScrollInfo(bar,&info);

	if(info.nPos+clientSize>scrollSize) {
		info.nPos=scrollSize-clientSize;
	}
	info.nMin=0;
	info.nMax=scrollSize-clientSize;
	if(info.nMax<=0) {
		info.nMax=0;
	}
	info.fMask=SIF_POS|SIF_RANGE;
	SetScrollInfo(bar,&info,TRUE);
}

POINT View::GetScrollPos() const
{
	RECT clientRect;
	GetClientRect(&clientRect);

	POINT point;
	point.x=GetScrollPosSub(SB_HORZ,Width(clientRect),m_scrollSize.cx);
	point.y=GetScrollPosSub(SB_VERT,Height(clientRect),m_scrollSize.cy);
	return point;
}

int View::GetScrollPosSub(int bar,int clientSize,int scrollSize) const
{
	int pos;
	if(clientSize>=scrollSize) {
		pos=0;
	} else {
		SCROLLINFO scrollInfo;
		scrollInfo.cbSize=sizeof(SCROLLINFO);
		scrollInfo.fMask=SIF_POS;
		GetScrollInfo(bar,&scrollInfo);
		pos=scrollInfo.nPos;
	}
	return pos;
}

void View::OnPrepareDC(HDC hDC)
{
	POINT scroll=GetScrollPos();
	::SetViewportOrgEx(hDC,-scroll.x,-scroll.y,NULL);
}

void View::OnInitialUpdate()
{
}

void View::OnCommand(WPARAM wparam,LPARAM lparam)
{
	switch(LOWORD(wparam)) {
	case ID_WINDOW_NEWWINDOW:
		CreateAlternateWindow(GetDocument());
		break;
	}
}

void View::OnUpdateCommand(UINT id,Command *command)
{

}

void View::SendMessageToAllViews(UINT message,WPARAM wparam,LPARAM lparam)
{
	SAMPLENAME::SendMessageToAllViews(message,wparam,lparam);
}

void View::SetStatusText(UINT textID)
{
	GetFrame()->SetStatusText(textID);
}

int View::GetViewNumber() const
{
	return m_viewNumber;
}

SAMPLE_NAMESPACE_END
