#include "prefix.h"

#include <windows.h>

#include "window.h"
#include "main.h"
#include "resource.h"

/*************************************************

	EBhENX

**************************************************/

SAMPLE_NAMESPACE_START

//EChEo^
static const char *s_windowClassName="SampleWindow";

LRESULT CALLBACK PreAttachWindowProc(HWND,UINT,WPARAM,LPARAM);
LRESULT CALLBACK SampleWindowProc(HWND,UINT,WPARAM,LPARAM);

LRESULT CALLBACK PreAttachWindowProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam)
{
	return ::DefWindowProc(hwnd,msg,wparam,lparam);
}

LRESULT CALLBACK SampleWindowProc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam)
{
	Window *window=(Window*)::GetWindowLong(hwnd,GWL_USERDATA);
	if(window->WindowProc(msg,wparam,lparam)) {
		return 0;
	}
	return ::DefWindowProc(hwnd,msg,wparam,lparam);
}

Window::Window()
{
}

Window::~Window() throw()
{

}

void Window::RegisterWindowClass()
{
	static bool registered=false;
	if(registered) return;

	registered=true;;

	WNDCLASS wc;

	wc.lpszClassName=s_windowClassName;
	wc.lpfnWndProc=(WNDPROC)PreAttachWindowProc;
	wc.style=CS_OWNDC|CS_VREDRAW|CS_HREDRAW;

	wc.hInstance=g_hInstance;

	wc.hIcon=::LoadIcon(g_hInstance,MAKEINTRESOURCE(IDR_MAIN));
	wc.hCursor=::LoadCursor(NULL,IDC_ARROW);
	wc.hbrBackground=reinterpret_cast<HBRUSH>(COLOR_APPWORKSPACE+1);
	wc.lpszMenuName=MAKEINTRESOURCE(IDR_MAIN);
	wc.cbClsExtra=0;
	wc.cbWndExtra=0;

	::RegisterClass(&wc);
}

bool Window::Create(const char *caption,UINT style,Window *parentWindow)
{
	RegisterWindowClass();

	HWND hwnd=::CreateWindow(s_windowClassName
	,caption
	,style
	,CW_USEDEFAULT,CW_USEDEFAULT
	,CW_USEDEFAULT,CW_USEDEFAULT
	,parentWindow->GetHWND()
	,NULL
	,g_hInstance
	,NULL);
	if(hwnd==NULL) return false;

	m_hwnd=hwnd;
	SetWindowLong(GWL_USERDATA,(LONG)this);
	SetWindowLong(GWL_WNDPROC,(LONG)SampleWindowProc);

	return true;
}

bool Window::WindowProc(UINT msg,WPARAM wparam,LPARAM lparam)
{
	switch(msg) {
	case WM_NCDESTROY:
		OnNcDestroy();
		break;

	default:
		return false;
	}
	return true;
}

BOOL Window::ShowWindow(int mode)
{
	return ::ShowWindow(GetHWND(),mode);
}

BOOL Window::DestroyWindow()
{
	return ::DestroyWindow(GetHWND());
}

BOOL Window::GetClientRect(RECT *rect) const
{
	return ::GetClientRect(GetHWND(),rect);
}

BOOL Window::SetWindowPos(Window *wndInsertAfter,int x,int y,int cx,int cy,UINT flags)
{
	return ::SetWindowPos(GetHWND(),wndInsertAfter->GetHWND(),x,y,cx,cy,flags);
}

LONG Window::SetWindowLong(int nIndex,LONG dwNewLong)
{
	return ::SetWindowLong(GetHWND(),nIndex,dwNewLong);
}

HWND Window::GetHWND() const
{
	if(this==NULL) {
		return NULL;
	}
	return m_hwnd;
}

void Window::OnNcDestroy()
{
	delete this;
}

BOOL Window::SetScrollInfo(int fnBar,SCROLLINFO *pSi,BOOL redraw)
{
	return ::SetScrollInfo(GetHWND(),fnBar,pSi,redraw);
}

BOOL Window::GetScrollInfo(int fnBar,SCROLLINFO *pSi) const
{
	return ::GetScrollInfo(GetHWND(),fnBar,pSi);
}

int Window::ScrollWindowEx(int dx,int dy,const RECT *scroll
		,const RECT *clip,HRGN updateRegion
		,RECT *updateRect,UINT flags)
{
	return ::ScrollWindowEx(GetHWND(),dx,dy,scroll
		,clip,updateRegion,updateRect,flags);
}

BOOL Window::UpdateWindow()
{
	return ::UpdateWindow(GetHWND());
}

LRESULT Window::SendMessage(UINT message,WPARAM wparam,LPARAM lparam)
{
	return ::SendMessage(GetHWND(),message,wparam,lparam);
}

BOOL Window::PostMessage(UINT message,WPARAM wparam,LPARAM lparam)
{
	return ::PostMessage(GetHWND(),message,wparam,lparam);
}

BOOL Window::InvalidateRect(const RECT *rect)
{
	return ::InvalidateRect(GetHWND(),rect,TRUE);
}

BOOL Window::ScreenToClient(POINT *point) const
{
	return ::ScreenToClient(GetHWND(),point);
}

BOOL Window::ClientToScreen(POINT *point) const
{
	return ::ClientToScreen(GetHWND(),point);
}

HWND Window::SetCapture()
{
	return ::SetCapture(GetHWND());
}

BOOL Window::DragDetect(POINT point)
{
	return ::DragDetect(GetHWND(),point);
}

BOOL Window::SetWindowText(const char *text)
{
	return ::SetWindowText(GetHWND(),text);
}

SAMPLE_NAMESPACE_END



